/*
 * Copyright (c) 2004 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.
 */

/* ----------------------------------------------------------------------
 *
 * AspectLib     -   Fast aspect map computation (nearly O(n))
 *
 * Main function : aspectComputesAspect
 *
 * Computes the list of segments visible from a given view point.
 *
 * Author : Sara Fleury - April 2004    
 * Sara.Fleury@laas.fr
 *
 *
 * PRINCIPLE : 
 * -----------
 * Every segment is incrementally added in a polar representation of the
 * aspect map. The aspect map is represented as an ordinate list of angles,
 * the extremities of the segments : every angle is at least a start or an
 * end of a segment (can be both in case of occultation).
 * 
 * When a segment [(t1,r1), (t2, r2)] is added  in this list, we
 * check for every angle t within [t1, t2] if the new segment hides or is
 * hiden by the segment(s) carries by t. In the first situation t is removed,
 * in the second one the new segment is cut. Special attention is required
 * at the new segment extremities.
 * 
 * KNOWN LIMITATION :
 * ------------------
 * The computation is exact if the segments do not cross each other.
 * Indeed, we consider that a segment is in front of another one
 * within ALL the angular map sector [ti,ti+1] if it is in front at ]ti]. 
 * Crossing WITHIN ]ti,ti+1[ will not be detected (however the order will
 * be corrected from angle ]ti+1]).
 * 
 * ALGORITHM PRINCIPLE :
 * ---------------------
 * Map = list of ordinate angles t with for each of them the id of the
 * current nearest (ie, visible) following segment carries by this angle (if any).
 * 
 * 1/ If an angle t carries neither a preceeding visible segment nor a following
 * visible segment (other than the new segment) it can be removed ; else ,
 * 2/ If the following segment is not any more visible it is replaced
 *    by the new one.
 * 
 * 
 * ALGORITHM IN DEPTH : 3 steps have to be considered :
 * --------------------
 * Step 1/ introduction of first extremity t1 has 3 cases :
 *   1/ t1 does not cut any segment -> t1 added
 *   2/ t1 cuts a seg but the new one is closest -> t1 added
 *   3/ t1 cuts a seg but the new one is farther -> t1 not added
 * 
 * Step 2/ intersection with the angles within ]t1,t2[ has 8 cases 
 *   according if t carries a previous and/or a next seg and if the(se)
 *   seg(s) are in front or behind the new one (the situation with neither
 *   a previous nor a following seg is not possible) :
 *   1/ No previous seg :
 *      1.1/ next seg behind new one -> t removed
 *      1.2/ next seg in front of new one -> t unchanged
 *   2/ Prev seg behind new one :
 *      2.1/ no following seg -> t removed
 *      2.2/ followig seg behind -> t removed
 *      2.3/ followig seg in front -> t unchanged
 *   3/ Prev seg in front of new one :
 *      3.1/ no following seg -> t carries the new seg
 *      3.2/ followig seg behind -> t carries the new seg
 *      3.3/ followig seg in front -> t unchanged
 *    
 * Step 3/ introduction of second extremity t2 has 3 cases :
 *   1/ t2 does not cut any segment -> t2 added
 *   2/ t2 cuts a seg but the new one is closest 
 *      -> t2 added and carries the OLD one
 *   3/ t2 cuts a seg but the new one is farther -> t not added
 *
 * ----------------------------------------------------------------------
 */

/* ---------------------------------------------------------------------- */
/* Makes the algorithm very verbose for debug purpose                     */
#define DEBUG 0
#ifndef VXWORKS
/* #define MESURE_TIME */
#endif

/* ---------------------------------------------------------------------- */


#ifndef VXWORKS
#include <portLib.h>
#include <sys/time.h>
#include <math.h>
#else
#include <vxWorks.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/*#include "h2mathLib.h"*/

#include "aspectLib.h"

#define ASPECT_DIST_INFINITY 10000

static int nbMaxSegs=0;

/* -- LOCAL STRUCTURES --------------------------------------------------
 */

typedef struct SEG_POLAR {
  ASPECT_POLAR_SEG polar;
  int seg;
} SEG_POLAR;

typedef struct ASPECT_ANGLE {
  double theta;
  int segN;
  double rhoP;
  double rhoN;
  struct ASPECT_ANGLE *next;
  struct ASPECT_ANGLE *prev;
} ASPECT_ANGLE;

/* -- LOCAL VARIABLES --------------------------------------------------
 */

static ASPECT_SEG *localCartMap=NULL;
static int *reversedSegs=NULL;

static ASPECT_ANGLE * aspectMap=NULL;
static ASPECT_ANGLE * aspectMapLast=NULL;
static int nbAngles=0;
static int nbAspectSegs;

/* -- LOCAL FUNCTIONS -------------------------------------------------- 
 */
#include "aspectLibStaticProto.h"

static int tmpNbMapSegs;
static BOOL testBugZero()
{
  BOOL zero=FALSE;
  int i;
  return FALSE;
  for (i=0; i<tmpNbMapSegs; i++) {
    if (fabs(localCartMap[i].x1) < 0.0001) {
      zero=TRUE;
      printf ("************* testZero : localCartMap[%d].x1 == 0\n", i);
    }
    if (fabs(localCartMap[i].x2) < 0.0001) {
      zero=TRUE;
      printf ("************* testZero : localCartMap[%d].x2 == 0\n", i);
    }
    if (fabs(localCartMap[i].y1) < 0.0001) {
      zero=TRUE;
      printf ("************* testZero : localCartMap[%d].y1 == 0\n", i);
    }
    if (fabs(localCartMap[i].y2) < 0.0001) {
      zero=TRUE;
      printf ("************* testZero : localCartMap[%d].y2 == 0\n", i);
    }
  }
  if (zero)
    return TRUE;
  return FALSE;
}

/* ---------------------------------------------------------------------- */

/* XXXXXX should be get from laser */
#define ASPECT_LASER_MAX_DIST 32 

/* ----------------------------------------------------------------------
 *
 * aspectSimulLaser - 
 *                    
 *
 *
 */

int aspectSimulLaser(double thMin, double thMax, double dth, 
		     int nbSegs,
		     ASPECT_POLAR_SEG *polar, ASPECT_SEG *cartLocal, 
		     int nbEchos, double *rho)
{
  int i, seg;
  double th;

  /* starts from thMin */
  th = thMin;
  seg = 0;
  i = 0;

  /* for each seg within [thMin, thMax] */
  while (seg<nbSegs && th < thMax) {

    /* after th and before ext1 : no echos */
    while (th < polar[seg].t1 && th < thMax) {
      if (i==nbEchos) {
	printf ("aspectSimulLaser: too many echos, stop at %d (%f/t1=%f)\n",
		i, th, polar[seg].t1);
	return 1;
      }
      rho[i++] = ASPECT_LASER_MAX_DIST;
      th += dth;
    }

    /* angle within [ext1,ext2] of seg : computes echo */
    while (th < polar[seg].t2 && th < thMax) {
      if (i==nbEchos) {
	printf ("aspectSimulLaser: too many echos, stop at %d (%f/t2 = %f)\n",
		i, th,  polar[seg].t2);
	return 1;
      }
      rho[i++] = angleDistToSeg(&cartLocal[seg], th);
      th += dth;
    }
    seg++;
  }

  /* after ext2 of last seg and before thMax : no echos */
  while (th < thMax) {
    if (i==nbEchos) {
      printf ("aspectSimulLaser: too many echos, stop at %d (%f/max = %f)\n",
	      i, th, thMax);
      return 1;
    }
    rho[i++] = ASPECT_LASER_MAX_DIST;
    th += dth;
  }

#if DEBUG>=1
    printf ("aspectSimulLaser: %d echos\n", i);
#endif
  return TRUE;
}


/* ----------------------------------------------------------------------
 *
 * aspectMapFillIn  - main function with computed arrays pre-allocated
 *
 *
 */

int aspectMapFillIn(ASPECT_CONFIG *sensorViewPoint, 
		    ASPECT_VIEW_PARAMETERS *viewParameters,
		    int nbInputMapSegs, 
		    const ASPECT_SEG cartMap[],
		    int outputCartInGlobalFrameFlag,
		    int aspectOutputTabSize,
		    ASPECT_SEG aspectCartToFill[],
		    ASPECT_POLAR_SEG aspectPolarToFill[],
		    int inputSegsId[],
		    double angleStep,
		    double *rho)
{
  return aspectComputesAspect(sensorViewPoint, 
			      viewParameters,
			      nbInputMapSegs, 
			      cartMap,
			      outputCartInGlobalFrameFlag,
			      aspectOutputTabSize,
			      aspectCartToFill,
			      aspectPolarToFill,
			      inputSegsId,
			      NULL, NULL, NULL,
			      angleStep,
			      rho);
}

/* ----------------------------------------------------------------------
 *
 * aspectMapAlloc  -  main function with cumputed arrays allocated locally
 *
 *
 */

int aspectMapAlloc(ASPECT_CONFIG *sensorViewPoint, 
		   ASPECT_VIEW_PARAMETERS *viewParameters,
		   int nbInputMapSegs, 
		   const ASPECT_SEG cartMap[],
		   int outputCartInGlobalFrameFlag,
		   ASPECT_SEG **aspectCartToAlloc,
		   ASPECT_POLAR_SEG **aspectPolarToAlloc,
		   int **inputSegsId,
		   double angleStep,
		   double *rho)
{
  return aspectComputesAspect(sensorViewPoint, 
			      viewParameters,
			      nbInputMapSegs, 
			      cartMap,
			      outputCartInGlobalFrameFlag,
			      0, NULL, NULL, NULL,
			      aspectCartToAlloc,
			      aspectPolarToAlloc,
			      inputSegsId,
			      angleStep,
			      rho);
}

/* ----------------------------------------------------------------------
 *
 * aspectComputesAspect - Main function that computes an aspect map 
 *                        from a map of segments.
 *
 *
 */

int aspectComputesAspect(ASPECT_CONFIG *sensorViewPoint, 
			 ASPECT_VIEW_PARAMETERS *viewParameters,
			 int nbInputMapSegs, 
			 const ASPECT_SEG cartMap[],
			 int outputCartInGlobalFrameFlag,
			 int aspectOutputTabSize,
			 ASPECT_SEG aspectCartToFill[],
			 ASPECT_POLAR_SEG aspectPolarToFill[],
			 int inputSegsIdToFill[],
			 ASPECT_SEG **aspectCartToAlloc,
			 ASPECT_POLAR_SEG **aspectPolarToAlloc,
			 int **inputSegsIdToAlloc,
			 double angleStep,
			 double *rho)
{
  ASPECT_POLAR_SEG *aspectPolarLocal=NULL;
  BOOL filterAngleView=FALSE;
  double minAngleView=-M_PI, maxAngleView=M_PI;
  int *inputSegsId = NULL;

#ifdef MESURE_TIME
  struct timeval timebegin;
  struct timeval timeend;

  gettimeofday(&timebegin, NULL);
#endif

  /* empty map */
  if (!nbInputMapSegs) {
    printf ("aspectComputesAspect: empty map\n");
    return 0;
  }

  tmpNbMapSegs = nbInputMapSegs;

  /* filter according to angle of view */
  if (viewParameters->halfAngleView >= M_PI)
    viewParameters->halfAngleView = 0;

  /* global to local */
  if (!aspectGlobalToLocal(sensorViewPoint, nbInputMapSegs, cartMap)) {
    fprintf(stderr, "aspectComputesAspect: aspectGlobalToLocal failed\n");
    return 0;
  }

  /* preparing angle of view filtering */
  if (viewParameters->halfAngleView != 0) {
    filterAngleView = TRUE;
    minAngleView = -viewParameters->halfAngleView;
    maxAngleView = viewParameters->halfAngleView;
  }
  else {
    minAngleView = -M_PI;
    maxAngleView = M_PI;
  }

  /* main computation */
  aspectComputMain(nbInputMapSegs, viewParameters->maxDistView, 
		   filterAngleView, minAngleView, maxAngleView);

  /* fill in output polar aspect maps */
  if (!aspectPolarToFill && !aspectPolarToAlloc) 
    aspectPolarToAlloc = &aspectPolarLocal;
  if (!fillInPolarMap (&nbAspectSegs, 
		       viewParameters->minAspectSegAngle,
		       viewParameters->keepReversedSegs,
		       aspectOutputTabSize, 
		       aspectPolarToFill, 
		       aspectPolarToAlloc,
		       inputSegsIdToFill,
		       inputSegsIdToAlloc,
		       &inputSegsId))
    return 0;

  /* fill in output cartesian aspect maps */
  polarToCart (nbAspectSegs, aspectCartToFill, aspectCartToAlloc, 
	       aspectPolarToFill ? aspectPolarToFill : *aspectPolarToAlloc,
	       inputSegsId);
  if (aspectPolarLocal) free(aspectPolarLocal);

  /* simul laser */ 
  if (rho)
    aspectSimulLaser(minAngleView, maxAngleView, angleStep, nbAspectSegs, 
		     aspectPolarToFill ? aspectPolarToFill : *aspectPolarToAlloc,
		     aspectCartToFill ? aspectCartToFill : *aspectCartToAlloc,
		     (int)((maxAngleView-minAngleView)/angleStep + 0.5), rho);

  /* convert back to global map */
  if (outputCartInGlobalFrameFlag) {
    if (sensorViewPoint->x != 0 || sensorViewPoint->y != 0 || 
	sensorViewPoint->theta != 0)
      localToGlobal(sensorViewPoint, nbAspectSegs, 
		    aspectCartToFill ? aspectCartToFill : *aspectCartToAlloc);
  }
  testBugZero();

  /* free aspect map */
  freeAspectMap();
  if (!inputSegsIdToFill && !inputSegsIdToAlloc) {
    free(inputSegsId);
  }

#ifdef MESURE_TIME
  gettimeofday(&timeend, NULL);
  printf ("%5.4lf ms\n", (double)(timeend.tv_sec*1000000 + timeend.tv_usec - 
				  (timebegin.tv_sec * 1000000 + timebegin.tv_usec))/1000);
#endif

  return nbAspectSegs;
}


/* ----------------------------------------------------------------------
 *
 * aspectComputMain - 
 *                    
 *
 *
 */

static int aspectComputMain(int nbInputMapSegs,
			    double maxDistView, 
			    BOOL filterAngleView,
			    double minAngleView, 
			    double maxAngleView)
{
  int segNum, oldSegN;
  int segPrev;
  SEG_POLAR segPolar;
  ASPECT_ANGLE *angleP;
  ASPECT_SEG *segCart;
  double rhoSeg;
  BOOL moduloList, moduloAngle;

  /* for each seg */
  for(segNum=0; segNum<nbInputMapSegs; segNum++) {

    segCart = &localCartMap[segNum];
    
    /** --------------------------------------------------
     ** Convert from cartesian to polar point. 
     ** Filter according to dist and angle of view.  
     **/

#if DEBUG>=2
    printf ("\n================== SEG %d (%g %g) (%g %g)\n", 
	    segNum, segCart->x1, segCart->y1, segCart->x2, segCart->y2);
#endif
    
    segPolar.seg = segNum;

    if (!aspectCartToPolarFilter(segCart, &segPolar.polar, 
				 maxDistView, 
				 filterAngleView,
				 minAngleView,
				 maxAngleView,
				 &reversedSegs[segNum]))
      continue;

    /** --------------------------------------------------
     ** Try to insert extremity 1 of the segment
     **/

    /* first segment of the aspect map */
    if (!aspectMap) {
      initAspectMap(&segPolar);
      continue;
    }

    /* look for the previous angle angleP */
    angleP = lookForPrevSeg(segPolar.polar.t1);
    oldSegN = -1;

    /* no previous seg */
    if (angleP->segN == -1) {
      angleP = insert1After(&segPolar, angleP, ASPECT_DIST_INFINITY);
    }

    /* there is a previous seg */
    else {

      /* that is farther (but cn still comming back) */
      if (!segCloser(&localCartMap[angleP->segN], segPolar.polar.t1, 
		     segPolar.polar.r1, &rhoSeg)) {
	oldSegN = angleP->segN;
	/* insert angle */
	angleP = insert1After(&segPolar, angleP, rhoSeg);
      }
      else {
#if DEBUG>=3
	printf("\n-- seg %d closer than ext1 %g\n", 
	       angleP->segN, segPolar.polar.t1);
#endif
      }
    }
    
    /* for the following angle, next seg becomes prev seg */
    segPrev = angleP->segN;
    
    /** --------------------------------------------------
     ** Test intersections of the segment with following
     ** polar angles.
     **/

    moduloList = FALSE;
    moduloAngle = FALSE;

    /* in case of second extremity angle greater than M_PI */
    if (segPolar.polar.t2 < segPolar.polar.t1) 
      moduloAngle = TRUE;

    /* if already end of list, loop already at begining with modulo */
    if (angleP == aspectMapLast)
      moduloList = TRUE;
    testBugZero();
    /* loop over angles intersected by the segment */
    while (angleP->next->theta + moduloList*2*M_PI < segPolar.polar.t2 + moduloAngle*2*M_PI) {
      
      /* considered angle */
      angleP = angleP->next;
      
      /* the previous seg was the new one (only one or closer) */
      if (/* segPrev */ /* XXX */ angleP->prev->segN == segPolar.seg) {

	/* the next seg will be also the new one (closer or only one) */
	if (segCloser(segCart, angleP->theta, angleP->rhoN, &angleP->rhoP)) {

	  /* then this angle is useless (however the segs can reappears later on) */
	  oldSegN = angleP->segN;
	  angleP = removeAspectAngle(angleP);
	}

	/* there is another closer following seg (lets forget the old seg behind) */
	else {
	  oldSegN = -1;
	}
      }

      /* the previous seg is not changed because closer */
      else {

	/* but the next one is the new one */
	/* because only one */
	if (angleP->segN == -1) {
	  oldSegN = -1; /* no hidden seg */
	  angleP->rhoN = angleDistToSeg(segCart, angleP->theta);
	  angleP->segN = segPolar.seg;
	}
	/* or because closer */
	else if (segCloser(segCart, angleP->theta, angleP->rhoN, &rhoSeg)) {
	  oldSegN = angleP->segN;  /* the hidden seg can reappears later */
	  angleP->rhoN = rhoSeg;
	  angleP->segN = segPolar.seg;
	}
	/* else the new one is not visible :  nothing to do */
	else {
	  oldSegN = -1; /* no seg hidden by new one */
	}
      }

      /* next angle */
      segPrev = angleP->segN;
	    
      /* start an other turn on trigo circle */
      if (angleP == aspectMapLast) 
	moduloList = TRUE;
      
    } /* while preeceding angles */


    /** --------------------------------------------------
     ** Finaly, try to insert extremity 2 of the segment
     **/

#if DEBUG>=3
    printf ("segPrev = %d    seg = %d     segN = %d  oldSegN = %d\n",
	    /* angleP->prev->segN */ segPrev , segPolar.seg, angleP->segN, oldSegN);
#endif

    /* the last seg is the new one (only one or closer) */
    if (angleP->segN  /* segPrev */  == segPolar.seg) {

      angleP = insert2After(&segPolar, angleP);

      /* however, there was another seg there that is only partly occulted now */
      if (oldSegN /* angleP->prev->segN */ != -1) {
	angleP->segN = oldSegN /* angleP->prev->segN */;
	angleP->rhoN = angleDistToSeg(&localCartMap[/* angleP->prev->segN  */oldSegN], angleP->theta);
      }
    }
    testBugZero();
  } /* for each segment */

  return 1;
}

/* ----------------------------------------------------------------------
 *
 * polarToCart
 *
 *
 */
static BOOL polarToCart(int nbSegs, 
			ASPECT_SEG aspectCartToFill[], 
			ASPECT_SEG **aspectCartToAlloc,
			ASPECT_POLAR_SEG aspectPolar[],
			int *inputSegsId)
{
  double cost, sint;
  int i;
  ASPECT_SEG *acl;

  /* no address to fill in */
  if (!aspectCartToFill && !aspectCartToAlloc) {
    printf ("polarToCart: no cartesian aspect map required\n");
    return FALSE;
  }
  
  /* address but not allocated yet */
  if (!aspectCartToFill) {
    if(!(*aspectCartToAlloc = malloc(nbSegs*sizeof(ASPECT_SEG)))) {
      printf ("polarToCart: failled alloc aspectCart\n");
      return FALSE;
    }
    acl = *aspectCartToAlloc;
  }
  else {
    acl = aspectCartToFill;
  }

#if DEBUG>=1
  printf ("\n-- local cartesian aspect map :\n");
#endif

  /* fill in cart */
  for (i=0; i<nbSegs; i++) {

    /* ext 1 */
    cost = cos(aspectPolar[i].t1);
    sint = sin(aspectPolar[i].t1);
    if (!reversedSegs[inputSegsId[i]]) {
      acl->x1 = aspectPolar[i].r1 * cost;
      acl->y1 = aspectPolar[i].r1 * sint;
    }
    else {
      acl->x2 = aspectPolar[i].r1 * cost;
      acl->y2 = aspectPolar[i].r1 * sint;
    }

    /* ext 2 */
    cost = cos(aspectPolar[i].t2);
    sint = sin(aspectPolar[i].t2);
    if (!reversedSegs[inputSegsId[i]]) {
      acl->x2 = aspectPolar[i].r2 * cost;
      acl->y2 = aspectPolar[i].r2 * sint;
    }
    else {
      acl->x1 = aspectPolar[i].r2 * cost;
      acl->y1 = aspectPolar[i].r2 * sint;
    }

#if DEBUG>=1
    printf ("%2d : (%g %g) (%g %g)  %s\n", 
	    i,
	    acl->x1,    acl->y1,	    
	    acl->x2,    acl->y2,
	    reversedSegs[inputSegsId[i]] ? "REVERSED":"not reversed");
#endif

    acl++;
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
 *
 *  localToGlobal
 *
 *  
 */
static void localToGlobal(ASPECT_CONFIG *sensorViewPoint, int nbseg, 
			  ASPECT_SEG *aspectCart)
{
  double cosView, sinView;
  double x, y;
  int i;

  cosView = cos(sensorViewPoint->theta);
  sinView = sin(sensorViewPoint->theta);

#if DEBUG>=3
  printf ("\n-- global cart :\n");
#endif

  for(i=0; i<nbseg; i++) {
    x = aspectCart[i].x1;
    y = aspectCart[i].y1;
    aspectCart[i].x1 = sensorViewPoint->x + x * cosView - y * sinView;
    aspectCart[i].y1 = sensorViewPoint->y + y * cosView + x * sinView;
    x = aspectCart[i].x2;
    y = aspectCart[i].y2;
    aspectCart[i].x2 = sensorViewPoint->x + x * cosView - y * sinView;
    aspectCart[i].y2 = sensorViewPoint->y + y * cosView + x * sinView;

#if DEBUG>=3
    printf ("%d : (%g %g) (%g %g)\n", i,
	    (aspectCart)[i].x1,    (aspectCart)[i].y1,	    
	    (aspectCart)[i].x2,    (aspectCart)[i].y2);
#endif

  }

}

/* ----------------------------------------------------------------------
 *
 *  printPolarMap
 *
 * 
 */
#if DEBUG>=3
static void printPolarMap (void)
{
  ASPECT_ANGLE *aa;
  int i;

  printf ("head = %g\n", aspectMap->theta);
  for (aa = aspectMap, i=0; aa->next != aspectMap; aa=aa->next, i++)
    printPolarAngle(aa);

  printPolarAngle(aa);

  printf ("tail = %g\n", aspectMapLast->theta);
}
#endif /* DEBUG>0 */

/* ----------------------------------------------------------------------
 *
 *  printPolarAngle
 *
 * 
 */
#if DEBUG>=3
static void printPolarAngle (ASPECT_ANGLE *aa)
{
  printf ("th %5.3f  \trhoP %5.2f  \tsegN %2d  \trhoN %5.2f\n",
	  aa->theta, 
	  aa->rhoP==ASPECT_DIST_INFINITY?99:aa->rhoP, 
	  aa->segN, aa->rhoN==ASPECT_DIST_INFINITY?99:aa->rhoN);
}
#endif /* DEBUG>0 */

/* ----------------------------------------------------------------------
 *
 *  filterSegs
 *
 * 
 */

static BOOL filterSegs(ASPECT_ANGLE *aa,
		       BOOL keepReversed,
		       double minAspectSegAngle)
{
  /* reverse seg not wanted */
  if (!keepReversed && reversedSegs[aa->segN]) {
#if DEBUG>=3
    printf ("aspect filterSegs: remove inverted %d\n", aa->segN);
#endif
    aa->segN = -1;
    return FALSE;
  }
   
  /* small seg angles not wanted */
  if (minAspectSegAngle != 0 && fabs(modulo2pi(aa->next->theta - aa->theta)) 
      < minAspectSegAngle) {
#if DEBUG>=3
    printf ("aspect filterSegs: remove small angle %d\n", aa->segN);
#endif
    aa->segN = -1;
    return FALSE;
  }
  return TRUE;
}

/* ----------------------------------------------------------------------
 *
 *  fillInPolarMap
 *
 * 
 */

static BOOL fillInPolarMap (int *nbSegs, 
			    double minAspectSegAngle,
			    BOOL keepReversed,
			    int aspectOutputTabSize,
			    ASPECT_POLAR_SEG aspectPolarToFill[],
			    ASPECT_POLAR_SEG **aspectPolarToAlloc,
			    int aspectInputSegsIdToFill[],
			    int **aspectInputSegsIdToAlloc,
			    int **isid)
{
  ASPECT_ANGLE *aa;
  ASPECT_POLAR_SEG *apsl;
  int i, segN;
  int count = 0;
 
  if (!aspectMap) {
#if DEBUG>=1
    printf ("fillInPolarMap: empty map !\n");
#endif
    *nbSegs = 0;
    return TRUE;
  }

  /* count and filter */
  for (aa = aspectMap; aa->next != aspectMap; aa=aa->next) {
    if (aa->segN != -1) {

      /* filter unwanted segs */
      if (filterSegs(aa, keepReversed, minAspectSegAngle))
	/* keep others */
	count++;
    }
  }

  /* last one */
  if (aa->segN != -1)
    if (filterSegs(aa, keepReversed, minAspectSegAngle))
      count++;

#if DEBUG>=3
  printf ("fillInPolarMap: %d segs\n", count);
#endif
  *nbSegs = count;

  /* output polar map to be allocated */
  if (!aspectPolarToFill) {
    if(!(*aspectPolarToAlloc = malloc(*nbSegs*sizeof(ASPECT_POLAR_SEG)))) {
      printf ("fillInPolarMap: failed alloc aspectPolar\n");
      return FALSE;
    }
    apsl = *aspectPolarToAlloc;
  }

  /* pre-allocated map */
  else {
    if (*nbSegs > aspectOutputTabSize) {
      printf ("fillInPolarMap: WARNING aspect map size %d too small, %d were required ! Some segs will be lost\n",
	      aspectOutputTabSize, *nbSegs);
      *nbSegs = aspectOutputTabSize;
    }
    apsl = aspectPolarToFill;
  }

  /* inputSegsId to be allocated */
  if (!aspectInputSegsIdToFill) {
    if(!(*isid = malloc(*nbSegs*sizeof(int)))) {
      printf ("fillInInputSegsId: failed alloc aspectInputSegsId\n");
      return FALSE;
    }
    if (aspectInputSegsIdToAlloc)
      *aspectInputSegsIdToAlloc= *isid;
  }

  /* pre-allocated array */
  else {
    if (*nbSegs > aspectOutputTabSize) {
      printf ("fillInInputSegsId: WARNING aspect map size %d too small, %d were required ! Some segs will be lost\n",
	      aspectOutputTabSize, *nbSegs);
      *nbSegs = aspectOutputTabSize;
    }
    *isid = aspectInputSegsIdToFill;
  }

  /* FILL IN MAP */
  i = 0;
  if (aspectMap->segN != -1) aa = aspectMap;
  else aa = aspectMap->next;

#if DEBUG>=1
  printf ("\n-- polar aspect map (t1, r1) (t2, r2) :\n");
#endif
  do {
    if (aa->segN == -1) {
      aa = aa->next;
      continue;
    }

    /* record segment */
    apsl->t1 = aa->theta;
    apsl->r1 = aa->rhoN;
    segN = aa->segN;
    if (*isid)
      (*isid)[i] = segN;
    aa = aa->next;
    apsl->t2 = aa->theta;
    apsl->r2 = aa->rhoP;

#if DEBUG>=1
    printf ("%2d (map %3d) : (%g %g) (%g %g)\n", i, segN,
	    apsl->t1,    apsl->r1,	    
	    apsl->t2,    apsl->r2);
#endif
    
    i++;
    apsl++;
  } while (i<*nbSegs);

  return TRUE;
}

/* ----------------------------------------------------------------------
 *
 *  segCloser  - test if seg is closer than polar angle
 *
 * Returns TRUE if seg closer. Comuptes also rho for seg at angle
 * intersection.
 */

static int segCloser(ASPECT_SEG *seg, double theta, double rho, double *segRho)
{
  *segRho = angleDistToSeg(seg, theta);
  if (*segRho >= rho)
    return FALSE;
  return TRUE;
}

/* ----------------------------------------------------------------------
 *
 *  removeAspectAngle
 *
 */
  
static ASPECT_ANGLE *removeAspectAngle (ASPECT_ANGLE *aa)
{
  ASPECT_ANGLE *prev=aa->prev;

#if DEBUG>=3
  printf ("\n--- removed ext2 %g after %g\n", aa->theta, prev->theta);
#endif

  if (aa == aspectMap)
    aspectMap = aa->next;
  else if (aa == aspectMapLast)
    aspectMapLast = aspectMapLast->prev;

  prev->next = aa->next;
  aa->next->prev = prev;
  free(aa);

#if DEBUG>=3
  printPolarMap ();
#endif

  return prev;
}

/* ----------------------------------------------------------------------
 *
 * lookForPrevSeg    -  Look for the angle in the aspect map which value 
 *                      just preceed the given angle.
 *                      WARNING :  angles within [-pi,pi]
 */
static ASPECT_ANGLE *lookForPrevSeg(double theta)
{
  ASPECT_ANGLE *aa = aspectMap;

  /* empty list */
  if (!aspectMap) 
    return NULL;

  /* begining or end of the list */
    /* XXXXX <= or < ??? */
  if (theta <= aspectMap->theta || theta > aspectMapLast->theta)
    return aspectMapLast;

  /* next of the list */
  do {
    aa = aa->next;
    /* XXXXX <= or < ??? */
    if (theta <= aa->theta)
      return aa->prev;
  } while (aa != aspectMap);

  printf ("lookForPrevSeg ERROR: %g not in [%g,%g]\n",
	  theta, aspectMap->theta, aspectMapLast->theta);
  return NULL;
}


/* ----------------------------------------------------------------------
 *
 * initAspectMap
 *
 *
 */
static ASPECT_ANGLE *initAspectMap(SEG_POLAR *segPolar) 
{
  ASPECT_ANGLE *a1;
  ASPECT_ANGLE *a2;

  /* init new angle a1 */
  if (!(a1 = (ASPECT_ANGLE *)malloc(sizeof(ASPECT_ANGLE)))) {
    printf ("insertHead1: not enough memory\n");
    return NULL;
  }
  a1->theta = segPolar->polar.t1;
  a1->segN = segPolar->seg;
  a1->rhoP = ASPECT_DIST_INFINITY;
  a1->rhoN = segPolar->polar.r1;

  
  /* init a2 angle a2 */
  if (!(a2 = (ASPECT_ANGLE *)malloc(sizeof(ASPECT_ANGLE)))) {
    printf ("insertHead2: not enough memory\n");
    return NULL;
  }
  a2->theta = segPolar->polar.t2;
  a2->segN  = -1;
  a2->rhoP  = segPolar->polar.r2;
  a2->rhoN  = ASPECT_DIST_INFINITY;


  /* brand new map */

  if (a1->theta <= a2->theta) {
    aspectMap = a1;
    aspectMapLast = a2;
  }
  else {
    aspectMap = a2;
    aspectMapLast = a1;
  }

  aspectMap->next = aspectMapLast;
  aspectMap->prev = aspectMapLast;

  aspectMapLast->next = aspectMap;
  aspectMapLast->prev = aspectMap;

  nbAngles = 2;
  nbAspectSegs = 1;

#if DEBUG>=3
  printf ("\n--- init map\n");
  printPolarMap ();
#endif

  return a1;
}

/* ----------------------------------------------------------------------
 *
 *  insert1After  -  Inserts first seg extremity just after the indicated 
 *                   angle in the angle list.
 *                   If NULL inserts at head of list.
 *                   If after last AND but smaller angle inserts at head.
 *                   WARNING : angle within [-M_PI,M_PI]
 */
static ASPECT_ANGLE * insert1After(SEG_POLAR *segPolar, ASPECT_ANGLE *prev, 
				   double otherRho)
{
  ASPECT_ANGLE *new;


  /* alloc polar angle */
  if (!(new = (ASPECT_ANGLE *)malloc(sizeof(ASPECT_ANGLE)))) {
#if DEBUG>=3
    printf ("insert1After: alloc failed\n");
#endif
    return NULL;
  }

  /* init angle */
  new->theta = segPolar->polar.t1;
  new->rhoN  = segPolar->polar.r1;
  new->rhoP  = otherRho;
  new->segN  = segPolar->seg;  

  /* insert after prev */
  new->next  = prev->next;
  new->prev  = prev;
  prev->next = new;
  new->next->prev = new;


#if DEBUG>=4
    printf ("insert %g after %g\n", 
	    new->theta, new->prev->theta);
#endif 

  /* head or tail of list */
  if (prev == aspectMapLast) {

    /* new head */
    if (segPolar->polar.t1 < aspectMapLast->theta) {
#if DEBUG>=4
      printf ("%g is head\n", new->theta);
#endif 
      aspectMap = new;
    }

    /* new tail */
    else {
      aspectMapLast = new;
#if DEBUG>=4
      printf ("%g is tail\n", new->theta);
#endif 
    }
  }

  /* record angle */
  nbAngles++;
  nbAspectSegs++;

#if DEBUG>=3
    printf ("\n--- insert ext1 %g after %g\n", new->theta, prev->theta);
    printPolarMap ();
#endif
  return new;
}

/* ----------------------------------------------------------------------
 *
 *  insert2After   - like insert1After for the second seg extremity
 *
 *
 */
static ASPECT_ANGLE * insert2After(SEG_POLAR *segPolar, ASPECT_ANGLE *prev)
{
  ASPECT_ANGLE *new;

  /* alloc polar angle */
  if (!(new = (ASPECT_ANGLE *)malloc(sizeof(ASPECT_ANGLE)))) {
    printf ("aspect insert2After : malloc failed \n");
    return NULL;
  }

  /* fill in structure */
  new->theta = segPolar->polar.t2;
  new->rhoP  = segPolar->polar.r2;
  new->segN  = -1;
  new->rhoN = ASPECT_DIST_INFINITY;

  /* insert */
  new->next  = prev->next;
  new->prev  = prev;
  prev->next = new;
  new->next->prev = new;

#if DEBUG>=4
    printf ("insert %g after %g\n", 
	    new->theta, new->prev->theta);
#endif 

  /* head or tail of list */
  if (prev == aspectMapLast) {

    /* new head */
    if (segPolar->polar.t2 < aspectMapLast->theta) {
      aspectMap = new;
#if DEBUG>=4
      printf ("%g is head\n", new->theta);
#endif 
    }
    /* new tail */
    else {
      aspectMapLast = new;
#if DEBUG>=4
      printf ("%g is tail\n", new->theta);
#endif 
    }
  }

#if DEBUG>=3
  printf ("\n--- insert ext2 %g after %g\n", new->theta, prev->theta);
  printPolarMap ();
#endif

  nbAngles++;
  return new;
}


/* ----------------------------------------------------------------------
 *
 * cartToPolar
 *
 *
 */
static int cartToPolar(ASPECT_SEG *cartSeg, 
		       ASPECT_POLAR_SEG *polarSeg,
		       BOOL *reverted)
{
  double x, y, rho, theta;

  /* distances */
  polarSeg->r1 = hypot(cartSeg->x1, cartSeg->y1);
  polarSeg->r2 = hypot(cartSeg->x2, cartSeg->y2);
  if (polarSeg->r1 < 0.001 && polarSeg->r2 < 0.001)
    return FALSE;

  /* angles */
  polarSeg->t1 = atan2(cartSeg->y1, cartSeg->x1);
  polarSeg->t2 = atan2(cartSeg->y2, cartSeg->x2);

  /* already in trigo */
  if (isTrigo(polarSeg->t1, polarSeg->t2)) {
    *reverted = FALSE;
    return TRUE;
  }

  /* put in trigo */
  *reverted = TRUE;
#if DEBUG>=2
  printf ("(inverted)\n");
#endif
  
  /* invert polar seg extremities */ 
  theta = polarSeg->t1;
  polarSeg->t1 = polarSeg->t2;
  polarSeg->t2 = theta;
  
  rho = polarSeg->r1;
  polarSeg->r1 = polarSeg->r2;
  polarSeg->r2 = rho;
  
  /* invert cartesian seg extremities */ 
  x = cartSeg->x1;
  y = cartSeg->y1;
  cartSeg->x1 = cartSeg->x2;
  cartSeg->y1 = cartSeg->y2;
  cartSeg->x2 = x;
  cartSeg->y2 = y;
  return TRUE;
}

/* ----------------------------------------------------------------------
 *
 * aspectCartToPolar
 *
 *
 */
BOOL aspectCartToPolarFilter(ASPECT_SEG *cartSeg, 
			     ASPECT_POLAR_SEG *polarSeg,
			     double maxDistView, 
			     BOOL filterAngleView,
			     double minAngleView,
			     double maxAngleView,
			     int *reverted)
{
  /* convert to polar */
  if (!cartToPolar(cartSeg, polarSeg, reverted))
    return FALSE;

  /* filter distance */
  if (maxDistView > 0 && 
      minDistToSeg(polarSeg, cartSeg) > maxDistView) {
#if DEBUG>=2
    printf ("too far\n");
#endif
    return FALSE;
  }

#if DEBUG>=2
  printf ("before clip : t1 %5.3f r1 %5.2f t2 %5.3f r2 %5.2f\n",
	  polarSeg->t1, polarSeg->r1, polarSeg->t2, polarSeg->r2);
#endif

  /* filter angle of view */
  if (filterAngleView) {

    if (!aspectClipPolarSeg(minAngleView, 
			    maxAngleView,
			    maxDistView,
			    polarSeg,
			    cartSeg))
      return FALSE;
  }

  return TRUE;
}
 

/* ----------------------------------------------------------------------
 *
 *  aspectClipPolarSeg  - 
 *
 *
 *  returns :  0 : out
 *             1 : clip ext 1
 *             2 : clip ext 2
 *             3 : clip both
 *             4 : unchanged
 */

int aspectClipPolarSeg(double minAngleView, 
		       double maxAngleView, 
		       double maxDistView,
		       ASPECT_POLAR_SEG *polar,
		       ASPECT_SEG *cartSeg)
{    
  int  clip;
  int trigo;
  ASPECT_SEG cartSegL;
  ASPECT_POLAR_SEG polarL;

  /* XXX Is it OK ? */
  if (minAngleView > 0 && minAngleView > maxAngleView) {
    minAngleView -= 2*M_PI;
    if (polar->t1 > maxAngleView)
      polar->t1 -= 2*M_PI;
    if (polar->t2 > maxAngleView)
      polar->t2 -= 2*M_PI;
  }

  if (maxAngleView < 0 && minAngleView > maxAngleView) {
    maxAngleView += 2*M_PI;
    if (polar->t1 > maxAngleView)
      polar->t1 += 2*M_PI;
    if (polar->t2 > maxAngleView)
      polar->t2 += 2*M_PI;
  }

  /* test trigo */
  if (isTrigo(polar->t1, polar->t2)) {
    polarL = *polar;
    trigo = TRUE;
  }
  else {
    polarL.t1 = polar->t1;
    polarL.r1 = polar->r1;
    polarL.t2 = polar->t2;
    polarL.r2 = polar->r2;
    trigo = FALSE;
  }

  /* computes cart if not given */
  if (cartSeg && trigo) {
    cartSegL = *cartSeg;
  }
  else {
    cartSegL.x1 = polarL.r1 * cos(polarL.t1);
    cartSegL.y1 = polarL.r1 * sin(polarL.t1);
    cartSegL.x2 = polarL.r2 * cos(polarL.t2);
    cartSegL.y2 = polarL.r2 * sin(polarL.t2);
  }

  /* filter distance */
  if (maxDistView > 0 && minDistToSeg(&polarL, &cartSegL) > maxDistView) {
#if DEBUG>=2
    printf ("too far\n");
#endif
    clip = 0;
    return 0;
  }

  /* extremity 1 out of [minAngleView, maxAngleView] */
  if (!thetaWithinRangeLeftRight(polarL.t1, minAngleView, maxAngleView)) {

    clip = 1;

    /* extremity 2 also out of [minAngleView, maxAngleView] */
    if (!thetaWithinRangeLeftRight(polarL.t2, minAngleView, maxAngleView)) {
      
      /* angle view (ie, minAngleView or maxAngleView) not within segment sector */
      if (!thetaWithinRangeLeftRight(minAngleView, polarL.t1, polarL.t2)) {
	clip = 0;
	return 0;
      }
      
      /* cut what is out */
      clip = 3;
      polarL.t2 = maxAngleView;
      polarL.r2 = angleDistToSeg(&cartSegL, maxAngleView);
      cartSegL.x2 = polarL.r2 * cos(polarL.t2);
      cartSegL.y2 = polarL.r2 * sin(polarL.t2);
    }
  
    /* clip ext1 */
    polarL.t1 = minAngleView;
    polarL.r1 = angleDistToSeg(&cartSegL, minAngleView);      
    cartSegL.x1 = polarL.r1 * cos(polarL.t1);
    cartSegL.y1 = polarL.r1 * sin(polarL.t1);
  } /* ext 1 out */
  

  /* ext 1 not clipped */
  else {

    /* only extremity 2 out of  [minAngleView, maxAngleView] */
    if (!thetaWithinRangeLeftRight(polarL.t2, minAngleView, maxAngleView)) {
      
      /* cut ext 2 */
      polarL.t2 = maxAngleView;
      polarL.r2 = angleDistToSeg(&cartSegL, maxAngleView);
      cartSegL.x2 = polarL.r2 * cos(polarL.t2);
      cartSegL.y2 = polarL.r2 * sin(polarL.t2);
      clip = 2;
    }

    /* nothing to clip */
    else {
      clip = 4;
    }
  }
    
  /* record result */
  if (trigo) {
    *polar = polarL;
    if (cartSeg)
      *cartSeg = cartSegL;
  }
  else {
    polar->t1 = polarL.t2;
    polar->r1 = polarL.r2;
    polar->t2 = polarL.t1;
    polar->r2 = polarL.r1;
    if (cartSeg) {
      cartSeg->x1 = cartSegL.x2;
      cartSeg->y1 = cartSegL.y2;
      cartSeg->x2 = cartSegL.x1;
      cartSeg->y2 = cartSegL.y1;
    }
  }

#if DEBUG>=2
  switch(clip) {
  case 0: printf ("out of view\n"); break;
  case 1: printf ("clip ext1\n"); break;
  case 2: printf ("clip ext2\n"); break;
  case 3: printf ("clip both\n"); break;
  case 4: printf ("no clip\n"); break;
  default: printf ("CLIP ERROR\n"); break;
  }
  printf ("after clip: t1 %5.3f r1 %5.2f  t2 %5.3f r2 %5.2f\n",
	  polar->t1, polar->r1, polar->t2, polar->r2);
  printf ("after clip: x1 %5.2f y1 %5.2f  x2 %5.2f y2 %5.2f\n",
	  cartSeg->x1, cartSeg->y1, cartSeg->x2, cartSeg->y2);
#endif
  
  return clip;
} 

/* ----------------------------------------------------------------------
 *
 *  aspectGlobalToLocal  - 
 *
 *
 */

static BOOL aspectGlobalToLocal(ASPECT_CONFIG *sensorViewPoint, 
				int nbSegs, const ASPECT_SEG cartMap[])
{
  int seg;
  double dx, dy;
  double cosView, sinView;

  /* alloc local maps */
  if (!allocLocalMaps(nbSegs))
    return FALSE;

  /* map already in local frame : juste a copy. */
  if (!sensorViewPoint || 
      (sensorViewPoint->x == 0 && sensorViewPoint->y == 0 && 
       sensorViewPoint->theta == 0)) {
    memcpy(localCartMap, cartMap, nbSegs*sizeof(ASPECT_SEG));
    return TRUE;
  }

  /* map to be convert in local frame */
  cosView = cos(sensorViewPoint->theta);
  sinView = sin(sensorViewPoint->theta);

  /* for each seg */
  for (seg=0; seg<nbSegs; seg++) {
    
    /* convert map from global to view point frame */
    dx = cartMap[seg].x1 - sensorViewPoint->x;
    dy = cartMap[seg].y1 - sensorViewPoint->y;
    localCartMap[seg].x1 = dx * cosView + dy * sinView;
    localCartMap[seg].y1 = dy * cosView - dx * sinView;
    dx = cartMap[seg].x2 - sensorViewPoint->x;
    dy = cartMap[seg].y2 - sensorViewPoint->y;
    localCartMap[seg].x2 = dx * cosView + dy * sinView;
    localCartMap[seg].y2 = dy * cosView - dx * sinView;
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
 *
 *  thetaWithinRangeLeftRight  - tests if theta belongs to [t1,t2]
 *
 *  XXXX is it OK ?
 */
static BOOL thetaWithinRangeLeftRight(double theta, 
				      double theta1, 
				      double theta2)
{
  /* theta equal or greater than theta1 ? */
  if (theta >= theta1)

    /* AND theta equal or smaller than theta2 ? */
    if (theta <=  theta2)
      return TRUE;
#if 0
  /* theta equal or greater than theta1 ? */
  if (modulo2pi(theta - theta1) >= 0)

    /* AND theta2 equal or greater than theta ? */
    if (modulo2pi(theta2 - theta) >= 0)
      return TRUE;
#endif
  return FALSE;
}

/* ----------------------------------------------------------------------
 *
 *  isTrigo - test weather theta2 is greater than theta1,
 *            ie, 0 <= theta2-theta1 < M_PI
 *
 *
 */
static BOOL isTrigo(double theta1, double theta2)
{
  /* trigo */
  if (modulo2pi(theta2 - theta1) >= 0)
    return TRUE;
  
  /* not trigo */
  return FALSE;
}

/* ----------------------------------------------------------------------
 *
 *  modulo2pi  -  get back an angle within [-pi,pi[
 *
 *  WARNING : we suppose that intially theta is within [-2pi,2pi[
 *
 *  (here it is OK because initially all angles are within [-pi,pi[
 *   so the modulo is required only after a sum or a diff of 2 angles).
 *
 */
static double modulo2pi(double theta) 
{
  if (theta < -M_PI) return (theta + 2*M_PI);
  if (theta >= M_PI) return (theta - 2*M_PI);
  return theta;
}

/* ----------------------------------------------------------------------
 *
 *  angleDistToSeg    -  computes distance to a given seg along a given angle
 * 
 * We suppose that this angle really intersect the segment. No check !
 *
 */
static double angleDistToSeg (ASPECT_SEG *seg, double theta) 
{ 
  double x01, y01, x12, y12, dist;
 
  x01 = - seg->x1; 
  y01 = - seg->y1; 
  x12 = seg->x1 - seg->x2; 
  y12 = seg->y1 - seg->y2; 

  dist = fabs((y01 * x12 - x01 * y12)  
	      / (x12 * sin(theta) - y12 * cos(theta))); 

  return (dist); 
} 

/* ----------------------------------------------------------------------
 *
 *  minDistToSeg    -  computes minimum distance to a given seg
 * 
 *
 */
static double minDistToSeg (ASPECT_POLAR_SEG *polar, ASPECT_SEG *seg)
{ 
  double vx, vy;
  double d1, d2, dist;

  /* vector of the segment */
  vx = seg->x2 - seg->x1;
  vy = seg->y2 - seg->y1;
  

  /* test if the perpendicular projection of the point is on the segment */

  /* distance  between P1 and the projection of P on S (scalar product P1P2.P1P) */
  d1 = ( - seg->x1)*vx + ( - seg->y1)*vy;
  
  /* distance  between P2 and the projection of P on S (scalar product P1P2.P2P)*/
  d2 = ( - seg->x2)*vx + ( - seg->y2)*vy;
  
  /* proj(P) on S is not in [P1,P2] */
  if ( d1 <= 0 || d2 >= 0) {

    /* the minimum distance is one of the 2 extremities */
    return MIN(polar->r1, polar->r2);
  }

  /* proj on the seg : the mini dist is the vecorial product P1P2^P1P */
  dist = (vx*seg->y1 - vy*seg->x1) / sqrt(vx*vx + vy*vy);
  
#if DEBUG>=2
  printf ("min seg dist = %g \n", dist);
#endif
  return fabs(dist); 
} 


/* ----------------------------------------------------------------------
 *
 *  aspectDistToSeg    -  computes minimum distance to a given seg
 * 
 *
 */
double aspectDistToSeg (ASPECT_SEG *seg) 
{ 
  double vx, vy;
  double d1, d2, dist;

  /* vector of the segment */
  vx = seg->x2 - seg->x1;
  vy = seg->y2 - seg->y1;
  

  /* test if the perpendicular projection of the point is on the segment */

  /* distance  between P1 and the projection of P on S (scalar product P1P2.P1P) */
  d1 = ( - seg->x1)*vx + ( - seg->y1)*vy;
  
  /* distance  between P2 and the projection of P on S (scalar product P1P2.P2P)*/
  d2 = ( - seg->x2)*vx + ( - seg->y2)*vy;
  
  /* proj(P) on S is not in [P1,P2] */
  if ( d1 <= 0 || d2 >= 0) {

    /* the minimum distance is one of the 2 extremities */
    return MIN(hypot(seg->x1, seg->y1), hypot(seg->x2, seg->y2));
  }

  /* proj on the seg : the mini dist is the vecorial product P1P2^P1P */
  dist = (vx*seg->y1 - vy*seg->x1) / sqrt(vx*vx + vy*vy);
  
#if DEBUG>=2
  printf ("min seg dist = %g \n", dist);
#endif
  return fabs(dist); 
} 

double aspectPtDistToSeg (ASPECT_SEG *seg) 
{ 
  double vx, vy;
  double d1, d2, dist;

  /* vector of the segment */
  vx = seg->x2 - seg->x1;
  vy = seg->y2 - seg->y1;
  

  /* test if the perpendicular projection of the point is on the segment */

  /* distance  between P1 and the projection of P on S (scalar product P1P2.P1P) */
  d1 = ( - seg->x1)*vx + ( - seg->y1)*vy;
  
  /* distance  between P2 and the projection of P on S (scalar product P1P2.P2P)*/
  d2 = ( - seg->x2)*vx + ( - seg->y2)*vy;
  
  /* proj(P) on S is not in [P1,P2] */
  if ( d1 <= 0 || d2 >= 0) {

    /* the minimum distance is one of the 2 extremities */
    return MIN(hypot(seg->x1, seg->y1), hypot(seg->x2, seg->y2));
  }

  /* proj on the seg : the mini dist is the vecorial product P1P2^P1P */
  dist = (vx*seg->y1 - vy*seg->x1) / sqrt(vx*vx + vy*vy);
  
#if DEBUG>=2
  printf ("min seg dist = %g \n", dist);
#endif
  return fabs(dist); 
} 

/* ----------------------------------------------------------------------
 *
 *  allocLocalMaps
 *
 *
 *
 */
static BOOL allocLocalMaps(int nbSegs)
{
  /* array large enough */
  if (nbMaxSegs >= nbSegs)
    return TRUE;

  /* free all arrays */
  freeLocalMaps();

  /* compute new map size */
  nbMaxSegs = ((nbSegs+100)/100)*100;

#if DEBUG>=0
  printf ("allocLocalMaps : alloc local map for %d elts \n", nbMaxSegs);
#endif  

  /* alloc local map */
  if (!(localCartMap = (ASPECT_SEG *)malloc(nbMaxSegs*sizeof(ASPECT_SEG)))) {
    printf ("allocLocalMaps: alloc local cart map failed\n");
    return FALSE;
  }

  if (!(reversedSegs = (int *)malloc(nbMaxSegs*sizeof(int)))) {
    printf ("allocLocalMaps: alloc invert failed\n");
    return FALSE;
  }

  /* init maps OK */
  return TRUE;
}

/* ----------------------------------------------------------------------
 *
 *  freeLocalMaps
 *
 *
 */
static void freeLocalMaps()
{
  /* free local cart map */
  if (localCartMap) {
    free(localCartMap);
    localCartMap = NULL;
  }

  if (reversedSegs) {
    free(reversedSegs);
    reversedSegs = NULL;
  }
}

/* ----------------------------------------------------------------------
 *
 *  freeAspectMap
 *
 *
 */
static void freeAspectMap()
{
  ASPECT_ANGLE *next, *aa = aspectMap;

  while(aa != aspectMapLast) {
    next = aa->next;
    free(aa);
    aa = next;
  }
  free(aspectMapLast);
  aspectMapLast = NULL;
  aspectMap = NULL;
}

/* ----------------------------------------------------------------------
 *
 *  aspectReadSegkitMap
 *
 *
 */

ASPECT_SEG * aspectReadSegkitMap(char *filename, int *nbsegs)
{
  FILE *pfile;
  int i;
  int nbSegs;
  ASPECT_SEG *segs=NULL;
  ASPECT_SEG *seg;
  double x, y, cs, sn, cst, snt;
  double enc[4];
  double vx, vy, d,r, o;
  double vr, vro, vo;
  char typ1[8], typ2[8];
  char coosys;

  /* Lecture fichier */
  if ((pfile = fopen(filename,"r")) == NULL) {
    printf ("seglocReadSegkitMap: cannot open %s\n", filename);
    return NULL;
  }
  
  /* Nombre de segments ? */
  if (fscanf(pfile, "%d", &nbSegs) != 1) {
    printf ("seglocReadSegkitMap: cannot read nbsegs\n");
    fclose(pfile);
    return NULL;
  }
  *nbsegs = nbSegs;

  /* Entete */
  if (fscanf(pfile, "%lf %lf %lf %lf %lf %lf %c", 
	     &x, &y, 
	     &cs, &sn,
	     &cst, &snt, 
	     &coosys) != 7) {
    printf ("seglocReadSegkitMap: cannot read header\n");
    fclose(pfile);
    return NULL;
  }
  if (fscanf(pfile, "%lf %lf %lf %lf", 
	     &enc[0], &enc[1],
	     &enc[2], &enc[3]) != 4) {
    printf ("seglocReadSegkitMap: cannot read enc\n");
    fclose(pfile);
    return NULL;
  }
	     
  /* Les segments */
  if ((segs = (ASPECT_SEG *) malloc (sizeof(ASPECT_SEG) * nbSegs)) == NULL) {
    printf ("seglocReadSegkitMap: cannot alloc segs\n");
    fclose(pfile);
    return NULL;
  }

  seg=segs;
  for(i=0; i<nbSegs; i++,seg++) {
    
    /* Premier point */
    if (fscanf(pfile, "%lf %lf", 
	       &seg->x1, &seg->y1) != 2) {
      printf ("seglocReadSegkitMap: cannot read first point (seg %d)\n", i);
      fclose(pfile);
      free(segs);
      return NULL;
    }
    /* second point */
    if (fscanf(pfile, "%lf %lf", 
	       &seg->x2, &seg->y2) != 2) {
      printf ("seglocReadSegkitMap: cannot read second point (seg %d)\n", i);
      fclose(pfile);
      free(segs);
      return NULL;
    }
    /* vecteur */
    if (fscanf(pfile, "%lf %lf %lf", 
	       &vx, &vy, &d) != 3) {
      printf ("seglocReadSegkitMap: cannot read vx ... (seg %d)\n", i);
      fclose(pfile);
      free(segs);
      return NULL;
    }
    /* o,r */
    if (fscanf(pfile, "%lf %lf", 
	       &r, &o) != 2) {
      printf ("seglocReadSegkitMap: cannot read o, r (seg %d)\n", i);
      fclose(pfile);
      free(segs);
      return NULL;
    }
    /* variances */
    if (fscanf(pfile, "%lf %lf %lf", 
	       &vr, &vro, &vo) != 3) {
      printf ("seglocReadSegkitMap: cannot read variances ... (seg %d)\n", i);
      fclose(pfile);
      free(segs);
      return NULL;
    }
    /* type extremite */
    if (fscanf(pfile, "%s %s", 
	       typ1, typ2) != 2) {
      printf ("seglocReadSegkitMap: cannot read extremity types (seg %d)\n", 
	      i);
      fclose(pfile);
      free(segs);
      return NULL;
    }
  }
    
  fclose(pfile);
  return segs;
}
