/*************************************************
 *                                               *
 * Nomadic Host Software Development Environment *
 *                                               *
 * The Nomadic Geometry Engine                   *
 *   contains all the important geometric        *
 *   reasoning routines                          *
 *                                               *
 * First Created: 06/05/92                       *
 * Last Modified: 10/04/92                       *
 *                                               *
 * Author:        David J. Zhu                   *
 *                                               *
 * Copyright 1992, Nomadic Technologies, Inc.    *
 *                                               *
 *************************************************/



/*
 * Revision 1.2  1993/07/21  22:14:11  kamason
 * Added rcs LOG comment
 */

/* Mod Log: 
 *  1. my own memory management for pline_list and vertex_list: 8/28;
 *  2. add eliminate_slanted_pline: 9/25;
 *  3. fixed closer_pline: 9/28;
 *  4. fixed poly_intersect_poly: 9/28;
 *  5. added multiple robot sensing: 2/27/93
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef VXWORKS
#define M_PI 3.141592653589793
#endif

#ifdef min 
#undef min
#endif
#ifdef max
#undef max
#endif




/*******
  includes for philippe's structures
*******/

#include <lasdef.h>
#include <las_struct_p.h>
#include <dbliste.h>
/******
  prototypes for Aspect.c
*******/

#include "aspect.h"
#include "aspect_p.h"
#include "aspect_private_p.h"
#include "aspect_private_common_p.h"

static int filter_sgl_seg(Sgl *s, double distmax, 
			  double anglemax, 
			  int testfreespace);

static struct pline_memory_type pline_memory;

static long      Max_range = 10000;
static AList *ListAngles = NULL;



/* other prototyping for functions that David added to this file */

struct polar_line_segment_type *Find_visible_segments
       (igla *locmap, igla *aspect, double distmax, double a1, double a2);

/*************
computes in the image aspect the set of
visible segments of locmap.
Locmap is expressed in the local 
coordinate system of the point of
view
*************/

/**
 ** Local_visibility : compute the aspect map 
 **                    (ie, the segments selected in the map for localization)
 **                    (Original method)
 **/
void Local_visibility(locmap,aspect,distmax)
     igla *locmap;
     igla *aspect;
     double distmax;
 
{
  struct polar_line_segment_type *pline;
 

  /* Conversion to pline */
  pline = Sgl_to_pline(locmap);
 
  /* Compute visible segments (visibility map) from the robot 
     estimated position */
  pline = visible_polar_line_list(pline); 
 
  /* Convert back to Sgl + filter segments to far away */
  Pline_to_Sgl(pline,locmap,aspect,distmax);
 
  free_pline_list(pline);
}
  


/**
 ** Local_visibility_filter : 
 **
 **  Same as above but more efficient :  previously to the computation of 
 **  the visibility map filters segments that are :
 **         - too far away (if distmax != 0)
 **         - "behind" the robot (if anglemax < PI)
 **/

/* XXX Would be nice to take into account position uncertainty
   XXX (by translating it to an angular uncertainty according to 
   XXX  the distances robot-seg1-seg2) */

void Local_visibility_filter(igla *locmap, 
			     igla *aspect, 
			     double distmax,
			     double anglemax)

{
  struct polar_line_segment_type *pline;

  /* Conversion to pline + filter segments */
  pline = Sgl_to_pline_filter(locmap,distmax,anglemax,0);


  /* Compute visible segments (visibility map) from the robot 
     estimated position */
  pline = visible_polar_line_list(pline); 

  /* Convert back to Sgl (without the distmax filter) */
  Pline_to_Sgl(pline,locmap,aspect,0.0);

  free_pline_list(pline);
}
  
/**
 ** Local_visibility_simple : 
 **
 **  Same as above but DO NOT compute the real visible segments.
 **  Only the filters are applied
 **         - too far away (if distmax != 0)
 **         - "behind" the robot (if anglemax < PI)
 **         - free side not visible (if testfreespace == 1)
 **/
void Local_visibility_simple(igla *locmap, 
			     igla *aspect, 
			     double distmax,
			     double anglemax,
			     int testfreespace)
{
  Sgl *s,*sl=NULL;
  PDBL *dl;
  int i=0;

  /* loop over locmap's segments */
  for(dl = locmap->lsgl; dl != NULL; dl = dls(dl), i++){

    /* getting the Sgl segment */
    s = (Sgl *)dlel(dl);

    /* Test if keep seg */
    if (!filter_sgl_seg(s,distmax,anglemax,testfreespace))
      continue;

    /* creating the new one  */
    sl = Sgl_cree();

    /* copying */
    *sl = *s;

    /* keep index of original sgl segment */
    sl->ifirst = i;

    /* adding to the aspect */
    aspect->lsgl = dlhcons(aspect->lsgl,sl);
    aspect->nsg ++;
  }
}


/**
 ** filter_sgl_seg :   Test acceptable segments
 **  
 **  (the segments have to be in the robot frame)
 **  
 **  Returns 0 for segments that are :
 **         - at more than distmax from the origine (if distmax != 0)
 **         - out of the angular view ]-anglemax,anglemax[ 
 **          (if 0 < anglemax < PI)
 **         - free side not visible (if testfreespace == 1)
 **           WARNING: not to be done if a real visibility map is computed
 **           afterwards
 **         
 **  Otherwise, returns 1
 **/
static int filter_sgl_seg(Sgl *s, double distmax, 
			  double anglemax, 
			  int testfreespace)
{
#if UNUSED
  double t1=1000, t2=1000, dt;
#else
  double t1, t2;
#endif

  /* removed if to far away */
  if (distmax != 0.0)
    if (!test_near_seg(s,distmax))
      return 0;
  
  /* Removed if back the robot */
  if (anglemax > 0.0 && anglemax < PI) {
    t1 = atan2(s->first.y,s->first.x);
    if (fabs(t1) > anglemax) {
      t2 = atan2(s->last.y,s->last.x);
      if (fabs(t2) > anglemax)
	return 0;
    }
  }
  
  /* Removed if free space not in the correct side
     (WARNING: not to be done if a visibility map is built afterwards) */

  if (testfreespace) {

    /* Segment have been already reversed in tr_abs_to_loc */
    if (s->flg)
      return 0;

#if UNUSED
    /* Useless test as segments have been reversed in tr_abs_to_loc ! 
       and flag s->flg has been set to 1. But perahps can be usefull 
       in other situations...). */
      if (t1 == 1000)
	t1 = atan2(s->first.y,s->first.x);
      if (t2 == 1000)
	t2 = atan2(s->last.y,s->last.x);
      dt = t2-t1;
      if(dt>PI)
	dt-=PI;
      else if(dt<-PI)
	dt+=PI;
      if (dt<0.0) {
	return 0;
      }
#endif
  }
  
  /* OK */
  return 1;
}

/**********
 visible segments within angular range (angle1, angle2)
 if angle1 == angle2, then the complete 2PI range.
 Added by David for local localization strategy
**********/

struct polar_line_segment_type *Find_visible_segments
       (igla *locmap, igla *aspect, double distmax, double a1, double a2)
{
  struct polar_line_segment_type *pline1, *pline2, *pline3;

  pline1 = Sgl_to_pline(locmap);

  pline1 = visible_polar_line_list(pline1);
  
  Pline_to_Sgl(pline1,locmap,aspect,distmax);

  pline2 = Sgl_to_pline(aspect);
  free_pline_list(pline1);

  if (a1 == a2) {
    return(pline2);
  }
  else {
    pline3 = clip_pline_list_angle_range(pline2,a1,a2); 
    free_pline_list (pline2);
    return(pline3);
  }    
}





/*********
  interface a map of structures Sgl (see lasdef.h)
  into polar lines.  
  locmap is already expressed in the local 
  coordinate system corresponding
  to the point of view.
  Conventions: 
  David: free space on the left side when
         walking from the begining to the end of 
	 a segment (at least for the plines...)
  Philippe: the  same.
  XXX Sara: but Phil, you re-reverse the segment (and set s->flg = 1)
  XXX      in func tr_abs_to_loc (map_util.c). Is it a pb ?
*********/


static struct polar_line_segment_type *Sgl_to_pline(igla *locmap)
{

  Sgl *s;
  long i;
  struct polar_line_segment_type *lp = NULL,*lr  = NULL,*ls;
  PDBL *dl;

  /***
    Note: 
    locmap->usr[i]  = pointer on the ith segment in locmap->lsgl.
  ***/

  i = 0;
  
  
  /* loop over locmap's segments */

  for(dl = locmap->lsgl; dl != NULL; dl = dls(dl)){

    /* creating and adding the new pline */
    ls = get_pline();
    ls->next_pline = 0;

    if(lp == NULL)
      lr = ls;
    else
      lp->next_pline = ls;
    lp = ls;

    /* getting the Sgl segment */
    s = (Sgl *)dlel(dl);

    /* keep index of original sgl segment */
    ls->id = i;
    i++;

    /* reading and copying the Sgl */
    
    ls->n = atan2(s->vy,s->vx);
    ls->r2 = s->last.r;
    ls->t2 = atan2(s->last.y,s->last.x);
    ls->r1 = s->first.r;
    ls->t1 = atan2(s->first.y,s->first.x);
    ls->x01 = -s->first.x;
    ls->y01 = -s->first.y;
    ls->x12 = s->first.x - s->last.x;
    ls->y12 = s->first.y - s->last.y;

    
  }
  
  return(lr);
}

/*********
	  Same as above but filter segment according to 
	  distance ... :

  removed segments that are :

  o further than distmax (>0.0)
  o "behind" de robot : out of the angular view [anglemax,-anglemax]
     (0<=anglemax<=PI)
  o miss oriented relatively to free space (only if testfreespace
     flag is set to 1). WARNING: This last test should not be used
     if a visibilty segment map is computed later.
*********/


static struct polar_line_segment_type 
*Sgl_to_pline_filter(igla *locmap, double distmax,
		     double anglemax, int testfreespace)
{

  Sgl *s;
  long i;
  struct polar_line_segment_type *lp = NULL,*lr  = NULL,*ls;
  PDBL *dl;

  /***
    Note: 
    locmap->usr[i]  = pointer on the ith segment in locmap->lsgl.
  ***/

  i = -1;
  
  
  /* loop over locmap's segments */

  for(dl = locmap->lsgl; dl != NULL; dl = dls(dl)){

    /* getting the Sgl segment */
    s = (Sgl *)dlel(dl);
    i++;

    /* Filter segments */
    if (!filter_sgl_seg(s, distmax, anglemax, testfreespace))
      continue;

    /* creating and adding the new pline */
    ls = get_pline();
    ls->next_pline = 0;

    if(lp == NULL)
      lr = ls;
    else
      lp->next_pline = ls;
    lp = ls;


    /* keep index of original sgl segment */
    ls->id = i;

    /* reading and copying the Sgl */
    
    ls->n = atan2(s->vy,s->vx);
    ls->r2 = s->last.r;
    ls->t2 = atan2(s->last.y,s->last.x);
    ls->r1 = s->first.r;
    ls->t1 = atan2(s->first.y,s->first.x);
    ls->x01 = -s->first.x;
    ls->y01 = -s->first.y;
    ls->x12 = s->first.x - s->last.x;
    ls->y12 = s->first.y - s->last.y;

    
  }
  
  return(lr);
}

static int test_near_seg(Sgl *seg, double dist)
{
  double t1,t2,d;
 
  if (seg->first.r < dist)
    return 1;
  if (seg->last.r < dist)
    return 1;

  t1 = seg->first.x * seg->vy  -  seg->first.y * seg->vx;
  t2 = seg->last.x  * seg->vy  -  seg->last.y  * seg->vx;

  /* the origin does not projects on the segment */
  if(t1*t2 >= 0) 
    return 0;
    
  d = fabs(seg->first.x*seg->vx + seg->first.y*seg->vy);
  if (d < dist)
      return 1;

  return 0;
}

/********
  back to Sgl  structures;
  We eliminate segments further than distmax
  from the pt of view.
********/
static void Pline_to_Sgl(struct polar_line_segment_type *pline,
			 igla *locmap,
			 igla *aspect,
			 double distmax)

{
  struct polar_line_segment_type *pl;
  Sgl *s,*sl=NULL;
  Sgl **tab;
  long integ = 1;

  pl = pline;

  tab = (Sgl **)locmap->usr;

  while(pl){
    /* 
       filtering small angle not allowing
       a segment to be perceived anyway
    */
    if(Small_angle_filter(pl)){

      /* getting back to the original segment */
      s = tab[pl->id];

      /* if s not reversed */
      if(s->flg == 0){
	
	/* creating the new one  */
	if (integ)
	  sl = Sgl_cree();
	/* copying the visible part */
	*sl = *s;
	sl->last.x = pl->r2*cos(pl->t2);
	sl->last.y = pl->r2*sin(pl->t2);
	sl->first.x = pl->r1*cos(pl->t1);
	sl->first.y = pl->r1*sin(pl->t1);

	/* Keep original segment index in the map */
	sl->ifirst = pl->id;
	
	/* filtering segment too far to be seen */
	if (distmax == 0.0 ||
	    dist_to_origin(sl->first.x,sl->first.y,sl->last.x,sl->last.y, 
			   s->vx,s->vy) < distmax)
	  {
	    /* adding to the aspect */
	    aspect->lsgl = dlhcons(aspect->lsgl,sl);
	    aspect->nsg ++;
	    integ = 1;
	  }
	else
	  integ = 0;
      }
    }
    pl = pl->next_pline;
  }
  DeleteListAngle();
  if(integ == 0)
    Sgl_free(sl);
}
    
static long create_more_pline_memory ()
{
    long i;

    if (pline_memory.length+100 > MAX_PLINE_MEM) {
	printf("out of pline memory\n");
	exit(1);
    }
    else {
	pline_memory.begin_ptr = pline_memory.length;
	pline_memory.end_ptr = pline_memory.length+99; 
    }
    for (i=pline_memory.length; i<pline_memory.length+100; i++) {
	pline_memory.pline_list[i] =
	  (struct polar_line_segment_type *)
	    malloc(sizeof (struct polar_line_segment_type));
	if (!pline_memory.pline_list[i]) {
	    printf("malloc error\n");
	    return(0);
	}
	else {
	  pline_memory.pline_list[i]->next_pline = 0;
	}
    }
    pline_memory.length = pline_memory.length + 100;
    return(1);
}

static struct polar_line_segment_type *get_pline (void)
{
    if (pline_memory.begin_ptr == pline_memory.end_ptr) {
	create_more_pline_memory();
    }
    pline_memory.begin_ptr = pline_memory.begin_ptr % pline_memory.length;
    pline_memory.begin_ptr = pline_memory.begin_ptr + 1;
    return(pline_memory.pline_list[pline_memory.begin_ptr - 1]);
}

static long free_pline (pline_ptr)
struct polar_line_segment_type *pline_ptr;
{
    if (pline_ptr) {
	pline_memory.end_ptr = pline_memory.end_ptr + 1;
	pline_memory.end_ptr = pline_memory.end_ptr % pline_memory.length;
	pline_ptr->next_pline = 0;
	pline_memory.pline_list[pline_memory.end_ptr] = pline_ptr;
	return(1);
    }
    else {
	return(0);
    }
}

static void free_pline_list (pline_list)
    struct polar_line_segment_type *pline_list;
{
    struct polar_line_segment_type *pline_ptr, *old_pline_ptr;

    pline_ptr = pline_list;
    while (pline_ptr) {
	old_pline_ptr = pline_ptr;
	pline_ptr = pline_ptr->next_pline;
	free_pline (old_pline_ptr); 
    } 
}

static long theta_within_range (theta, theta1, theta2)
    double theta, theta1, theta2;
{
  double t1,t2;
  int resWhile;
  int resNorm;

  if (theta1 > theta2) {
    double tmp;
    printf ("============ theta1 %f > theta2 %f\n", theta1, theta2);
    tmp = theta1;
    theta1 = theta2;
    theta2 = tmp;
  }

#if 1
  while (theta < theta1) {
    theta = theta + TWO_PI;
  }
  while (theta > theta2) {
    theta = theta - TWO_PI;
  }
  if ((theta > theta1) && (theta < theta2)) 
    resWhile = 1;
  else
    resWhile = 0;
#else  /* Avant c'etait comme ca mais ca ne marche pas car il semble
	  que parfois [theta1 -> theta2] n'est pas dans le sens trigo.
	  (pas normal...).
	  Dans ce cas il ne faut surtout pas prendre le complementaire. */
  t1 = normalize_angle(theta2 - theta1 ,0.);
  t2 = normalize_angle(theta - theta1 ,0.);
  if( t1 > t2 && t2 > 0.)
    resNorm = 1;
  else 
    resNorm = 0;
#endif
  return resWhile;
}

static long theta_within_range_left (theta, theta1, theta2)
    double theta, theta1, theta2;
{
    double del = 0.00001;
    double t1,t2;
    int printflag=0;

#if 0
  if (theta1 > theta2) {
    double tmp;
    printf ("<<<<<<<<< theta1 %f > theta2 %f  ", theta1, theta2);
    printflag = 1;
  }
  else 
    printflag = 0;
#endif

#if 1
    while (theta < (theta1-del)) {
      theta = theta + TWO_PI;
    }
    while (theta > (theta2+del)) {
      theta = theta - TWO_PI;
    }
    if ((theta >= (theta1-del)) && (theta < theta2)) {
      if (printflag)
	printf ("  %f in\n", theta);
      return (1);
    }
    else {
      if (printflag)
	printf ("  %f out\n", theta);
      return (0);
    }
#else
    t1 = normalize_angle(theta2 + 2*del - theta1  ,0.);
    t2 = normalize_angle(theta - theta1 + del,0.);
    if( t1 > t2)
      return (1);
    else 
      return (0);
#endif
}

static long theta_within_range_right (theta, theta1, theta2)
    double theta, theta1, theta2;
{
    double del = 0.00001,t1,t2;

  if (theta1 > theta2) {
    double tmp;
    printf (">>>>>>>>>>> theta1 %f > theta2 %f\n", theta1, theta2);
    tmp = theta1;
    theta1 = theta2;
    theta2 = tmp;
  }

#if 1
    while (theta < (theta1-del)) {
      theta = theta + TWO_PI;
    }
    while (theta > (theta2+del)) {
      theta = theta - TWO_PI;
    }
    if ((theta > theta1) && (theta <= (theta2+del)))
      return (1);
    else 
      return (0);
#else
    t1 = normalize_angle(theta2 + 2*del - theta1  ,0.);
    t2 = normalize_angle(theta - theta1 + del,0.);
    if( t1 >= t2 && t2 > 0)
      return (1);
    else 
      return (0);
#endif
}

static long theta_within_range_both (theta, theta1, theta2)
    double theta, theta1, theta2;
{
    double del = 0.00001;
    double t1,t2;
  
    if (theta1 > theta2) {
      double tmp;
      printf (">>>>>>>>>>> theta1 %f > theta2 %f\n", theta1, theta2);
      tmp = theta1;
      theta1 = theta2;
      theta2 = tmp;
    }
  
#if 1
    while (theta < (theta1-del)) {
      theta = theta + TWO_PI;
    }
    while (theta > (theta2+del)) {
      theta = theta - TWO_PI;
    }
    if ((theta >= (theta1-del)) && (theta <= (theta2+del)))
      return (1);
    else 
      return (0);
#else
    t1 = normalize_angle(theta2 - theta1 + 2*del,0.);
    t2 = normalize_angle(theta - theta1 + del,0.);
    if( t1 >= t2 )
      return (1);
    else 
      return (0);
#endif
}

static long theta_overlap (th1, th2, thp1, thp2)
    /* th1 and th2 are modified to be within thp1 and thp2 if they overlap */
    double *th1, *th2, thp1, thp2;
{
    long overlap = 0;

    if (theta_within_range_left(thp1, *th1, *th2)) {
	if (theta_within_range_right(thp2, *th1, *th2)) {
	    *th1 = thp1;
	    *th2 = thp2;
	    overlap = 1;
	}
	else {
	    if (*th2 < thp1)
	      *th1 = thp1 - TWO_PI;
	    else {
		if (*th1 > thp1)
		  *th1 = thp1 + TWO_PI;
		else
		  *th1 = thp1;
		}
	    overlap = 1;
	}
    }
    else {
	if (theta_within_range_right(thp2, *th1, *th2)) {
	    if (*th2 < thp2)
	      *th2 = thp2 - TWO_PI;
	    else {
		if (*th1 > thp2)
		  *th2 = thp2 + TWO_PI;
		else
		  *th2 = thp2;
		}
	    overlap = 1;
	}
    }
    if ((theta_within_range_left(*th1, thp1, thp2)) &&
	(theta_within_range_right(*th2, thp1, thp2))) {
	overlap = 1;
    }
    return(overlap);
}

static double dist_to_pline (pline, theta)
    struct polar_line_segment_type *pline;
    double theta;
{
    double x01, y01, x12, y12, dist;

    if (theta_within_range_both(theta, pline->t1, pline->t2)) {
	x01 = pline->x01;
	y01 = pline->y01;
	x12 = pline->x12;
	y12 = pline->y12;
	dist = fabs((y01 * x12 - x01 * y12) 
		    / (x12 * sin(theta) - y12 * cos(theta)));
    }
    else
      dist = Max_range;
    return (dist);
}


static long closer_pline (pline1, pline2, th1, th2)
    struct polar_line_segment_type *pline1, *pline2;
    double th1, th2;
{
    double r11, r12, r21, r22, del = 0.001;

    if (th1 == pline1->t1) 
      r11 = pline1->r1;
    else
      r11 = dist_to_pline (pline1, th1);

    if (th1 == pline2->t1) 
      r21 = pline2->r1;
    else
      r21 = dist_to_pline (pline2, th1);

    if (th2 == pline1->t2) 
      r12 = pline1->r2;
    else
      r12 = dist_to_pline (pline1, th2);

    if (th2 == pline2->t2) 
      r22 = pline2->r2;
    else
      r22 = dist_to_pline (pline2, th2);

    if (((r11 < (r21 + del)) && (r12 < (r22 + del))) ||
	((r11 < (r22 + del)) && (r12 < (r21 + del))))
      return (1);
    else
      return (0);
}

static struct polar_line_segment_type *clip_pline_list_angle_range
  (struct polar_line_segment_type *pline_list, double theta1, double theta2)
{
    struct polar_line_segment_type *pline_ptr, *new_pline, 
                                   *new_pline_ptr, *new_pline_list;
    double th1, th2;
    long i = 0;

    new_pline_ptr = 0; new_pline_list = 0;
    pline_ptr = pline_list;
    while (pline_ptr) {
	th1 = pline_ptr->t1;
	th2 = pline_ptr->t2;
	if (theta_overlap(&th1, &th2, theta1, theta2)) {
	    new_pline = get_pline();
	    if ((new_pline == pline_ptr) && (i < 4)) {
	      printf("begin ptr %ld,  end ptr %ld\n",
		     pline_memory.begin_ptr, pline_memory.end_ptr) ;
	      i++;
	    }
	    if (th1 > th2) th1 = th1 - TWO_PI;
	    if ((th2 - th1) > TWO_PI) th1 = th1 + TWO_PI;
	    new_pline->t1 = th1; new_pline->t2 = th2;
	    new_pline->r1 = dist_to_pline(pline_ptr, th1);   
	    new_pline->r2 = dist_to_pline(pline_ptr, th2);   
	    new_pline->x01 = pline_ptr->x01; 
	    new_pline->x12 = pline_ptr->x12;
	    new_pline->y01 = pline_ptr->y01; 
	    new_pline->y12 = pline_ptr->y12;
	    new_pline->n   = pline_ptr->n;
	    new_pline->next_pline = 0;
	    if (new_pline_ptr) {
		new_pline_ptr->next_pline = new_pline;
	    }
	    else {
		new_pline_list = new_pline;
	    }
	    new_pline_ptr = new_pline;
	}
	pline_ptr = pline_ptr->next_pline;
      }
    return (new_pline_list);
}

/********
  filling an array theta_list of angles of vertices in
  pline_list.
********/


static struct polar_line_segment_type *eliminate_null_plines (pline_list)
    /* elimiante all plines that have empty theta interval, i.e., t1 == t2 */
    struct polar_line_segment_type *pline_list;
{
    struct polar_line_segment_type *pline_ptr, *prev_pline_ptr, *free_ptr;
    double del = 0.01;
    double t;

    prev_pline_ptr = 0;
    pline_ptr = pline_list;
    while (pline_ptr) {
	free_ptr = NULL;
	/*****
	if (((pline_ptr->t2 - pline_ptr->t1) > del) &&
	    ((pline_ptr->t2 - pline_ptr->t1) < (TWO_PI - del))) {
	    prev_pline_ptr = pline_ptr;
	}
	******/
	t = normalize_angle(pline_ptr->t1 - pline_ptr->t2,-M_PI);
	if( fabs(t) > del)
	  prev_pline_ptr = pline_ptr;

	else { /* null pline */
	    if (prev_pline_ptr) {
		prev_pline_ptr->next_pline = pline_ptr->next_pline;
		free_ptr = pline_ptr;
	    }
	    else {
		pline_list = pline_ptr->next_pline;
		free_ptr = pline_ptr;
	    }
	}
	pline_ptr = pline_ptr->next_pline;
	if (free_ptr)
	  free_pline (free_ptr);
    }
    return (pline_list);
}

/**
 ** visible_polar_line_list  -  computes the "visibility graph"
 **                             more precisely, removed the part of segments
 **                             that are behind other segments.
 **
 ** Algorithms: 
 ** An ordered list of the angles of view (from the robot point of view) of 
 ** every segment extremity is computed. 
 ** Every angular sector between two successive angles T_i T_i+1 is considered.
 ** For each sector, every segment of the polar_line that intersects the 
 ** sector is considered. 
 ** If the current segment is nearest than the previous one, the sector
 ** [T_i T_i+1] of the previous one is removed (ie, this segment is cutted  
 ** into 2 parts). And respectively. The nearest segment is memorized and
 ** the loop go on.
 **/

static struct polar_line_segment_type *visible_polar_line_list (pline_list)
    struct polar_line_segment_type *pline_list;
    /* pline_list is modified by this function */
{
    struct polar_line_segment_type *pline_ptr, *new_pline, 
                                   *closest_pline, *change_ptr;

    double  theta1, theta2;
    AList *l;

    long i;

    /* 
       extract_theta_list(pline_list, thlist);
       sort_flist(thlist);
       delete_equal_flist(thlist); 
    */
    

    /*    print_flist(thlist); */
    Construct_list_angles(pline_list);

    l = ListAngles;

    
    i = 1;
    while (l != NULL) {

      /* Considers each angular sector (theta i , theta i+1) 
	 (with theta1 = theta i and theta2 = theta i+1) */

	/* theta1 = thli;
	theta2 = thlist[i+1];
	*/
      theta1 = l->th;
      if(l->next == NULL)
	theta2 = ListAngles->th;
      else
	theta2 = l->next->th;

      closest_pline = 0;
      pline_ptr = pline_list;

      /* For each segment */
      while (pline_ptr) {

	/* If the angular sector intersect the segment */
	if ((theta_within_range_left(theta1,pline_ptr->t1,pline_ptr->t2)) &&
	    (theta_within_range_right(theta2,pline_ptr->t1,pline_ptr->t2))) {
	  
	  while (theta2 < pline_ptr->t1) {
	    theta1 = theta1 + TWO_PI;
	    theta2 = theta2 + TWO_PI;
	  }
	  while (theta1 > pline_ptr->t2) {
	    theta1 = theta1 - TWO_PI;
	    theta2 = theta2 - TWO_PI;
	  } 
	  
	  if (!closest_pline) {
	    closest_pline = pline_ptr;
	  }
	  /* pline_ptr closer than closest */
	  else {
	    if (closer_pline(pline_ptr,closest_pline,theta1,theta2)) {
	      change_ptr = closest_pline;
	      closest_pline = pline_ptr;
	    }
	    /* pline_ptr is blocked */
	    else {
	      change_ptr = pline_ptr;
	    }
	    /* modify pline pointed to by change_ptr to eliminate 
	       the portion that is blocked */
	    if (theta_within_range(theta1,change_ptr->t1,
				   change_ptr->t2)) {
	      if (theta_within_range(theta2, change_ptr->t1,
				     change_ptr->t2)) {

		/* split into two plines */
		new_pline = get_pline();
		new_pline->x01 = change_ptr->x01;
		new_pline->x12 = change_ptr->x12;
		new_pline->y01 = change_ptr->y01;
		new_pline->y12 = change_ptr->y12;
		new_pline->t1 = theta2;
		new_pline->r1 = dist_to_pline(change_ptr,theta2);  
		new_pline->t2 = change_ptr->t2;
		new_pline->r2 = change_ptr->r2;
		new_pline->n  = change_ptr->n;
		new_pline->id  = change_ptr->id;
		change_ptr->t2 = theta1;
		change_ptr->r2 = dist_to_pline(change_ptr,theta1); 
		/* insert new_pline into the pline_list */
		new_pline->next_pline 
		  = change_ptr->next_pline;
		change_ptr->next_pline = new_pline;
	      }
	      else {
		change_ptr->t2 = theta1;
		change_ptr->r2 = dist_to_pline(change_ptr,theta1);
	      }
	    }
	    else {
	      change_ptr->t1 = theta2;
	      change_ptr->r1 = dist_to_pline(change_ptr,theta2);
	    }
	  }
	}
	pline_ptr = pline_ptr->next_pline;
      }
      l = l->next;
    }
    pline_list = eliminate_null_plines(pline_list);
    return(pline_list);
}


/***********
  normalize the angle a between p and p + 2 Pi
***********/

static double normalize_angle(double a,double p)
   
{
  double b;

/* XXX fmod does not seem usefull with the following tests
   It is really important to have a very efficent function here
   as according to gprof it is call half a million times for a map 
   of 350 segs and the code spends about 25% time here !

   b = fmod(a,2*M_PI); */
  b = a;

  while(b <= p)
    b += 2*M_PI;

  while(b > p + 2*M_PI)
    b -= 2*M_PI;

  return(b);
}

/**********
  inserts the angle a in the list
  of angles  l in non decreasing
  order.
  All angles are  normalize between 
  0 and 2 Pi
***********/

static AList *Insert_angle(AList *l, double a)
{
  double b;
  AList *la,*li;

  b = normalize_angle(a,0.);
  li = (AList *)malloc(sizeof(AList));
  li->next = NULL;
  li->th = b;

  /* the list is empty */
  if(l == NULL)
    return(li);

  /* same angle : we don't include it and FREE it ; 
     THANX TO ZHU for those stupid bugs */
  if(fabs(b - l->th) < 0.00001){
    free(li);
    return(l);
  }

  /*li is at the begining of the list */
  if(l->th > b){
      li->next = l;
      return(li);
    }
  for(la = l; la->next != NULL; la = la->next){
    
    if(fabs(la->next->th - b) < 0.00001){
      free(li);
      return(l);
    }
    if(la->next->th > b){
      li->next = la->next;
      la->next = li;
      return(l);
    }
  }
  
  /* li is at the end of the list */
  
  la->next = li;
  return(l);
  
}

static void DeleteListAngle()

{
  AList *la,*li;
  for(la = ListAngles; la != NULL;){
    li = la->next;
    free(la);
    la = li;
  }
  ListAngles = NULL;
}



/*************
  constructs the global
  list ListAngles from 
  the critical points of the list
  pline in non-decreasing order.
  All angles are between 0 and 2 Pi
*************/

static void Construct_list_angles(PolLine *pline)
{
  PolLine *ptr;

  ptr = pline;

  while(ptr){
    ptr->t1 = normalize_angle(ptr->t1,0.);
    ptr->t2 = normalize_angle(ptr->t2,0.);

    ListAngles = Insert_angle(ListAngles,ptr->t1);
    ListAngles = Insert_angle(ListAngles,ptr->t2);

    ptr = ptr->next_pline;
  }

}


/*************
  computes the shortest distance between the segment x1,y1,x2,y2 
  and the origin.
  nx,ny is the outward normal of the segment.
*************/

static double dist_to_origin(double x1,double y1,double x2,double y2,double nx,double ny)
{
  double t1,t2,dist;

  t1 = x1*ny - y1*nx;
  t2 = x2*ny - y2*nx;

  if(t1*t2 < 0) /* the origin projects on the segment */
    {
      dist = fabs(x1*nx + y1*ny);
      return(dist);
    }

  else{ /* minimum distance to one extrimity */

    t1 = sqrt(x1*x1 + y1*y1);
    t2 = sqrt(x2*x2 + y2*y2);

    if(t1 < t2)
      return(t1);
    return(t2);
  }
}


/************
  filters small angles (< 1.5 degrees)
  The segmentation couldn't provide a segment
  from it anyway (needs at least 1.5 degrees for 4 points)
*************/

static long Small_angle_filter(PolLine *pline)
{
  
  double t;

  t = normalize_angle(pline->t1 - pline->t2,-M_PI);

  if(fabs(t) < 1.5*0.0174)
    return(0);
  return(1);
}
