/************************************************************************
  This file maintains a list of possible trasformations
  PossibleTransf of structures "Hyp" defined in 
  localize.h 

  each struture has a field called Tr[3] which defines the transform
  from the robot encoder position and the absolute location 
  corresponding to the hypothesis.

  calling rob_enc[4] = (x,y,sterring,turret), the root encoder position
          abs_loc[4] = (",",  " , "  "     ), the absolute location
          
     we have

     abs_loc[0] = cos(Tr[2])*rob_enc[0] - sin(Tr[2])*rob_enc[1] + Tr[0]
     abs_loc[1] = sin(Tr[2])*rob_enc[0] + cos(Tr[2])*rob_enc[1] + Tr[1]
     abs_loc[2] = rob_enc[2] + Tr[2];
     abs_loc[3] = rob_enc[3] + Tr[2]

  At the end of the localization process, each hypothesis defines an absolute
  location abs_loc(hyp).

  -The function TransformListHyp(list_hypo,rob_enc)  
  transforms the absolute locations of every hypothesis 
  in list_hypo into  the transformation Tr. 

  rob_enc and vrob_enc is the encoder location of the robot at the time 
  the localization was made.

  
  After the robot will have moved, the absolute location corresponding 
  to each hypothesis will then be computed as above.
  
  
  Then, the process goes like this:

  -getting the new robot location and its variance after it moved
  -getting laser data at this location

  FilterBadHypo:

  -for each hypothesis, 
       -computes the corresponding absolute location
       -checks if this location is within the boudary limits 
       -check if the local data is valid for this location,
        (i.e, check if there is any data seen behind a static object).

  The  list of remaining hypothesis is stored in 
  the static local variable "PossibleTransf".

  
      
  Note about the way the structure Hyp is used in the list PossibleTransfs:
  the field ->lsm is no longer a list of matchings but a reference to
  the corresponding hypothesis in the original list ListHypo. 

**********************************************************************/

/**
  05/10/94 add the function GetListMatch() 
  returning the list (in PDBL format) of matchings (see localize.h 
  structure Smatch*) of the best current hypothesis
**/


#include <stdio.h>
#include <math.h>
#ifdef VXWORKS
#define M_PI 3.141592653589793
#endif
#include <stdlib.h>




#include <dbliste.h>
#include <lasdef.h>
#include <las_prot.h>
#include <kal_prot.h>
#include <image_mgr_p.h>
#include <definelocal.h>
#include <odometer.h>
#include <odometer_p.h>

#include "localize.h"
#include "match_p.h"
#include "map_util.h"
#include "disambiguate_p.h"
#include "LaserAc_p.h"

double VarLoc[6];

/****************
  List of possible transformations "encoder position -> absolute position"
****************/

static PDBL *PossibleTransfs = NULL;
static long NumberTransfs = 0;

/*************
  Map structure and an auxilliary for 
  its local representation (in robot coordinate system)
*************/
static igla *Map;
static igla *LocalMap;

/**************
  Structure for the boundary.
  A location will be rejected if the robot doens't
  belong to the interior of this boundary
***************/
static igla *boundary;

/************
  for memorizing the best hypothesis
  computed by the Localization process
*************/
static Hyp *Best = NULL;

/* prototype of added functions (by David Zhu) */
long DeleteHypo(long Trx, long Try, long Trth, long *possloc);




/**********************************************************************
-The function TransformListHyp()  
  transforms the absolute locations of every hypothesis 
  in list_hypo into a transformation "encoder position -> absolute position"
  See header of this file.

  lhyp = list of hypothesis to be transformed
  map = map !
  local_map = empty (but allocated !) struture for containing a local 
              representation of the map.
  bound = structure containing the boundary (can be NULL).
  xrob = encoder position of the robot when the localization was made.
  possloc = array of all possible locations.
  WARNING: possloc has to be allocated for 500 hypothesis = 2001 long integers.
  If there is more than 500 hypothesis, the remaining ones are not copied.

  Returns the index of the best location in the array possloc.
***********************************************************************/

long TransformListHyp(PDBL *lhyp,igla *map, igla *local_map,igla *bound,
		      double *xrob,long *possloc, int maxarraysize)
{
  PDBL *l;
  Hyp *hyp,*hyptr;
  double score = 0;
  long best_index = -1;
  
  Best = NULL;
  /***********
    freeing previous list
  ***********/

  if(PossibleTransfs != NULL)
    {
      for(l = PossibleTransfs; l != NULL; l = dls(l)){
	free(dlel(l));
      }
      dlib_list(PossibleTransfs);
      PossibleTransfs = NULL;
      NumberTransfs = 0;
      Best = NULL;
    }

  /********
    storing the map and the local version of it
  ********/

  Map  = map;
  LocalMap = local_map;

  /********
    storing the boudary, if any
  ********/
  boundary = bound;

  /**********
    transforming the input list from absolute localization to transformation
    robot encoder position -> absolute position
  ***********/



  for(l = lhyp; l != NULL; l = dls(l)){
    hyp = (Hyp *)dlel(l);

    /*******
      I added a field called "->length_matched in which
      the total length matched is computed by this function
    *******/
    Compute_length_matched(hyp);
    if(hyp->nmatched >= 2)
      {
	/*****
	  creation of the transform hypothesis
	*****/

	hyptr = (Hyp *)malloc(sizeof(Hyp));

	/****
	  storing info from the original hypothesis
	****/
	hyptr->nmatched = hyp->nmatched;
	hyptr->length_matched = hyp->length_matched;
	hyptr->lsm = hyp->lsm;

	/********
	  looking for the best hypothesis
	  selected by the localization
	********/
	
	if(hyp->length_matched >= score)
	  {
	    score  = hyp->length_matched;
	    Best = hyptr;
	  }
	  
	
	/* if(hyp->nmatched >= score)
	  {
	    score  = hyp->nmatched;
	    Best = hyptr;
	  }*/ 
	/********
	  computing the transform
        ********/

	CalcTransf(hyp->Tr,hyp->Vtr,xrob,hyptr->Tr,hyptr->Vtr);

	/******
	  adding to the list
        ******/

	PossibleTransfs = dlhcons(PossibleTransfs,hyptr);

	/*****
	  copying into the array possloc
	  WARNING: the array has to be big enough !!!
	  = 4*number of possible location + 1
	******/

	if(NumberTransfs < maxarraysize){
	  possloc[4*NumberTransfs + 1] = hyptr->Tr[0]*1000;
	  possloc[4*NumberTransfs + 2] = hyptr->Tr[1]*1000;
	  possloc[4*NumberTransfs + 3] = hyptr->Tr[2]*1000;
	  possloc[4*NumberTransfs + 4] = hyptr->nmatched;
	  
	  /*****
	    storing the best location
	    *****/

	  if(Best == hyptr) {
	    best_index = NumberTransfs;


	    /* XXXXXX On reforce Glob_pos et Vpos car il semble
	       qu'on ait perdu ce dernier .... */
/* 	    printf("TransformListHyp %f %f %f (%f %f %f)\n", */
/* 		   hyp->Tr[0], hyp->Tr[1], hyp->Tr[2], */
/* 		   hyp->Vtr[0], hyp->Vtr[2], hyp->Vtr[5]); */
	    Glob_pos[0] = hyp->Tr[0];
	    Glob_pos[1] = hyp->Tr[1];
	    Glob_pos[2] = hyp->Tr[2];
	    Vpos[0] = hyp->Vtr[0];
	    Vpos[1] = hyp->Vtr[1];
	    Vpos[2] = hyp->Vtr[2];
	    Vpos[3] = hyp->Vtr[3];
	    Vpos[4] = hyp->Vtr[4];
	    Vpos[5] = hyp->Vtr[5];
	  }
	}
	else {
	  printf("TransformListHyp Warning: more than %d possibilities !\n", 
		 maxarraysize);
	  break;
	}

	NumberTransfs++;

	
      }
  }
  
  if(best_index > -1)
    /*** initializing the robot position and variance from the best hypothesis */
    init_tr_st(Best->Tr,Best->Vtr);
  
  
  /*** returning the result *****/
  
  possloc[0] = NumberTransfs;
  /*printf("\n");
  printf("%ld possible locations\n",NumberTransfs);*/
  return(best_index);
}


PDBL *GetListMatch(void)
{
  return(Best->lsm);
}

/* Mainly for debug purpose */
Hyp *GetBestHyp(void)
{
  return Best;
}

/* Mainly for debug purpose */
PDBL *GetPossibleTranfs()
{
  return PossibleTransfs;
}

/****************
  Filters bad hypothesis by:
  

 Returns -1 if no possibilities
***************/




long FilterBadLocations(long n_data,long *possloc, double distmax, int maxarraysize)
{
  PDBL *l;
  static double pos[4];
  static double vpos[6];
  Hyp *hyptr,*curr_hyp=NULL;
  long currentnumber = 0;
  long best = -1;
  long score = -1;
  igla *data;


  /****
    getting the pointer on the data structure
  ****/

  data = PtoIm(n_data);

  /*****
    copying the encoder values data->enc in the image data 
  ******/

  /*Set_enc_image(data,state);*/

  /******
    copying the encoder values in the odometer data
    Side effect: Glob_pos will be recomputed according to
    the current transform saved in Loc2Abs[] in odom_sim.c
  *******/
  
  /*set_encoders(data->enc);*/

  for(l = PossibleTransfs; l != NULL; l = dls(l)){
    hyptr = (Hyp *)dlel(l);

    if(hyptr->nmatched < 2)
      /*** means this possibility has already been discarded ***/
      continue;

    /************
      computing the absolute position corresponding
      to the transform hyptr->Tr and the current encoder 
      position rob_enc and its variance Vrob_enc.
      Computes also the variance on the new position(not used yet)
    *************/
  
    Calc_abs_pos(rob_enc,Vrob_enc,hyptr->Tr,hyptr->Vtr,pos,vpos);

    /*********************
      test if the tested position is inside the boundary
    *********************/

    if(!InsidePoly(pos,boundary)){
      /*** discarding this possibility ***/
       hyptr->nmatched = 0;
       if(hyptr == Best)
	 printf("FilterBadLocations Warning:  you are loosing your best hypothesis\n");
      NumberTransfs--;
     }
    else{
      /*********
       computing the map in the local coordinate system
      *********/
      tr_abs_to_loc(Map,LocalMap,pos,data->coosys);

      /****************
       test if something is seen behind a static object of the Map
      *****************/
  
      if( !hypo_is_valid(data,LocalMap,distmax)){
	/*** discarding this possibility ***/

	hyptr->nmatched = 0;
	NumberTransfs--;
      }
    }

    if(hyptr->nmatched >= 2){
      printf(" Remaining hypothesis x = %.5f y = %.5f theta = %.5f \n",pos[0],pos[1],180*pos[2]/M_PI);
      curr_hyp = hyptr;

      /*****
      copying into the array possloc
      WARNING: the array has to be big enough !!!
      = 4*number of possible location + 1
      ******/
      if(currentnumber < maxarraysize){
	/*** everything is expressed in millimeters and hundredth of degrees ****/
	possloc[4*currentnumber + 1] = hyptr->Tr[0]*1000;
	possloc[4*currentnumber + 2] = hyptr->Tr[1]*1000;
	possloc[4*currentnumber + 3] = hyptr->Tr[2]*18000/M_PI;
	possloc[4*currentnumber + 4] = hyptr->nmatched;

      }
      else {
	printf("FilterBadLocations Warning: more than %d possibilities !\n",
	       maxarraysize);
	break;
      }
      /***
	storing the best location
      ***/
       if(hyptr->nmatched >= score)
	{
	  score = hyptr->nmatched;
	  best = currentnumber;
	}
      
      
      currentnumber++;

     
    }

  }

  if(NumberTransfs > 1){
    printf("%ld remaining locations \n",NumberTransfs);
  }

  else if(NumberTransfs == 1){
    printf("Only possible remaining location\n");
    if (curr_hyp == NULL) {
      printf ("FilterBadLocations : error (curr_hyp not initialized)\n");
    }
    Best = curr_hyp;
  }
  
  else{
    printf("All the possibilities have been discarded:\n");
    printf("Maybe the encoder drift is too important:\n");
    printf("or the map is out of date\n");
    printf("Rerun the Localization Process \n");
  }
  possloc[0] = currentnumber;

  /*****
    freeing the data structure
  ******/

  Gfree_image(n_data);

  return(best);
}


/**************
delete all hypothesis corresponding to Trx Try and Trth
**************/


long DeleteHypo(long Trx, long Try, long Trth, long *possloc)
{
  PDBL *l;
  Hyp *hyptr;
  long currentnumber = 0;
  long deleted = 0;

  for(l = PossibleTransfs; l != NULL; l = dls(l)){
    hyptr = (Hyp *)dlel(l);
    if (hyptr->nmatched >= 2) {
      if ((abs(Trx - (long)(hyptr->Tr[0]*1000.0)) == 0) && 
	  (abs(Try - (long)(hyptr->Tr[1]*1000.0)) == 0) && 
	  (abs(Trth - (long)(hyptr->Tr[2]*18000.0/M_PI)) == 0)) {
	deleted = 1;
	NumberTransfs--;
	hyptr->nmatched = 0;
      }
      else {
	possloc[4*currentnumber + 1] = hyptr->Tr[0]*1000;
	possloc[4*currentnumber + 2] = hyptr->Tr[1]*1000;
	possloc[4*currentnumber + 3] = hyptr->Tr[2]*18000/M_PI;
	possloc[4*currentnumber + 4] = hyptr->nmatched;
	currentnumber++;
      }
    }
  }
  possloc[0] = currentnumber;
  if (deleted)
    return(1);
  else
    return(0);
}


/** returns x,y and theta in meters and degrees **/
long BestCurrentLocation(double *x,double *y,double *theta)
{
  static double pos[4];
  static double vpos[6];


  if(Best == NULL)
    return(0);
  /****
    getting the latest position of the robot and its variance.
    Encoder values in rob_enc and Vrob_enc
  ****/

  
  
  /*******
    computes the current location corresponding 
    to the Best transform so far.
  *******/

  Calc_abs_pos(rob_enc,Vrob_enc,Best->Tr,Best->Vtr,pos,vpos);
  *x = pos[0]; *y = pos[1]; *theta = pos[2]*180./3.14159;
  
  return(1);
}




/************
  compute the length of the total overlaping
  for the hypothesis hyp 
************/


void Compute_length_matched(Hyp *hyp)
{
 PDBL *l;
 Smatch *m;
 igla ig;
 Sgl *s1,*s2;
 double length_matched;
 double angle_matched, angle_not_matched;
 double underlapPercent;


 /******
   setting the position of the auxilliary image ig
 ********/

 ig.x = hyp->Tr[0]; ig.y = hyp->Tr[1]; ig.cs = cos(hyp->Tr[2]);
 ig.sn = sin(hyp->Tr[2]); ig.coosys = 'S';
 

 /********
   initializing the total length
 ********/

 hyp->length_matched = 0.;

 
 /*******
   reading the list of matchings
 ********/
 for(l = hyp->lsm; l != NULL; l = dls(l)){
   
   /****
     getting the matching structure
   ****/

   m = (Smatch *)(l->el);
   
   /*******
     getting the perceptual segment
   *******/

   s1 = m->sgp;
   
   /******
     getting the map segment 
   ******/

   s2 = m->sgm;

   /********
     computing the overlap
   *********/
/*    angle_matched = angleOverlap(&ig,s1,s2, &angle_not_matched);  */
/*    underlapPercent = angle_not_matched/(angle_not_matched+angle_matched); */
/*    length_matched2 = Test_overlap2(&ig,s1,s2, &length_not_matched);   */
   length_matched = MaxOverlap(&ig,s1,s2); 
/*    printf ("l %.2f, u %.2f (%d), a %.3f\n",  */
/* 	   length_matched,  */
/* 	   length_matched*(underlapPercent/(1-underlapPercent)), */
/* 	   (int)(100*underlapPercent), */
/* 	   angle_matched); */
/*    if (length_matched <= 0.0) { */
/*      printf ("Negative length match\n"); */
/*    } */
   hyp->length_matched += length_matched;

 }

}




long LasSegmentation(long nbp,long *laser_val)
   
{
  long im;
  

  im = SegmCreate_image(nbp,laser_val);
  if(im >= 0 ){
    if(PtoIm(im) != NULL){
      printf("Image number %ld : %d points + %ld segments \n",
	     im,PtoIm(im)->nbp,PtoIm(im)->nsg); 
      
      Set_pos_image(PtoIm(im),Glob_pos[0],Glob_pos[1],Glob_pos[2],Glob_pos[3]);
      return(im);
    }
  }
  return(-1);
}  




















