#include <stdio.h>
#include <math.h>
#ifdef VXWORKS
#define M_PI 3.141592653589793
#endif
#include <stdlib.h>
#include <string.h>

#include <dbliste.h>
#include <lasdef.h>
#include <las_prot.h>
#include <las_io_p.h>
#include <image_mgr_p.h>
#include <odometer.h>
#include <odometer_p.h>
#include <odom_funct_p.h>

#include "lasers_robots.h"
#include "map_mgr_p.h"
#include "map_builder_p.h"
#include "map_builder_private_p.h"
#include "map_unc.h"
#include "map_util.h"
#include "localize.h"
#include "LaserAc_p.h"
#include "match_p.h"
#include "disambiguate_p.h"
#include "user.h"
#include "com_common.h"

/*********************
  Control parameters
********************/

/** 
  threshold for relocalizing when 
  the "absolute localization" didn't work
**/ 

static double Thres = 12.;

/***
  thresholds for the auto-fusion of the map
***/

static double gap_ang = 8.; /* acceptable angular gap (in degrees) */


/*** 
  defines the minimum aperture defined by the data segment
  Not really useful in real case: can be filter out by the segmentation 
  setting
 ***/

static double DATA_ANG = 1.;

/*****
   maximum variance in angle for a segment 
*****/

static double MAX_VAR_ANGLE = 0.001;

/*****
   maximum variance in distance for a segment 
*****/

static double MAX_VAR_DIST = 0.1;


/*** minimum length ***/
#ifdef notmetric
static double mid_distance = 3.0; /* acceptable distance of the mid-point
				of one segment from the other segment */
static double min_length = 25.; /** inches^2 **/
/** square of the maximum distance for linking **/
static double link_dist = 9.;
#else
static double min_length = 0.04; /** meter^2 **/
static double link_dist = 0.00064;
static double mid_distance = 0.08; /* acceptable distance of the mid-point
				of one segment from the other segment */
#endif


/****************
  private auxilliary structures
****************/

#define NB_MAX_POSSIBILITIES 100
static long Possibilities[4*NB_MAX_POSSIBILITIES+1];
static igla *local_map = NULL;

static igla *Map = NULL;
static long Map_ind;

static int AutoSave = 0;
static char File[100] = "Nomad200MapFile.tmp";


extern PDBL TEST_PDBL;


/****************
**************
Main function
*************
****************/


/****
  returns an array of segments composing the new map.
  the first 4 integers: x,y,steering,turret corrected
  then:
  number of segments (array[4])
  for i=1 t0  array[4]
     x1[i] = array[4*i + 1]
     y1[i] = array[4*i + 2]
     x2[i] = array[4*i + 3]
     y2[i] = array[4*i + 4]
  end for


  Units: tenth of inch, tenth of degree.

  Orientation: free space on right side when going from
  (x1,y1) to (x2,y2).

  WARNING: this array is allocated inside the function.
           you're responsible for freeing it.
*****/


long *DataIntegration(long data, laserConfiguration *sick_config)
{
  int err = 0;

  if(data < 0)
    err =1;
  else if(PtoIm(data) == NULL)
    err = 1;
  else if(PtoIm(data)->lsgl == NULL)
    err = 1;
  if(err)
    {
      printf("No segments found; check your laser\n");
   
    }
  else{
    TreatData(data, sick_config);
  }

  return(GetCurrentMap());
  
 
}

/***************
 * FUNCTION:     GetCurrentMap
 * PURPOSE:        returns an array of segments composing the current map.
                   the first 4 integers: x,y,steering,turret corrected
		   then:
		   number of segments (array[4])
		   for i=1 t0  array[4]
		   x1[i] = array[4*i + 1]
		   y1[i] = array[4*i + 2]
		   x2[i] = array[4*i + 3]
		   y2[i] = array[4*i + 4]
		   end for
		   
		   
		   Units: tenth of millimeters, milliradiants.
		   
		   Orientation: free space on right side when going from
		   (x1,y1) to (x2,y2).
		   
		   WARNING: this array is allocated inside the function.
		   you're responsible for freeing it.
 * ARGUMENTS:    None
 * ALGORITHM:    
 * RETURN:       long *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long *GetCurrentMap(void)
{
  long *map_int;
  PDBL *l;
  Sgl *s;
  int i = 1;
  if(Map == NULL)
    return(NULL);

  map_int = (long *)malloc(4*(Map->nsg + 2)*sizeof(long));

  map_int[0] = Glob_pos[0]*1000.;
  map_int[1] = Glob_pos[1]*1000.;
  map_int[2] = Glob_pos[2]*1800/M_PI;
  map_int[3] = Glob_pos[3]*1800/M_PI;
  

  for(l = Map->lsgl; l != NULL; l = dls(l)){
    s = (Sgl *)dlel(l);
    map_int[4*i+1] = 1000.*s->first.x ;
    map_int[4*i+2] = 1000.*s->first.y;
    map_int[4*i+3] = 1000.*s->last.x;
    map_int[4*i+4] = 1000.*s->last.y;
    i++;
  }
  map_int[4] = i-1;
  return(map_int);
}

/***************
 * FUNCTION:     GetCurrentMapAddr
 * PURPOSE:      returns the address of Map variable
		   
                 In order to get its own usefull data 
		 (GetCurrentMap not always adapted)

 * ARGUMENTS:    None
 * ALGORITHM:    
 * RETURN:       long *
 * SIDE EFFECT:  none
 * CALLS:        
 * CALLED BY:    segloc module
 ***************/
igla *GetCurrentMapAddr(void)
{
  return Map;
}

/*** here comes the image after scanning + local processing **/
/* CALLED BY: DataIntegration */

static void TreatData(long data, laserConfiguration *sick_config)
{
  double mean_ang;

  /* allows plus or minus 30 degrees around
     the estimated position */
  
  double dev = 30.;
  igla *data_ig;
  PDBL *lmatch;
  long Relocalized = 0;


  data_ig = PtoIm(data);
  
  /** filtering **/

  FilterOut(data_ig,FilterData);
   
  /** turns on the filtering for relocation **/
  /** just in case it was turned off by somebody ***/

    LocFilterOn(); 

  /** Map doesn't exists yet **/

  
  if(Map == NULL){
    

    /** Initialize the map in Map_mgr.c ****/
    SetMap(data);
    Map = data_ig;
    Map_ind = data;

    /* sets the boundary to null */
    SetBoundary(-1);

    /** 
      initialize the odometer to the position of
      the robot at this time; 
      Sets the current global position to be 0,0,0 
      The turret angle is then computed so it is consistant
      with the encoder values
    **/

    init_encoders(data_ig->enc);
    
    /** preparing the local map for next time **/

    local_map = NULL;

    return;
  }
 
    /** creating the auxilliary **/
    local_map = Imgl_create_copy(Map);

    /** setting for the localization **/
    SetLocalMap(local_map);    
 

  /** auto saving if on */
  
  if(AutoSave){
    printf("auto saving map in %s \n",File);
    wrt_file_sg_c(Map,File);
  }
  
  
  /****
    In simulation, we need to 
    set the position of the State vector from the
    user side.
    position = point of view of "data".
    The variance is computed in odom_sim 
    according to the general model of 
    the odometers. The move is assumed = (turn + straight line) from
    the previous position 
  ******/

  /* WARNING SIMULATION */

  set_encoders(data_ig->enc, data_ig->var);

  
    
  /** setting the angle restriction **/
  
  mean_ang = 180*Glob_pos[2]/M_PI;
  Set_loc_angle(mean_ang,dev);

  /* Local localizing */

  if(Local_localization(data,Glob_pos[0],Glob_pos[1],Possibilities,0,
			NB_MAX_POSSIBILITIES,sick_config) >= 0){
    
    Relocalized = 1;

    /* gets the list of matchings */
    lmatch = GetListMatch();

    /* 
       in this case Glob_pos and Vpos has been computed 
       by the localization process
    */
  }

  else {

    Relocalized = 0;
    lmatch = MapTrData(data_ig,Glob_pos,Vpos,Thres);
  
    /** re-initializing the transform from robot encoders to global position **/
    init_location_st(Glob_pos,Vpos);
        
  }

  /* appends the new data */
  
  Append2Map(data,lmatch,Glob_pos,Vpos);

  AutoFusion();

  /** freeing the local_map structure in Map_mgr.c **/
  /* I don'tthink we need a pointer local_map here though ... */
  FreeLocalMap();
  local_map = NULL;

    
  /** if !relocalized:  we need to free the list of matchings **/
  if(!Relocalized)
    free_matchings(lmatch);

  lmatch = NULL;

  
}

/** takes the map and creates a local auxilliary map used for aspect
 computation **/
 
void UpdateLocalMap()
{
  /** freeing the local_map structure in Map_mgr.c **/
  
  FreeLocalMap();

  /** preparing the local map for next time **/
  /** creating the auxilliary **/
  local_map = Imgl_create_copy(Map);

  /** setting for the localization */
  SetLocalMap(local_map);



}

  
void Erase_map(void)
{
  if(Map_ind >= 0)
    if(Map != NULL){
      Gfree_image(Map_ind);
      Map = NULL;
      if(local_map != NULL)
	FreeLocalMap();
    }
  Map_ind = -1;
  UnsetMap();
}    


void SetAutoSave(void)
{
  AutoSave = 1;
}

void UnsetAutoSave(void)
{
  AutoSave = 0;
}

void SetAutoSaveFile(char *file)
{
  strcpy(File,file);
}

/****** 
  Appends the new data "data" to the map using the set of 
  matchings in lsm and the robot position (xp) plus its variance
  (vxp)
****/


static void Append2Map(long data,PDBL *lsm,double *xp, double *vxp)
{
  igla *data_ig;
  PDBL *l,*lm;
  Sgl *sg,*sgp,*sgm;
  Smatch *m;
  Sgl smap;
  static double test[4];

  data_ig = PtoIm(data);
  
  /* marking the data segment with ->flg = 0 */

  for(l = data_ig->lsgl; l != NULL; l = dls(l)){
    sgp = (Sgl *)(dlel(l));
    sgp->flg = 0;
  }

  /* merging matched segments (+ setting their flg to 1) */

  for(lm = lsm; lm != NULL; lm = dls(lm)){
    m = (Smatch *)(dlel(lm));

    sgp = m->sgp;
    sgp->flg = 1;
    /* 
       retreiving the global segment from the aspect
       
    */

    sgm = (m->sgm)->sfirst;
    
    /** sg = projection of sgp in xp coord. syst. **/

    project_sg_unc(sgp,&smap,xp,vxp);
    
    /* for now we don't deal with connections */
    
    /* merging sgm+smap -> sgm*/

    abs_segments_fusion(sgm,&smap,sgm);
    
    /* computing extremities leading to the maximum length of sgm */
    overlap_abs(sgm,&smap,test);

    Maximum_length_seg(sgm,&smap,test);

    
  }
  /** LEAK ?? */
  for(l = data_ig->lsgl; l != NULL; l = dls(l)){
    sgp = (Sgl *)(dlel(l));
    if(sgp->flg == 0){
      sg = (Sgl *)malloc(sizeof(Sgl));
      project_sg_unc(sgp,sg,xp,vxp);
      Map->lsgl = dlhcons(Map->lsgl,sg);
      Map->nsg++;
    }

  }

}

/*************
  computes in sg1 the union of sg1 - sg2 which are supposed
  "superimposed". Takes the maximum extension 
  of the overlapping
**************/

static void Maximum_length_seg(Sgl *sg1,Sgl *sg2,double *test)
{
  
  if(test[0] < 0)
 
	/**** begining at sg2->first ( unless sg1->typf == V : not
	 implemented ) ***/
	{
	    sg1->first = sg2->first;
	    	  
	 }

  if(test[3] > 0) 

    /**** ending at sg2->last unless sg1->typl == V *****/
    {
	  
      sg1->last = sg2->last;
            
    }
  
  
  /***** otherwise, the good extremities are already in sg1 ***/
  
  /***** projection of the extremities on the new sg1 ***/
  
  proj_orth(sg1,&(sg1->first),&(sg1->first));
  proj_orth(sg1,&(sg1->last),&(sg1->last));
      
}     

/*** 
  in case we could not relocate, tries to match directly
  - (still relies on the aspect) 
  - for each data segment :
      -computes the probabilistic distance with all segments
       in the aspect. Takes the closest
       if the distance is less than thres: 
         - computes a new position for the robot
         - adds the matching to a list of matchings
      -appends data to map
***/

PDBL *MapTrData(igla *data,double *pos,double *vpos,double thres)
{
  
  PDBL *lasp,*ldat;
  igla *aspect;
  PDBL *lmatch = NULL;
  Sgl *s_map,*s_data,*s_best;
  static igla ig; 
  double dist, min_dist=100000;
  Smatch *match;

  double tmp_pos[4];
  double tmp_vpos[6];

  memcpy(tmp_pos, pos, 4*sizeof(double));
  memcpy(tmp_vpos, vpos, 6*sizeof(double));

  aspect = PtoIm(Get_aspect_num());

  /** 
    well, this is strange ...historically, the overlapping test
    in sg_tr_sg_dist() needs an igla structure to work with. 
    should probably  be changed ..
  **/

  ig.coosys = 'S'; /** coord. sys = steering **/
  ig.x = tmp_pos[0]; ig.y = tmp_pos[1]; ig.cs = cos(tmp_pos[2]); ig.sn = sin(tmp_pos[2]);
  
  for(ldat = data->lsgl; ldat != NULL; ldat = dls(ldat)){
    s_data = (Sgl *)ldat->el;
    s_best = NULL;
    for(lasp = aspect->lsgl; lasp != NULL; lasp = dls(lasp)){
      s_map = (Sgl *)lasp->el;
      
      /** 
	  test : keeping only the best. 
	  Might not be the best way to do this. Experience will tell ... 
      **/
      
      /* XXXX Possible BUG : sg_tr_sg_dist modifies the position 
	 This function was previously called directly with pos and vpos 
	 parameters which are references  on Glob_pos and Vpos !!!
	 So they were modified also if the distance was not OK !? */

      /* OLI : no, I think sg_tr_sg_dist (overlap_test + Dist_maha)
	 does not modify the position.
      */
      if( (dist = sg_tr_sg_dist(&ig,tmp_pos,tmp_vpos,s_map,s_data,1)) >= 0.)

	if(dist <= thres){
	  
	  /* XXXX second test XXX */
	  /** relocalizing **/
	  Update2_pos(tmp_pos,tmp_vpos,s_data,s_map);

	  if (pos_jump_filter(pos, vpos, tmp_pos)) {
	    if( dist < min_dist || (!s_best)){
	      min_dist = dist;
	      s_best = s_map;

	    }
	  }

	}
      
    }

    if(s_best){
      /** adding to the list of matchings **/
      match = (Smatch *)malloc(sizeof(Smatch));
      match->sgp = s_data;
      match->sgm = s_best;
      lmatch = dlhcons(lmatch,match);
      /** relocalizing **/
      Update2_pos(pos,vpos,s_data,s_best);
    }
  }
  return(lmatch);
}


/***
  tries to merge segments in the map itself in case they've been
  forgotten by the localization process.
  Computes intersections.
  NOTE : this will actually add crossed references between segments intersecting each other.
  However, this is purely local and shouldn't be used nor be relied upon by the user. Some functions
  like RemoveSegment will break those links without telling (which they should !!) . I will have to go 
  through all these cases ... one day !  
***/

static void AutoFusion(void)
{
  
  PDBL *l,*l2;
  Sgl *s,*s2;
  double x,y;
  static double test[4];
  double min_f, min_l,dist;
  Sgl *slast,*sfirst;

  /** 
    first, we clear the intersections since the merging 
    could have broken them
    Note: all that can be done in one pass; this is a little
    bit dumb but easier to debug for a first try. 
  **/

  for(l = Map->lsgl; l != NULL; l = dls(l)){
 
    s = (Sgl *)l->el;
    s->slast = s->sfirst = NULL;
    s->typl = s->typf = 'B';
  }
  



  /** merging **/
  
  for(l = Map->lsgl; l != NULL; l = dls(l)){
    s = (Sgl *)l->el;
    for(l2 = dls(l); l2 != NULL; l2 = dls(l2)){
      s2 = (Sgl *)l2->el;
      /** testing angles < gap_ang degrees**/
      if(fabs(norm_ang(s2->o - s->o,-M_PI)) < gap_ang * 0.017){
	/** middle point of s2 over s1 **/
	x = (s2->first.x + s2->last.x)/2;
	y = (s2->first.y + s2->last.y)/2;
	if(fabs(s->vx*x + s->vy*y + s->r) < mid_distance){

	  /** testing overlapping (see lasutil.c) **/
	  overlap_abs(s,s2,test);
	  if(test[2] >= 0 && test[1] <= 0){
	    /** merging in s **/
	    abs_segments_fusion(s,s2,s);
	    /** maximum length **/
	    Maximum_length_seg(s,s2,test);
	    /** removing s2 from the map **/
	    l2 = dlp(l2);
	    Sgl_free(s2);
	    Map->lsgl = dldel(Map->lsgl,dls(l2));
	    Map->nsg--;
	  }
	}
      }
     }
  }

  /** extremity computation **/ 

  for(l = Map->lsgl; l != NULL; l = dls(l)){
    s = (Sgl *)l->el;

    if(s->typf == 'V' && s->typl == 'V')
      /* this segment already have what it needs **/
      continue;

    min_f = min_l = 100000.;
    slast = sfirst = NULL;

    for(l2 = dls(l); l2 != NULL; l2 = dls(l2)){
      s2 = (Sgl *)l2->el;

      /** testing angles **/
      /** we ask for 25 degrees difference **/

      if(fabs(norm_ang(s2->o - s->o,-M_PI)) > 45. * 0.017){
	if(s->typf == 'B'){ /* looking for a first extremity 
			     if missing */
	  if(s2->typl == 'B'){
	    if((dist = (s->first.x - s2->last.x)*(s->first.x - s2->last.x)
	       + (s->first.y - s2->last.y)*(s->first.y - s2->last.y)) 
	       < min_f){
	      if(OkIntersectionF(s,s2)){
		sfirst = s2;
		min_f = dist;
	      }
	    }
	  }
	}

	if(s->typl == 'B'){ /* looking for a last extremity if missing*/
	  if(s2->typf == 'B'){
	    if((dist = (s->last.x - s2->first.x)*(s->last.x - s2->first.x)
	       + (s->last.y - s2->first.y)*(s->last.y - s2->first.y)) 
	       < min_l){
	      if(OkIntersectionF(s2,s)){
		slast = s2;
		min_l = dist;
	      }
	    }
	    
	  }
	}
      }
    }

    if(min_f < link_dist){
      intersg(s,sfirst,&(s->first));
      sfirst->last = s->first;
      s->sfirst = sfirst;
      sfirst->slast = s;
      sfirst->typl = s->typf = 'V';
    }

    if(min_l < link_dist){
      intersg(s,slast,&(s->last));
      slast->first = s->last;
      s->slast = slast;
      slast->sfirst = s;
      slast->typf = s->typl = 'V';
    }
    
  }
}
    



/*** test to see if the intersection point is not too far from any of the extremity points ****/ 

static int OkIntersectionF(Sgl *s1,Sgl *s2)
{
  static p2d pt;

  intersg(s1,s2,&pt);

  if((s1->first.x - pt.x)*(s1->first.x - pt.x) + (s1->first.y - pt.y)*(s1->first.y - pt.y) < link_dist)
    if((s2->last.x - pt.x)*(s1->last.x - pt.x) + (s2->last.y - pt.y)*(s2->last.y - pt.y) < link_dist)
      return(1);
  
  return(0);
}
    



static void free_matchings(PDBL *lm)
{
  PDBL *l;

  for(l = lm; l != NULL; l = dls(l)){
    free(l->el);
  }
}


/*** helper for filtering ****/
void FilterOutData(int image)
{
  FilterOut(PtoIm(image),FilterData);

}


/*** Filtering function. the function "filter" can be anything returning
     1 if you want to filter out the segment s considered ***/


static void FilterOut(igla *image,long (*filter)())
{
  PDBL *l;
  Sgl *s;

  /**  special loop for the begining of  the list ***/
  
  while(1){
    l = image->lsgl;
    /*if there is no segment left */
    if(! l )
      return;
    s = (Sgl *)l->el;
    if(filter(s)){
      l = dls(l);
      /* in case there are no next */
      if(l)
	dlp(l) = NULL;
      free(image->lsgl);
      image->lsgl = l;
      Sgl_free(s);
      image->nsg--;
    }
    else
      break;
  }
  
  /* now, the first element is "stable" */
  for(l = dls(image->lsgl); l != NULL; l = dls(l)){
    s = (Sgl *)l->el;
    if(filter(s)){
      l = dlp(l);
      dldel(image->lsgl,dls(l));
      Sgl_free(s);
      image->nsg--;
    }
  }

}
  
/** 
  returns 1 if the extremity points of the segment s 
  define a cone of less than DATA_ANG 
**/
  
long FilterData(Sgl *s)
{
  double x1,y1,x2,y2;
  double o1,o2;

  /* the next 4 lines are just for getting back 
     the segment coordinates */

  x1 = s->first.x; 
  y1 = s->first.y; 
  x2 = s->last.x; 
  y2 = s->last.y; 

  /** here you can add whatever you want **/

  if((x1-x2)*(x1-x2) + (y1 - y2)*(y1-y2) < min_length)
    return(1); /** too short **/

  o1 = atan2(y1,x1);
  o2 = atan2(y2,x2);
  
  if(fabs(norm_ang(o1 - o2,-M_PI)) < DATA_ANG*M_PI/180.)
     return(1); /** not enough aperture **/
  
  /* Filter uncertain segments */
  /* XXX to be tested and good parameters to be find-out */
/*   if(s->vo > MAX_VAR_ANGLE) */
/*     return(1); */

/*   if (s->vr > MAX_VAR_DIST) */
/*     return(1); */

  return(0); /** segment ok **/
}
 


/***************
 * FUNCTION:     RetrieveMap
 * PURPOSE:      retreives the map stored in 
                 *file in internal map builder format.
		 The map had to be saved by write_file_sg()
		 (see doc localization).
		 The local static variables Map and Map_ind
		 are set so one can continue building up
		 this same map => allows incremental map
		 building.
		 NOTE: this uses ReadBuiltMap() in Map_mgr.c.
		 This function will set automatically the map
		 for relocalization. Then, localization
		 software and map building will share the same map,
		 which is good.
		 Recomputes the local map so it is ready for localization/incrementation.

 * ARGUMENTS:    char *file :
 * ALGORITHM:    
 * RETURN:      0 if failure 1 on success 
 * SIDE EFFECT: erases any previous map from memory  
 * CALLS:        
 * CALLED BY:    
 ***************/
    


long  RetrieveMap(const char *file)
{
  int m ;
  Erase_map();
  if((m = ReadBuiltMap(file)) < 0)
    return 0;
  
  Map_ind = m;
  Map = PtoIm(Map_ind);
  UpdateLocalMap();
  return 1;
}





/***************
 * FUNCTION:     Export2MapBuilder
 * PURPOSE:      called by Map_mgr.c to export 
                 its map structure to the map builder.
 * ARGUMENTS:    long n :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    ReadMap()/Map_mgr.c
 ***************/

void Export2MapBuilder(long n)
{
  Map_ind = n;
  Map = PtoIm(Map_ind);
}




/***************
 * FUNCTION:     GetMapIndex
 * PURPOSE:      returns the index on the current map image
 * ARGUMENTS:    None
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long GetMapIndex(void)
{
  return (Map_ind);
}

/***************
 * FUNCTION:     TestRemoveSg
 * PURPOSE:      returns 1 if the extremities of test match the 
                 extremities of seg.
		 Used by the list searching function dl_member()
		 called by RemoveSegment().
 * ARGUMENTS:    Sgl *test :
 *               Sgl *seg :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    RemoveSegment()
 ***************/


long TestRemoveSg(Sgl *test, Sgl *seg)
{

    if(fabs(test->first.x - seg->first.x) <= 1/REAL2LONG &&
     fabs(test->first.y - seg->first.y) <= 1/REAL2LONG &&
     fabs(test->last.x - seg->last.x) <= 1/REAL2LONG &&
     fabs(test->last.y - seg->last.y) < 1/REAL2LONG )


    return(1);
  return(0);
}

/***************
 * FUNCTION:     RemoveSegment
 * PURPOSE:      removes the map segment matching the
                 coordinates (x1,y1) and (x2,y2)
		 units are in meter
		 free space on the right going from (x1y1) to
		 (x2, y2).
 * ARGUMENTS:    double x1 :
 *               double y1 :
 *               double x2 :
 *               double y2 :
 * ALGORITHM:    
 * RETURN:       1 if the segment has been found and removed, 0 otherwise
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

int RemoveSegment(double x1,double y1,double x2,double y2)
{

  static Sgl test;
  PDBL *l;


   test.first.x = x1;
   test.first.y = y1;
   test.last.x = x2;
   test.last.y = y2;


  /*** searching in the map segment list ***/

  l = dl_member(Map->lsgl,&test,TestRemoveSg);

  if(l != &TEST_PDBL){
    Map->lsgl = dbdel_el(Map->lsgl,l,Sgl_free);
    Map->nsg -= 1;
    return 1;
  }
  return 0;
}

/** finds the two segments defined respectively by x11,y11,x12,y12 and 
    x21,y21,x22,y22 in the map then intersects them. The intersection point will 
    be the last point of the first segment and the first of the second segment */


/***************
 * FUNCTION:     IntersectsSegments
 * PURPOSE:      
 * ARGUMENTS:    
 *               
 *               
 *               
 * ALGORITHM:    
 * RETURN:       1 if segment have been found and merged, 0 otherwise
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/
int IntersectsSegments(long x11,long y11,long x12,long y12, long x21,long y21,long x22,long y22)
{
  static Sgl test1, test2;

  /** copying coordinates in an internal structure */
#ifdef notmetric 
  test1.first.x = (double)x11/10.;
  test1.first.y = (double)y11/10.;
  test1.last.x = (double)x12/10.;
  test1.last.y = (double)y12/10.;

#else
   test1.first.x = (double)x11/1000.;
   test1.first.y = (double)y11/1000.;
   test1.last.x = (double)x12/1000.;
   test1.last.y = (double)y12/1000.;
#endif

    /** copying coordinates in an internal structure */
#ifdef notmetric 
  test2.first.x = (double)x21/10.;
  test2.first.y = (double)y21/10.;
  test2.last.x = (double)x22/10.;
  test2.last.y = (double)y22/10.;

#else
   test2.first.x = (double)x21/1000.;
   test2.first.y = (double)y21/1000.;
   test2.last.x = (double)x22/1000.;
   test2.last.y = (double)y22/1000.;
#endif

  return IntersectsSegmentsD(test1.first.x, test1.first.y,
			     test1.last.x, test1.last.y,
			     test2.first.x, test2.first.y,
			     test2.last.x, test2.last.y);
}


int IntersectsSegmentsD(double x11,double y11,double x12,double y12, double x21,double y21,double x22,double y22)
{
  
  static Sgl test;
  PDBL *l;
  Sgl *s1,*s2;
  p2d pt;
  int fst;
  /** copying coordinates in an internal structure */
   test.first.x = x11;
   test.first.y = y11;
   test.last.x = x12;
   test.last.y = y12;

  /*** searching in the map segment list ***/

  l = dl_member(Map->lsgl,&test,TestRemoveSg);
  if(l == &TEST_PDBL)
    /* not found */
    return 0;

  s1 = (Sgl *)(l->el);
  
    /** copying coordinates in an internal structure */
   test.first.x = x21;
   test.first.y = y21;
   test.last.x = x22;
   test.last.y = y22;

  /*** searching in the map segment list ***/

  l = dl_member(Map->lsgl,&test,TestRemoveSg);

  if(l == &TEST_PDBL)
    /* not found */
    return 0;

  s2 = (Sgl *)(l->el);

  /** intersection point in pt **/
  intersg(s1,s2,&pt);

  /** testing which point of s1 should be pt */
  if(((pt.x - s1->first.x) * (pt.x - s1->first.x) + (pt.y - s1->first.y) * (pt.y - s1->first.y)) < ((pt.x - s1->last.x) * (pt.x - s1->last.x) + (pt.y - s1->last.y) * (pt.y - s1->last.y))){
    s1->first = pt;
    fst = 1;
  }
  else{
    s1->last = pt;
    fst = 0;
  }
  /* now there is no choice : */

  if(fst){
    s2->last = pt;

  }
  else
  {
    s2->first = pt;

  }
  
  /* testing the coordinates change */
  /*
  p1[0] = pt.x;
  p1[1] = pt.y;
  
  p1[2] = s1->o;


  p2[0] = 0.0;
  p2[1] = 0.0;
  p2[2] = 0.0;
  ChangeMapCoordinates(p1,p2); */
  return 1;
}


/***************
 * FUNCTION:     MergeSegments
 * PURPOSE:      merges a set of segments in seg. seg[0] = number of segments.
 * ARGUMENTS:    seg[0] = number of segments.
 *               seg[1..4] : x1 y1 x2 y2  for 1st segment
 *               seg[5..8] : x1 y1 x2 y2  for 2nd segment
 *               ...
 * ALGORITHM:    
 * RETURN:       1 if segment have been found and merged, 0 otherwise
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/
int MergeSegments(double *seg)
{
  PDBL *l;
  static Sgl test;
  Sgl *sg1,*sg2;
  int i;
  double t[4];
  if(seg[0] <2) /* nothing to do ! */
    return 0;


  /* getting the first segment */
  
  

  test.first.x = seg[1];
  test.first.y = seg[2];
  test.last.x =  seg[3];
  test.last.y =  seg[4];

  
  /*** searching in the map segment list ***/

  l = dl_member(Map->lsgl,&test,TestRemoveSg);

  if(l == &TEST_PDBL)
    /* not found */
    return 0;

  sg1 = (Sgl *)(l->el);
  

  for( i = 1; i < seg[0]; i++){
    
    test.first.x = seg[i*4+1];
    test.first.y = seg[i*4+2];
    test.last.x =  seg[i*4+3];
    test.last.y =  seg[i*4+4];

    
    /*** searching in the map segment list ***/
    
    l = dl_member(Map->lsgl,&test,TestRemoveSg);
    
    if(l == &TEST_PDBL)
      /* not found */
      return 0;
    
    sg2 = (Sgl *)(l->el);
    
    
    /** merging in sg1 **/
    abs_segments_fusion(sg1,sg2,sg1);

    
    /* maximum length for sg1 */
    overlap_abs(sg1,sg2,t); /* just a helper for Maximum_length_seg() */
    Maximum_length_seg(sg1,sg2,t);

    /**removing sg2 from the map **/
    Map->lsgl = dbdel_el(Map->lsgl,l,Sgl_free);
    Map->nsg -= 1;
    
    return 1;
  }

  return 0;
}

long * RotateMap(double angle)

{
  PDBL *l;
  Sgl *s;
  double pos[3];

  pos[0] = 0;
  pos[1] = 0 ;
  pos[2] = angle;
  for(l = Map->lsgl; l != NULL; l = dls(l)){
    s = (Sgl *)dlel(l);
    project_sg_unc(s,s,pos,NULL); 
      
  }

  /** robot position **/

  RotateRobotCoordinates(pos,Glob_pos,Vpos,rob_enc,Loc2Abs,Vloc2abs);

  /*return GetCurrentMap();*/
  return NULL; /* Should return something ! */

}


/*
  take two coordinates system : 
   -starting point and angle in the current map coordinate 
   system
   - what we want this stating point/angle to become in the new 
   coord sys.
*/
long * ChangeMapCoordinates(double *start, double *end)
{
  
  PDBL *l;
  Sgl *s;
  double pos[3];
  
  
  project_coord_sys_unc(end,NULL,start,NULL,pos,NULL);

  for(l = Map->lsgl; l != NULL; l = dls(l)){
    s = (Sgl *)dlel(l);
    project_sg_unc(s,s,pos,NULL); 
      
  }

  /** robot position **/

  RotateRobotCoordinates(pos,Glob_pos,Vpos,rob_enc,Loc2Abs,Vloc2abs);

 
  /*return GetCurrentMap();*/

  return NULL;
}














