#include <stdio.h>
#include <math.h>
#ifdef VXWORKS
#define M_PI 3.141592653589793
#endif
#include <stdlib.h>

#include <dbliste.h>
#include <lasdef.h>
#include <las_struct_p.h>
#include <lasegm_p.h>
/*#include <laserinit_p.h>*/
#include <lasutil_p.h>
#include <las_unc_p.h>
#include <definelocal.h>

#include "map_util.h"


static double sp[4];
static p2d p1,p2,p3,p4;

/*****************
   test if one extremity of s1 is included in s2
 s1 is projected trough the transform in ig 
 returns 1 if yes, 0 if no.
 Uses the function Sg_pos in las_struct.c.
****************/

static double computePointsAngle(p2d *p1, p2d *p2, p2d *p3);

/***************
 * FUNCTION:     Test_overlap
 * PURPOSE:      Projects s1 through the transform 
                 defined in *ig (see Sg_pos() in las_struct.c)
		 calling pt1 and pt2 the two extremities of
		 the projected segment, this function tests
		 if the orthogonal projection of either pt1 or
		 pt2 on s2 line belongs to the segment s2.
		 In case the projection of s1 is parallel to
		 s2, this is equivalent to test the overlaping
		 of the projected s1 and s2.
		 returns 1 if overlaping, 0 otherwise.
		 Mainly used for localization and map building.
 * ARGUMENTS:    igla *ig : frame for s1 (ie, robot position)
 *               Sgl *s1 : perceived segment in ig's (ie, robot's) frame 
 *               Sgl *s2 : modele segment in global frame
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS:    Sg_pos/las_struct.c    
 * CALLED BY:    
 ***************/


#define TEST_ANGLE_OVERLAP

#ifdef TEST_ANGLE_OVERLAP
int Test_overlap(igla *ig,Sgl *s1,Sgl *s2)
{
  return Test_overlap_angle(ig,s1,s2);
}

#else /* ! TEST_ANGLE_OVERLAP */

int Test_overlap(igla *ig,Sgl *s1,Sgl *s2)
     

{
    double vx,vy,d1,d2,d3,d4;

    Sg_pos(ig,s1,sp);
    p1 = s2->first;
    p2 = s2->last;

    vx = p2.x - p1.x;
    vy = p2.y - p1.y;


    d4 = (sp[2] - p2.x)*vx + (sp[3] - p2.y)*vy;
    d1 = (sp[2] - p1.x)*vx + (sp[3] - p1.y)*vy;
    d2 = (sp[0] - p2.x)*vx + (sp[1] - p2.y)*vy;
    d3 = (sp[0] - p1.x)*vx + (sp[1] - p1.y)*vy;

    if( (d4* d1) < 0)
	return(1);

    if( (d4*d2) < 0)
	return(1);

    
    
    if((d1 * d3) < 0)
	return(1);

    if((d2 * d3) < 0)
	return(1);

    return(0);

}

#endif  /* TEST_ANGLE_OVERLAP */

/***************
 * FUNCTION:     Test_overlap_angle
 * PURPOSE:      Similar to Test_overlap but consider the ANGLE overlap
                 and test also the UNDERLAPING (ie, the part of segment s1 
		 that does not 
 * ARGUMENTS:    igla *ig : frame for s1 (ie, robot position)
 *               Sgl *s1 : perceived segment in ig's (ie, robot's) frame 
 *               Sgl *s2 : modele segment in global frame
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS:    Sg_pos/las_struct.c    
 * CALLED BY:    
 ***************/

/* WARNING : very sensible value
   (eg, for 5 degres many lost solutions) */
#define MIN_ANGLE_OVERLAP_DEGRE   1


/* Tests also part of perceived segments that does not coincid with 
   its associated model segment.
   WARNING 1:  ITS NOT SYMETRICAL:
               s1 must be the perceived segment
               s2 must be the model segment
   WARNING 2: DURING MAP BULDING DO NOT BE TO SEVERE !!

   The following is the ratio between the overlap and the underlap 
   of the perceived segment with its matching model.
   EG : 1 means that only half of the segment overlap its model.
        5 means that 5/6 of the segment does not overlap 
*/
#define MAX_ANGLE_UNDERLAP_OVER_OVERLAP 5 /* 1 normal mode
			                     5 map building */

int Test_overlap_angle(igla *ig,Sgl *s1,Sgl *s2)
{
  double angleLap;
  double angleUnderlap;

  angleLap = angleOverlap(ig, s1, s2, &angleUnderlap);
  if (angleLap<=0.0) 
    return 0;

  /* Au moins MIN_ANGLE_OVERLAP_DEGRE degres de recouvrement */
  if (angleLap < MIN_ANGLE_OVERLAP_DEGRE * 3.1415/180.) {
/*     printf ("Test_overlap: reject angle\n"); */
    return 0;
  }
  /* Au moins MAX_ANGLE_OVERLAP de recouvrement */
  /* WARNING: not very pertinent test during map building procedure */
  if (angleUnderlap > MAX_ANGLE_UNDERLAP_OVER_OVERLAP * angleLap)  {
/*     printf ("Test_overlap: reject underlap\n"); */
    return 0;
  }
  return 1;
}


/***************
 * FUNCTION:     lengthOverlap
 * PURPOSE:      Similar to Test_overlap and MaxOverlap but returns the 
                 signed overlap length (negative if segments are 
		 head-to-foot) or 0 if no overlap. Computes also
		 the underlap length 'lengthUnderlap' (ie, length of 
		 projected s1 that does) not overlap s2. The sign is 
		 identical to the returned lengthOverlap.
 * ARGUMENTS:    igla *ig : frame for s1 (ie, robot position)
 *               Sgl *s1 : perceived segment in ig's (ie, robot's) frame 
 *               Sgl *s2 : modele segment in global frame
 * ALGORITHM:    
 * RETURN:       double
 * SIDE EFFECT:  
 * CALLS:    Sg_pos/las_struct.c    
 * CALLED BY:    
 ***************/

double lengthOverlap(igla *ig,Sgl *s1,Sgl *s2, double *lengthUnderlap)
     
{
    double vx,vy,d1,d2,d3,d4;
    double lengthOverlap;

    Sg_pos(ig,s1,sp);
    p1 = s2->first;
    p2 = s2->last;

    vx = s2->vy;
    vy = -s2->vx;


    d1 = (sp[0] - p1.x)*vx + (sp[1] - p1.y)*vy;
    d2 = (sp[0] - p2.x)*vx + (sp[1] - p2.y)*vy;
    d3 = (sp[2] - p1.x)*vx + (sp[3] - p1.y)*vy;
    d4 = (sp[2] - p2.x)*vx + (sp[3] - p2.y)*vy;

    /* XXXXX LENGTH NORMALIZATION MISSIONG : 
       length should be divided by sqrt(vx*vx + vy*vy)  !!!! */


    /* No overlap */
    if ( (d1 <= 0 && d3 <= 0) ||
	 (d2 >= 0 && d4 >= 0) ) {
      lengthOverlap = 0.0;
      *lengthUnderlap = d3 - d1;
      return lengthOverlap;
    }

    /* Extremity 1 outside without head-to-foot */
    if (d1 <= 0)  {/* then d3 > 0 (overlap), and d2 <= 0 (geometricaly)  */
      if (d4 <= 0) {
	lengthOverlap = d3;
	*lengthUnderlap = -d1;
      }
      else {
	lengthOverlap = d3 - d4;
	*lengthUnderlap = d4-d1;
      }
      return lengthOverlap;
    }

    /* (From now d1 > 0) */

    /* Extremity 2 outside without head-to-foot */
    if (d4 >= 0)  {/* then d2 < 0 (overlap), d3 >= 0 (geometricaly) */
      *lengthUnderlap = d4;
      lengthOverlap = -d2;
      return lengthOverlap;
    }

    /* (From now d4 < 0) */

    /* Extremity 1 outside with head-to-foot */
    if (d2 >= 0) {
      if (d3 >= 0) {
	*lengthUnderlap = -d2;
	lengthOverlap = d4;
      }
      else {
	*lengthUnderlap = d3 - d2;
	lengthOverlap = d4 - d3;
      }
      return lengthOverlap;
    }

    /* (From now d2 < 0) */

    /* Extremity 2 outside with head-to-foot */
    if (d3 <= 0) {
      *lengthUnderlap = d3;
      lengthOverlap = -d1;
      return lengthOverlap;
    }

    /* (From now d3 > 0) */

    /* Both extremities inside (with or without head-to-foot) */
    *lengthUnderlap = 0.0;
    lengthOverlap = d3 - d1;
    return lengthOverlap;
}

/***************
 * FUNCTION:     angleOverlap
 * PURPOSE:      Similar to lengthOverlap but returns the signed
                 angle of view of the overlaping segment seen from ig.
                 (negative is the segments s1 and s2 are head-to-foot,
		 and 0 if there's no overlaping).
 * ARGUMENTS:    igla *ig : frame for s1 (ie, robot position)
 *               Sgl *s1 : perceived segment in ig's (ie, robot's) frame 
 *               Sgl *s2 : modele segment in global frame
 * ALGORITHM:    
 * RETURN:       double
 * SIDE EFFECT:  
 * CALLS:    Sg_pos/las_struct.c    
 * CALLED BY:    
 ***************/

double angleOverlap(igla *ig,Sgl *s1,Sgl *s2, double *angleUnderlap)
     
{
    double vx,vy,d1,d2,d3,d4;
    double lengthOverlap, lengthUnderlap;
    double angleOverlap;
    double underlapPercent;
    p2d pr;

    pr.x = ig->x;
    pr.y = ig->y;

    Sg_pos(ig,s1,sp);
    p1 = s2->first;
    p2 = s2->last;

    vx = s2->vy;
    vy = -s2->vx;

    d1 = (sp[0] - p1.x)*vx + (sp[1] - p1.y)*vy;
    d2 = (sp[0] - p2.x)*vx + (sp[1] - p2.y)*vy;
    d3 = (sp[2] - p1.x)*vx + (sp[3] - p1.y)*vy;
    d4 = (sp[2] - p2.x)*vx + (sp[3] - p2.y)*vy;

    /* No overlap */
    if ( (d1 <= 0 && d3 <= 0) ||
	 (d2 >= 0 && d4 >= 0) ) {
      lengthUnderlap = d3 - d1;
      *angleUnderlap = 1000; /* XXXX */
      return 0;
    }

    /* Extremity 1 outside without head-to-foot */
    if (d1 <= 0) { /* then d3 > 0 (overlap), and d2 <= 0 (geometricaly)  */
      p3 = p1;
      if (d4 <= 0) {
	lengthOverlap = d3;
	lengthUnderlap = -d1;
      }
      else {
	lengthUnderlap = d4-d1;
	lengthOverlap = d3 - d4;
      }
      underlapPercent = lengthUnderlap/(lengthUnderlap+lengthOverlap);
      p4.x = p3.x + lengthOverlap * vx;
      p4.y = p3.y + lengthOverlap * vy;
      angleOverlap = computePointsAngle(&p3, &pr, &p4);
      if (lengthOverlap < 0)
	angleOverlap = -angleOverlap;
      *angleUnderlap = underlapPercent / (1-underlapPercent) * angleOverlap;
      return angleOverlap;
    }

    /* (From now d1 > 0) */

    /* Extremity 2 outside without head-to-foot */
    if (d4 >= 0) { /* then d2 < 0 (overlap), d3 >= 0 (geometricaly) */
      p4 = p2;
      lengthOverlap = -d2;
      lengthUnderlap = d4;
      underlapPercent = lengthUnderlap/(lengthUnderlap+lengthOverlap);
      p3.x = p4.x - lengthOverlap * vx;
      p3.y = p4.y - lengthOverlap * vy;
      angleOverlap = computePointsAngle(&p3, &pr, &p4);
      if (lengthOverlap < 0)
	angleOverlap = -angleOverlap;
      *angleUnderlap = underlapPercent / (1-underlapPercent) * angleOverlap;
      return angleOverlap;
    }

    /* (From now d4 < 0) */

    /* Extremity 1 outside with head-to-foot */
    if (d2 >= 0) {
      p3 = p2;
      if (d3 >= 0){
	lengthUnderlap = -d2;
	lengthOverlap = d4;
      }
      else {
	lengthUnderlap = d3 - d2;
	lengthOverlap = d4 - d3;
      }
      underlapPercent = lengthUnderlap/(lengthUnderlap+lengthOverlap);
      p4.x = p3.x + lengthOverlap * vx;
      p4.y = p3.y + lengthOverlap * vy;      
      angleOverlap = computePointsAngle(&p3, &pr, &p4);
      if (lengthOverlap < 0)
	angleOverlap = -angleOverlap;
      *angleUnderlap = underlapPercent / (1-underlapPercent) * angleOverlap;
      return angleOverlap;
    }

    /* (From now d2 < 0) */

    /* Extremity 2 outside with head-to-foot */
    if (d3 <= 0) {
      p4 = p1;
      lengthOverlap = -d1;
      lengthUnderlap = d3;
      underlapPercent = 100*lengthUnderlap/(lengthUnderlap+lengthOverlap);
      p3.x = p4.x - lengthOverlap * vx;
      p3.y = p4.y - lengthOverlap * vy; 
      angleOverlap = computePointsAngle(&p3, &pr, &p4);
      if (lengthOverlap < 0)
	angleOverlap = -angleOverlap;
      *angleUnderlap = underlapPercent / (1-underlapPercent) * angleOverlap;
      return angleOverlap;
    }
    /* (From now d3 > 0) */

    /* Both extremities inside (with or without head-to-foot) */
    lengthOverlap = d3 - d1;
    lengthUnderlap = 0.0;
    underlapPercent = 0.0;
    p3.x = p1.x + d1 * vx;
    p3.y = p1.y + d1 * vy; 
    p4.x = p3.x + lengthOverlap * vx;
    p4.y = p3.y + lengthOverlap * vy;      
    angleOverlap = computePointsAngle(&p3, &pr, &p4);
    *angleUnderlap = 0.0;
    if (lengthOverlap < 0)
      angleOverlap = -angleOverlap;
    return angleOverlap;
}

/*****************
  computePointsAngle  -  computes the angle (p1,p2,p3)   
 *****************/

static double computePointsAngle(p2d *p1, p2d *p2, p2d *p3)
{
  double l1, l2;
  double angle, scal;
  p2d v1;
  p2d v2;

  v1.x = p1->x - p2->x;
  v1.y = p1->y - p2->y;
  v2.x = p3->x - p2->x;
  v2.y = p3->y - p2->y;

  scal = v1.x * v2.x + v1.y * v2.y;
  
  l1 = sqrt(v1.x*v1.x + v1.y*v1.y);
  l2 = sqrt(v2.x*v2.x + v2.y*v2.y);

  angle = acos(scal/l1/l2);

  /* Une bonne approximation: */
/*   printf ("(~ %.4f) ", sqrt(l13_2 - (l12-l32) * (l12-l32)) / MIN(l12,l32)); */
  return angle;
}

/*****************

  v being the direction of s1,

  this function computes an array of 4 integers:
  
  t[0] = (s2->first - s1->first).v
  t[1] = (s2->first - s1->last).v


  t[2] = (s2->last - s1->first).v
  t[3] = (s2->last - s1->last).v

  
****************/

void overlap_abs(Sgl *s1,Sgl *s2,double *t)
     

{
    double vx,vy;


    p1 = s1->first;
    p2 = s1->last;
  
    p3 = s2->first;
    p4 = s2->last;

    vx = p2.x - p1.x;
    vy = p2.y - p1.y;

    t[0] = (p3.x - p1.x)*vx + (p3.y - p1.y)*vy;
    t[1] = (p3.x - p2.x)*vx + (p3.y - p2.y)*vy;
    t[2] = (p4.x - p1.x)*vx + (p4.y - p1.y)*vy;
    t[3] = (p4.x - p2.x)*vx + (p4.y - p2.y)*vy;

}


/*************
  computes a segment image in local coordinates igloc 
  xp from a segment image in absolute coordinate igabs;

  only ->first,->last, ->vx,->vy and ->d  of the local 
  version are computed

  igloc is assumed allocated and initialized
  the list of segments in igloc is already 
  constructed and has to be of the same length 
  as the one of igabs
*************/


void tr_abs_to_loc(igla *igabs,igla *igloc,double *xp,char type)
{
    PDBL *la,*ll;
    Sgl *sa,*sl;
    double cs,sn,x,y,r;
  

    /*********** local system = steering ********/

    if(type == 'S'){
	cs = cos(xp[2]); 
	sn = sin(xp[2]);
    }
  
    /*********** local system = turret **********/

    else if(type == 'T'){
	cs = cos(xp[3]);
	sn = sin(xp[3]);
    }
    else {
      printf ("tr_abs_to_loc error : unknown type\n");
      return;
    }

    igloc->nsg = igabs->nsg;
    for(la = igabs->lsgl,ll = igloc->lsgl; la != NULL; la = dls(la)){
    
	sa = (Sgl *)dlel(la);
	sl = (Sgl *)dlel(ll);
	sl->flg = 0;
	sl->first.x = x = (sa->first.x - xp[0])*cs + (sa->first.y - xp[1])*sn;
	sl->first.y = y = -(sa->first.x - xp[0])*sn + (sa->first.y - xp[1])*cs;
	sl->first.r  = sqrt(x*x + y*y);
	
	
	

	sl->last.x = x = (sa->last.x - xp[0])*cs + (sa->last.y - xp[1])*sn;
	sl->last.y = y = -(sa->last.x - xp[0])*sn + (sa->last.y - xp[1])*cs;
	sl->last.r = r =  sqrt(x*x + y*y);
	

	sl->vx = sa->vx*cs + sa->vy*sn;
	sl->vy = -sa->vx*sn + sa->vy*cs;
	sl->d = -x*sl->vx - y*sl->vy;

	/* copying the variances */

	sl->vr = sa->vr; sl->vo = sa->vo; sl->vro = sa->vro;
 
	/*
	  making sure the new segment is going from right to left in
	  the new coord. sys.
	  */

	if (sl->d < 0)
	{
	    sl->last = sl->first;
	    sl->first.x = x;
	    sl->first.y = y;
	    sl->first.r = r;
	
	    sl->d = -sl->d;
	    sl->vx = -sl->vx;
	    sl->vy = -sl->vy;
	    /*indicating a reversed segment */
	    sl->flg = 1;
	}
	

	ll = dls(ll);
    
    }  
}

/*******
  this function tests if sloc is behind smap with a 
  gap more or equal than thr;
  the distance is understood as the orthogonal projection 
  of the farthest point of the part of sloc possibly 
  occulted by smap.

  sloc and smap are assumed expressed in the local coord. sys.
  from where sloc has been seen.

*******/

int test_behind(Sgl *sloc,Sgl *smap, double thr)
{

    static double t[4];
    static p2d pt;
    p2d *pl1,*pl2,*pm1,*pm2;
    static Sgl dt;
    double d;

    /* sin of 6 degrees */
    static double di = 0.1;


    pl1 = &(sloc->first); pl2 = &(sloc->last);
    pm1 = &(smap->first); pm2 = &(smap->last);

  

    /* origin -> sloc->first ^ origin -> smap->first */
    t[0] = (pl1->x*pm1->y - pl1->y*pm1->x);
    /* t[0] = sin(angle between locf and mapf) */
    t[0] = t[0]/(pl1->r * pm1->r);

    /* origin -> sloc->first ^ origin -> smap->last */
    t[1] = pl1->x*pm2->y - pl1->y*pm2->x;
    /* t[1] = sin(angle between locf and mapl) */
    t[1] = t[1]/(pl1->r  * pm2->r);

    /* origin -> sloc->last ^ origin -> smap->first */
    t[2] = pl2->x*pm1->y - pl2->y*pm1->x;
    /* t[1] = sin(angle between locl and mapf) */
    t[2] = t[2]/(pl2->r * pm1->r);

    /* origin -> sloc->last ^ origin -> smap->last */
    t[3] = pl2->x*pm2->y - pl2->y*pm2->x;
    /* t[3] = sin(angle between locl and mapl) */
    t[3] = t[3]/(pl2->r * pm2->r);


    if(t[0] >= di && t[2] <= -di ) {
    
	/* begining of smap between sloc->first and sloc->last */

	dt.vx = -pm1->y ;
	dt.vy = pm1->x;
	dt.d = 0.;
	intersg(&dt,sloc,&pt);
	d = -(pt.x - pm1->x)*smap->vx - (pt.y - pm1->y)*smap->vy;
	if(d > thr)
	    return(1);
    }
    
    else if(t[0] < -di && t[1] > di){
	/* begining of sloc between smap->first and smap->last */
	d = (pm1->x - pl1->x)*smap->vx + (pm1->y - pl1->y)*smap->vy;
	if(d > thr)
	    return(1);
    }
    else{
	/* no overlaping */
	return(0);
    }

    if( t[1] >= di  && t[3]  <= -di){
    
	/* end of smap between sloc->first and sloc->last */
	dt.vx = -pm2->y ;
	dt.vy = pm2->x;
	dt.d = 0.;
	intersg(&dt,sloc,&pt);
	d = -(pt.x - pm1->x)*smap->vx - (pt.y - pm1->y)*smap->vy;
	if(d > thr)
	    return(1);
    }

    else if (t[3]  > di && t[2] < -di){
    
	/* end of sloc between smap->first and smap->last */
	d = (pm1->x - pl2->x)*smap->vx + (pm1->y - pl2->y)*smap->vy;
	if(d > thr)
	    return(1);
    }

  


    return(0);

}


/***********
  returns the maximum overlap  
  (perpendicular projection of segment s1 on line s2)
***********/

double MaxOverlap(igla *ig,Sgl *s1,Sgl *s2)
     

{

    double vx,vy,d1,d4,d2,d3,dmax;

    Sg_pos(ig,s1,sp);
    p1 = s2->first;
    p2 = s2->last;

    vx = s2->vy;
    vy = -s2->vx;

    d1 = (sp[0] - p1.x)*vx + (sp[1] - p1.y)*vy;
  
    d2 = (sp[0] - p2.x)*vx + (sp[1] - p2.y)*vy ;
    
    d3 = (sp[2] - p1.x)*vx + (sp[3] - p1.y)*vy ;
 
    d4 = (sp[2] - p2.x)*vx + (sp[3] - p2.y)*vy;
  
    if(d1 >= 0 && d2 <= 0 && d3 >= 0 && d4 >=  0) 
    {
	return  -d2;
    }

    if(d1 >= 0 && d2 <= 0 && d3 >= 0 && d4 <=  0) 
    {
	return  d4 - d2;
    }
    
      
    if(d1 <= 0 && d2 <= 0 && d3 >=  0 && d4 >= 0)
    {
	dmax = d1 - d2;
	return(dmax);
    }

    if(d1 <= 0 && d3 >= 0 && d2 <= 0 && d4 <= 0)
      return d3;

    /* XXX No overlap (or head-to-foot segments) */
#ifdef VERBOSE
    printf ("No overlap or head-to-foot segments\n");
#endif
    return (-1.);


}



/***********
  returns 1 if the polygone is counterclockwise
  -1 otherwise
  particular case: less than 3 segments returns 0
***********/

long poly_orientation(igla *pl)
{
    PDBL *l;
    double x1,y1,x2,y2,x3,y3;
    Sgl *sg1;
    double ang = 0,ag1,ag2,ag_deb;

    l = pl->lsgl;

    if (Db_length(l) < 3)
	return(0);

    sg1 = (Sgl *)(l->el);

    x1 = sg1->first.x;
    y1 = sg1->first.y;
    x2 = sg1->last.x;
    y2 = sg1->last.y;
    ag1 = ag_deb = atan2(y2 - y1,x2-x1);

    do{
	l = dls(l); 
	sg1 = (Sgl *)l->el;
	x3 = sg1->last.x;
	y3 = sg1->last.y;
	ag2 = atan2(y3 - y2, x3 - x2);
	ang += norm_ang(ag2 - ag1,-M_PI);
	ag1 = ag2;
	x2 = x3; y2 = y3;
    }while(l->next != NULL);

    ang += norm_ang(ag_deb -ag1,-M_PI);

    if(ang > 0)
	return(1);
    else
	return(-1);

}













