
/*******
Smatch.c
********/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#ifdef VXWORKS
#define M_PI 3.141592653589793
#include <ioLib.h>
#endif

#ifndef VERBOSE
#define VERBOSE 0
#endif

#ifdef UNIX
# if VERBOSE>1
# define MEASURETIME
# endif
#endif

#ifdef MEASURETIME
#include <sys/time.h>
#endif


#define ANG_GROUP_HISTO   /* angular grouping based on histogram 
                                     (not on kalman) */

#include <dbliste.h>
#include <mat.h>
#include <las_prot.h>
#include <lasutil_p.h>
#include <kal_prot.h>
#include <image_mgr_p.h>
#include <definelocal.h>

#include "localize.h"
#include "match_p.h"
#include "match_private_p.h"
#include "map_unc.h"
#include "map_util.h"

static int Ang_match_histo(igla *igp,igla *igm);

/* ---------------------------------------------------------------------- */
/* PROTOTYPES OF LOCAL FUNCTIONS */
/* ---------------------------------------------------------------------- */

/* ---------------------------------------------------------------------- */

static double Thres = 12.; /* roughness of the matching in position 
			   = 95% for a 3 dof variable */

static double Thr = 6. ;  /* roughness for the angular matching 

			   = 98% for a 1 dof variable */
static double MaxAngle = 0.3; /* Max angle for groupe compatibility */
static double Thra = 0.5; /* sine of min angle between 2 segments for
			     creating an hypothesis;
			     default value corresponds to 30 degrees */
static  double MinLength = 0.1; /* square of the min length for 
			    a data segment to be taken into
			    account while generating hypothesis */
static  double MaxDistance = 22500.; /* square of the max distance for 
				a data segment to be taken into
				account while generating hypothesis
                                                 */
static Smatch *match; /* array containing all possible 
			 natchings between the map dand the data */
static long Nmct; /* global number of possible matchings */
static int Npercepts; /* Number of perceived segments */
static igla *Loc_map = NULL; /* empty image for local representation
				of the map */
static igla *Data; 
static PDBL *Phyp = NULL; /* list of angular groupings */
static PDBL *Lhypo = NULL; /*list of locations before
			    filtering */

/* prototype of added functions (by David Zhu) */
void SetMaxDistance(double maxd);

/**********
  approximate angle for local localization
**********/

static double MeanAngle;
static double GapAngle;

/*****
  to be set to 1 if 
  one wants to restrict the seach 
  to GapAngle around MeanAngle
*****/

static long LocAng = 0;

/************
  variables for memory management
**********/


static long Previous_number_matchings = 0;
static long Previous_map_nsg = 0;


static long Speed_level = 0;

/**
   Array with all prehyps (all compatible model and perceived segments)
**/
static char *Local_array = NULL;


                /**********************************
		  functions for setting parameters
		 **********************************/

/****
  setting the speed level; roughly it goes like this:
  level 0 : 
      slowest pace; at each step of the algorithm a probabilistic test
      is used as well as a kalman filtering for estimating the location
      of each hypothesis
  level 1 :
      the matching test is now heuristic only. The estimation are still 
      computed with kalman filtering all the way.
  level 2 : 
      only the initial position of each hypothesis is using probabilitic
      computation. Subsequent matches are only integrated ib the list of
      matches but don't contribute to ther location estimation.
      However, at the end, the feasible locations (and
      only them) will be plainly computed using the entire list of matches.
  level 3: the fastest. Doesn't use any probabilitic computation
      for initializing nor updating the locations, At the end we still
      compute plainly all feasible hypothesis such that the output is both
      more accurate and suitable for map building purposes.
****/ 
  
void SetSpeedLevel(int n)
{
  if(n < 0 || n > 3){
    printf("Error: Speed level must be between 0 and 3\n");
    printf("Level = %ld unchanged\n",Speed_level);
  }
  else
    Speed_level = n;
}

/****************
  Sets the mean angle and the possible deviation of the robot
  around it.

  angles  in degrees.

  0 <= gap <= Pi.

  If gap > 179 degrees, nothing is done and 
  the angle restriction is released.

*****/

void Set_loc_angle(double mean,double gap)
{
  
  GapAngle = gap*M_PI/180.;

  MeanAngle = mean*M_PI/180.;

  if(gap > 179.){
    LocAng = 0;
    return;
  }
  LocAng = 1;
}

/****
  Gets the mean angle (mean) and the possible deviation (gap)
  around it that have been defined with Set_loc_angle.

  Warning : angles in radians !!!!

  If gap > 179 degrees returns 0 (value with no meaning)
  else returns 1
***/

int Get_loc_angle(double *mean, double *gap)
{
  *mean = MeanAngle;
  *gap = GapAngle;
  return LocAng;
}

void Unset_loc_angle(void)
{
  LocAng = 0;
}

/*************
  setting of the minimum length a data segment 
  must have to be used for the localization
*************/

void SetMinLength(double minl)
{
  MinLength = minl*minl;
}

/*************
  setting of the max distance a data segment 
  must have to be used for the localization
*************/

void SetMaxDistance(double maxd)
{
  MaxDistance = maxd*maxd;
}


/******** 
  freeing and initializing the structures 
  Warning: the list Lhypo is not perfectly private in the sens
  that it goes back to the user who may 
  want to keep it; If he wants to clear the memory,
  the user will have to 
  call Free_list_hyp() before calling this function again.

  igp : local perception data
  map: map to be matched
  igm: sub-set of map for generating hypothesis( for example, local aspect
       or subdivision of the map into rooms...)

       igm may be equal to map. It leads to the most complex problem.

  Returns : the number of possible matchings (or 0).
***********/

int Init_Matching(igla *igp, igla *igm,igla *map)
{
  PDBL *l;


  /*****
    freeing the local version of the map 
  *****/


  if(Loc_map != NULL && map->nsg > Previous_map_nsg){
    free(Loc_map->usr);
    Imgl_free(Loc_map);
    Loc_map = NULL;
  }

  Previous_map_nsg = map->nsg;
  Data = igp;
  
  /****
    allocating  the new local version of the map
  ****/
  if(Loc_map == NULL)
    Loc_map = Imgl_create_copy(map);
  
  /**********
    ordering the segments of sgp:
    non-decreasing order of angle-reliability
  **********/

  angle_reliability_order(igp);
  
  /********
    discarding data segments less than
    MinLength  (inches) long
  ********/

  SelectionPercepts(igp);
  
  Lhypo = NULL;

  /* Freeing the pre-hypothesis list Phyp */

  if(Phyp != NULL){


    l = Phyp;

        /*** for each element of the list ****/

    do{

      /**** freeing the matching list*****/

      dlib_list(Prhel(l)->lsm);
      free(Prhel(l));
      l = dls(l);
    }while(l != Phyp);
    dlib_list(Phyp);
    Phyp = NULL;

  }


  
  /********* creating the new list of possible matchings *****/

  /* Returns the number of possible matchings */
  return Cree_list_match(igp,igm);


 
}


void CleanUp()

{
    PDBL *l;
    if(Loc_map){
      free(Loc_map->usr);
      Imgl_free(Loc_map);
    }
    Loc_map = NULL;
    Previous_map_nsg = 0;
    if(Phyp != NULL){


    l = Phyp;


        /*** for each element of the list ****/

    do{

      /**** freeing the matching list*****/

      dlib_list(Prhel(l)->lsm);
      free(Prhel(l));
      l = dls(l);
    }while(l != Phyp);
    dlib_list(Phyp);
    Phyp = NULL;

  }
 
}

/******* freeing a list of hypothesis *****/

void Free_list_hyp(PDBL *lhyp)
{
  PDBL *l;
  if(lhyp != NULL){
    for(l = lhyp; l != NULL; l = dls(l)){
      dlib_list(((Hyp *)dlel(l))->lsm);
      dlib_list(((Hyp *)dlel(l))->laux);
      free((Hyp *)dlel(l));
      
    }
  }

  dlib_list(lhyp);


}


/***************************
  pre-computes an array "match" of all 
  possible matchings between perception data 
  and the map.
  Only segments with ->flg == 0 will generate 
  possible matchings. This allow to use
  some kind of filtering in order to
  reduce the processing time.(see SelectionPercepts() );

  Each time this function is called,
  the array "match" is reallocated only if
  needed.

  Returns the number of possible matchings (or 0)
****************************/

static long Cree_list_match(igla *igp,igla *igm)

{
  
  PDBL *lp,*lm;
  Sgl *sp,*sm;
  Smatch *mct;

  if(igp == NULL || igm == NULL)
    return(0);

  
  if(igm->nsg * igp->nsg > Previous_number_matchings){
    free(match);
    Previous_number_matchings = igm->nsg * igp->nsg;
    match = (Smatch *)malloc(Previous_number_matchings * sizeof(Smatch));
  }
  mct = match;

  Nmct = 0;
  
  /*** loop over perceptual data */

  for( lp = igp->lsgl; lp != NULL; lp = dls(lp)){
    sp = Sgel(lp);
    
    /*** only segments with ->flg = 0 are taken into account ***/

    if(sp->flg == 0){

      /*** loop over map segments ***/

      for(lm = igm->lsgl; lm != NULL; lm = dls(lm)){
	sm = Sgel(lm);
	/*** only segments with ->flg = 0 are taken into account ***/

	if(sm->flg == 0){

	  Fill_match(mct,sp,sm);
	  mct++;
	  Nmct++;
	}
      }
    }
  }
  
  return(Nmct);
}     
    

/***********
Angular_group
Takes two images : igp and igm 
Computes a list of pre-hypothesis of matching between
igp and igm. All elements in a pre-hyp are compatible in angle
according to a Mahalanobis distance test.
It is  a partitionning : one matched pair  belongs to one 
pre-hyp.
The list is ordered  according to the angle of each pre-hypothesis.
Angles are between 0 and 2 Pi.
If flag LocAng==1 then groups which angle is out of 'MeanAngle +/- GapAngle'
interval are not considered (see Set_loc_angle and Unset_loc_angle).
************/


PDBL *Angular_group(igla *igp,igla *igm)
  
{
  PDBL *l, *lm;
  Prehyp *p;
  int n=0,nmatch;
  Smatch *match;
#ifdef MEASURETIME
  struct timeval tv;
   u_long msec1, msec2;
#endif
   FILE *pfile=NULL;

  /* Computes Phyp */
  Phyp = NULL;

#ifdef MEASURETIME
  gettimeofday(&tv, NULL);
  msec1 = (tv.tv_usec / 1000) + (tv.tv_sec * 1000);
#endif

#ifdef ANG_GROUP_HISTO
  if (!Ang_match_histo(igp, igm)) {
    printf ("Angular_group: Ang_match_histo failed\n");
    return NULL;
  }
#else
  if (!Ang_match(igp,igm)) {
    printf ("Angular_group: Ang_match failed\n");
    return NULL;
  }
#endif

#ifdef MEASURETIME
  gettimeofday(&tv, NULL);
  msec2 = (tv.tv_usec / 1000) + (tv.tv_sec * 1000);
  printf ("time %ld\n", msec2-msec1);
#endif  

#if VERBOSE>6
  pfile = fopen ("/home/sara/tmp", "w");
  if (pfile==NULL) 
    printf ("cannot open /home/sara/tmp\n");
#endif

  /* Give a number to each group (only for debug purpose) */
  l = Phyp;
  do {
    p = Prhel(l);
    p->numgp = n++;
    l = dls(l);

    nmatch = 0;
    for(lm = p->lsm; lm != NULL; lm = dls(lm)){
      match = Smel(lm);
      match->numgp = p->numgp;
      nmatch++;
      if (pfile)
	fprintf (pfile, "%f %d\n", match->eca, p->numgp);
    }
#if VERBOSE>2
    printf ("gpe %d : %d match\n", p->numgp, nmatch);
#endif
  } while(l != NULL && l != Phyp);
#if VERBOSE>0
  printf ("(%d gps)\n", n);
#endif
  if (pfile)
    fclose(pfile);

  return(Phyp);

}
  
/*********
Ang_match : see Angular_group
Returns 1 if OK, 0 otherwise
**********/
static long Ang_match(igla *igp,igla *igm)
    

{
  
  PDBL *hyp,*aux;
  Sgl *sp,*sm;
  Smatch *mct;
  long test;
  long i;
  Prehyp *phyp;

  if(igp == NULL || igm == NULL) {
    printf ("Ang_match : no data or no aspect map\n");
    return(0);
  }
  
  mct = match;

  for(i= 0; i < Nmct; i++){
    
    /* ignoring gaps more than GapAngle around MeanAngle when
       LocAng set to 1
       */
    
    if(LocAng)
      if(fabs(norm_ang(mct->eca - MeanAngle,-M_PI)) > GapAngle){
	mct++;
	continue;
      }
    
    sp = mct->sgp; sm = mct->sgm;
    
    hyp = Phyp;
    
    if(hyp == NULL){
      
      Add_Phyp(NULL,mct);
      
      /* we close the list */
      
      Phyp->next = Phyp;
      Phyp->last = Phyp;
      
    }
    
    
    else{
      do {

	/* testing if the angle defined by the matching mct is 
	   compatible (mahalobis angular distance) with a
	   pre-hypothesis pointed  by hyp or is to be put just after, 
	   just before (angle ordering).
	*/
  
	test = test_ang_mct(mct,hyp);
	
	
	if(test == 1) /* No compoatible pre-hyp in the list yet.
                         we create it and insert it before hyp
                      */
	  {
	    Add_Phyp(hyp,mct); 
	    break;
	  }
	
	else if (test == 3) /* mct is actually at the end of the list;
			       because the list is circular, it is equivalent to
			       insert mct before Phyp and keep Phyp as the 
			       begining of the list
			       */
	  
	  
	  {
	    aux = Phyp;
	    Add_Phyp(Phyp,mct);
	    Phyp = aux;
	    break;
	    
	  }
	
	
	/* Else  = we found a pre-hyp compatible.
                   mct has been added to the corresponding
                   pre-hypothesis in test_ang_mct()
	   */
	else if(test == 2)
	  break;
	hyp = dls(hyp);
      }while(hyp != NULL && hyp != Phyp);
      
    }
    
    mct++;
  }
  
  /****
    at this point, for each Prehyp hyp, hyp->lsm 
    is pointing the end of the list of matching.
    We come back to the begining
    ****/
  
  hyp = Phyp;
  if(hyp == NULL) {
    printf ("Ang_match : no hyp found\n");
    return(0);
  }
  do{
    phyp = (Prehyp *)dlel(hyp);
    phyp->lsm = dldeb(phyp->lsm);
    hyp = dls(hyp);
  }while(hyp != NULL && hyp != Phyp);
  return(1);
  
}

/****************
  computes all possible hypothesis 
  from the list of pre-hypothesis 
  lphyp

  Building of the list Lhypo

***************/

PDBL *Gen_Exp_Lhyp(PDBL *lphyp)
     

{
  PDBL *l;
  Prehyp *p;
  
  if(lphyp == NULL) {
    printf ("Gen_Exp_Lhyp : no initial hyp !\n");
    return(0);
  }
  l = lphyp;
#if VERBOSE>4
  printf("Thinking |");
  fflush(stdout);
#endif
  do{
    p = Prhel(l);
#if VERBOSE>7
    printf ("Gen_Exp_Lhyp gp %d\n", p->numgp);
#endif
    if(p->nmatched > 1)
      Gen_Expand(l);
    l = dls(l);
  } while(l != NULL && l != lphyp);
  
  return(Lhypo);
}


/*************************
  computes all possible hypothesis from 
  the pre-hypothesis in l (l is the list pointer
  on the actual pre-hypothesis)
*************************/
    
static void Gen_Expand(PDBL *l)
    
{
  
  PDBL *l1,*l2;
  Smatch *m1,*m2;
  Hyp *hyp=NULL;
#if VERBOSE>3
  long n = 0;
#endif
  Prehyp *phyp;
  int created = 0;
  int width;

  
  phyp = Prhel(l);

  set_array_prehyp(phyp);

  phyp->licomp = NULL;
  
  /* Loop over the matchings */

  for(l1 = phyp->lsm; l1 != NULL; l1 = dls(l1)){
    
    m1 = Smel(l1);
        
    for(l2 = dls(l1); l2 != NULL; l2 = dls(l2)){
      
      m2 = Smel(l2);
    
#if VERBOSE>3
      switch(n){
      case 0:
	printf("\b/");
	fflush(stdout);
	break;
      case 100:
	printf("\b-");
	fflush(stdout);
	break;
      case 200:
	printf("\b\\");
	fflush(stdout);
	break;
      case 300:
	printf("\b|");
	fflush(stdout);
	n = -1;
	break;
      default:
	break;
      }
      n++;
#endif

      if( m1->sgp != m2->sgp && m1->sgm != m2->sgm){

	  /******** m1 m2 non parallel (angle > arcSin(Thra)) *********/
    
	  if(fabs((m1->sgp)->vx*(m2->sgp)->vy - (m1->sgp)->vy*(m2->sgp)->vx) > Thra){

	    /***********
	       Test if m1 or m2 are already in a same 
	       hypothesis
	       Hypo_generated looks faster than dl_test_inter.
	    *************/

	    if(!Hypo_generated(m1,m2)){
	    
	      /* if(dl_test_inter(m1->Lhyp,m2->Lhyp) == 0 ){  */
	    
	      /****** creating an hypothesis ******/

	      if(!created){
		hyp = (Hyp *)malloc(sizeof(Hyp));
		hyp->lsm = NULL;
		hyp->laux = NULL;
		created = 1;
	      }

		/*******
		   Computes the robot position with this hypothesis
		********/

	      if(Speed_level < 3)
		Localize2(m1->sgp,m1->sgm,m2->sgp,m2->sgm,hyp->Tr,hyp->Vtr);
	      else
		Localize_np(m1->sgp,m1->sgm,m2->sgp,m2->sgm,hyp->Tr);

		


		/*******
		  adding hyp to the list of hypotheses 
		  generated by the pre-hypothesis phyp
		********/

		/* phyp->licomp = dlhcons(phyp->licomp,hyp); */ 
	      
		
			      

	      /* Expanding */
#ifdef ANG_GROUP_HISTO
	      width = 0;
#else 
	      width = 2;
#endif
	      HypExpand(m1,m2,hyp->Tr,hyp->Vtr,l,hyp, width);
	      
	      /******* adding hyp to head of Lhypo ******/
	      if(hyp->nmatched > 1){
		Lhypo = dlhcons(Lhypo,hyp);
		created = 0;
	      }
	      else
		{
		  

		}

	    }
	    
	  }
	  
	}
    }
  }

  if(created){
    dlib_list(hyp->lsm);
    dlib_list(hyp->laux);
    
    free(hyp);
  }

}

static long Hypo_generated(Smatch *m1,Smatch *m2)
{
  /* PDBL *lhyp1,*l;
  Hyp *h;

  lhyp1 = m2->Lhyp;

  for(l = lhyp1; l != NULL; l = dls(l))
    {
      h = (Hyp *)l->el;
      if(dlelm(h->lsm,m1) != NULL)
	return(1);
    }

    return(0);
   */

  if((long)(Smelt(Local_array,m1->n,m2->n)) == 1)
    return(1);

  return(0);
}
 

/**************** 
   takes a complete hypothesis defined by the 2 non-parallel matchings
   m1 and m2, producing the location tr with an accuracy vt.
   Then looks into the pre-hypothesis pointed to by phyp 
   if there are matchings compatible with the location tr and the accuracy vt.
   In case some matchings have been placed in the wrong pre-hypothesis
   because of unaccuracy on their angle value, we also look into 
   pre-hypothesis located around phyp plus or minus width
   Compatible matchings are added to the final hypothesis hyp.
****************/

static long useAutoVar = 0;


static void HypExpand(Smatch *m1,Smatch *m2,double *tr,double *vt,PDBL *phyp,Hyp *hyp,long width)
   
{
  long i;
  PDBL *lp,*ls;
  static double vtest[6];
  /* for debug test purposes */ 
  double *pVtest;

  if(!useAutoVar){
  /*** vtest will be used for testing the compatibility *****/
      mxs_copy(vt,vtest,3,3);
      pVtest = &vtest[0];
  }
  else
      /*uses pos variance and updates it along the way */
      pVtest = vt;


  Expand(m1,m2,tr,vt,vtest,Prhel(phyp)->lsm,hyp,1);
  
  if(hyp->nmatched < 2)
    return;
  
#if VERBOSE>8
  printf("  new hyp\n");
#endif /* VERBOSE>4 */

  /** filling the array for controling hypothesis generation **/

  Fill_array_prehyp(hyp->laux);

  lp = ls = phyp;

  /* we Expand also along the closest angular hypothesis */

  for(i = 0; i < width; i++){

    lp = dlp(lp);
    if(lp == ls)
      return;

#if VERBOSE>4
    printf ("HypExpand gpe %d (ang=%5.3f) tr=%5.3f utr=%5.3f\n", 
	    Prhel(phyp)->numgp, Prhel(phyp)->ang, tr[2], 3*sqrt(vt[5]));
    printf ("   -> gpe %d  ang=%5.3f dang=%5.3f dtr=%5.3f\n",
	    Prhel(lp)->numgp, Prhel(lp)->ang, 
	    fabs(norm_ang(Prhel(lp)->ang-Prhel(phyp)->ang,-M_PI)),
	    fabs(norm_ang(Prhel(lp)->ang-tr[2],-M_PI)));
#endif

    /* Test if the angle of the solution is near another angular group */
    if( fabs(norm_ang(Prhel(lp)->ang - tr[2],-M_PI)) < 0.1){
/*  Previously the test was with the angle OF THE GROUP of the solution */
/*   if( fabs(norm_ang(Prhel(lp)->ang - Prhel(phyp)->ang,-M_PI)) < 0.05){ */

#if VERBOSE>4
      printf ("HypExpand to gpe %d\n", Prhel(lp)->numgp); 
#endif /* VERBOSE>4 */
      Expand(m1,m2,tr,vt,vtest,Prhel(lp)->lsm,hyp,0);
    }
    
    ls = dls(ls);
    if (ls == lp) 
      return;

#if VERBOSE>4
    printf ("   -> gpe %d  ang=%5.3f dang=%5.3f dtr=%5.3f\n",
	    Prhel(ls)->numgp, Prhel(ls)->ang, 
	    fabs(norm_ang(Prhel(ls)->ang-Prhel(phyp)->ang,-M_PI)),
	    fabs(norm_ang(Prhel(ls)->ang-tr[2],-M_PI)));
#endif

    if( fabs(norm_ang(Prhel(ls)->ang - tr[2],-M_PI)) < 0.1) {
/*  Previously the test was with the angle OF THE GROUP of the solution */
/*  if( fabs(norm_ang(Prhel(ls)->ang - Prhel(phyp)->ang,-M_PI)) < 0.05) { */

#if VERBOSE>4
	printf ("HypExpand to gpe %d\n", Prhel(ls)->numgp);
#endif /* VERBOSE>4 */
	Expand(m1,m2,tr,vt,vtest,Prhel(ls)->lsm,hyp,0);
      }
  }

}



static void Expand(Smatch *m1,Smatch *m2,double *xp,double *vp,double *vtest,PDBL *lmatch,Hyp *hyp,long op)

{
	
  PDBL *l1;
  Sgl *st1,*st2;

  long t;

  Smatch *m;
  igla ig;
  
  /* To avoid that a same percepted could be used several times */
#define NB_MAX_ACQ_SEGS 200  
  static  int acqSegUsed[NB_MAX_ACQ_SEGS];
  int numSeg;

  ig.x = xp[0]; ig.y = xp[1]; ig.cs = cos(xp[2]); ig.sn = sin(xp[2]);
  ig.coosys = 'S';

  /* First expansion */

  if(op){
    hyp->nmatched = 0;
    
    

    if(Test_overlap(&ig,m1->sgp,m1->sgm) == 1){
      (hyp->nmatched)++;
  
    }
    else
      return; 

    if(Test_overlap(&ig,m2->sgp,m2->sgm) == 1){
      (hyp->nmatched)++; 
      
    }
    else
      return; 
  
  
  hyp->laux = dlhcons(NULL,m1);
  hyp->laux = dlhcons(hyp->laux,m2);
  hyp->lsm = dlhcons(NULL,m1);
  hyp->lsm = dlhcons(hyp->lsm,m2);


  /* To test if an acquired segment is not used several times */
  if (Npercepts > NB_MAX_ACQ_SEGS) {
    printf ("Expand ERROR: too many percepted segments (%d / %d)\n", 
	    Npercepts, NB_MAX_ACQ_SEGS);
    return; 
  }
  for(numSeg=0; numSeg<Npercepts; numSeg++)
    acqSegUsed[numSeg] = -1;
  acqSegUsed[m1->sgp->ifirst] = m1->sgm->ifirst;
  acqSegUsed[m2->sgp->ifirst] = m2->sgm->ifirst;

  }

  for(l1 = lmatch; l1 != NULL; l1 = dls(l1)){

    m  = Smel(l1);
    
    if(m != m1 && m != m2){
      
      st1 = m->sgp;
      st2 = m->sgm;

      /* Test if an acquired segment is not used several times */
      if (acqSegUsed[st1->ifirst] != -1) {
#if VERBOSE>8
	printf ("acq seg %ld already used (mod segs %d and %ld) gp %d op %ld\n", 
		st1->ifirst, acqSegUsed[st1->ifirst], st2->ifirst,
		m->numgp, op);
	acqSegUsed[st1->ifirst] = st2->ifirst;
#endif /* VERBOSE>4 */
	continue; 
      }

	/* here come all tests one can want to
	   be verified ; 

	   for now 
	   sg_test_match_fus tests:
	   -numerical compatibility
	   -overlaping of the perceptual and the map segments
	   
	   sg_test+match_fus returns : 0 if no numerical compatibility
	                               1 if overlaping

	   Overlaping is not tested if numerical compatibility
	   is not satisfied

	   */


      if((t = sg_test_match_fus2(&ig,xp,vp,vtest,st1,st2,Thres,Speed_level)) != 0){

	/* numerical compatibility */

	if(op == 1){

	  /* not to generate an hyp from m and any of this set 
	     of matchings */
	  hyp->laux = dlhcons(hyp->laux,m);
	}
	

	  /* overlaping */

	if(t == 1 ){

	  /* including m in hyp */
	  
	  hyp->lsm = dlhcons(hyp->lsm,m);
	
	  (hyp->nmatched)++;


	  /* Record that the segment is used */
	  acqSegUsed[st1->ifirst] = st2->ifirst;

	}
      }
    }
  }
}






long compatible_hyp(Hyp *hyp1,Hyp *hyp2)
{
  
  if(Dist_maha(idta3,didta3z,didta3x,hyp1->Tr,hyp1->Vtr,
	       hyp2->Tr,hyp2->Vtr,NULL,NULL,NULL,3,3,3) < 40.)
    return(1);
  return(0);
}

void fusion_hypo(Hyp *hyp1,Hyp *hyp2)
{
  PDBL *l1;
  Smatch *m;

  /*****
    loop over hyp2 matchings 
  ******/

  for (l1 = hyp2->lsm; l1 != NULL; l1 = dls(l1)){
    /**** 
      if the matching pointed to by l1 doesn't belong to
      hyp1 matchings
    ****/

    if(dlelm(hyp1->lsm,dlel(l1)) == NULL){
      /****
	integrating the matching in hyp1
      *****/

      m = (Smatch *)dlel(l1);

      Update2_pos(hyp1->Tr,hyp1->Vtr,m->sgp,m->sgm);
      
      hyp1->lsm = dlhcons(hyp1->lsm,m);

      hyp1->nmatched ++;

    }
  }
}




/*  
    test_validity_hypos : filters "ListHypo" hypotheses :

    - Filters any solution better than or equal to "thres" matchings 

    - Eliminates any solution leading the robot to be
    outside the boundary "bound", if any.
   
    - If op == 1
         eliminates solutions leading to an object ig1 to be seen behind
	 a static obstacle of ig2. "behind" means "further than distmax
	 behind any static obstacle"
	      
  */


PDBL *test_validity_hypos(igla *ig1,igla *ig2,igla *bound,
			 PDBL *phyp,long op,long thres,double distmax)
{
  PDBL *l,*list_valid = NULL;
  Hyp *hyp;

  if(phyp == NULL)
    return(NULL);

  l = phyp;
  do{
    
    hyp = (Hyp *)dlel(l);
    if(hyp->nmatched >= thres){

      
      /***
	testing if the solution is within
	the boundary "bound"
      ***/
      if(!InsidePoly(hyp->Tr,bound))
	 {
	   l = dls(l);
	   continue;
	 }

      /* XXXX calcul inutile ? */
      /* XXX IS IT REALLY USEFULL IF op == 0  ??!!!
	 XXX Loc_map seems to be usefull only for hypo_is_valid test.
	 XXX (Moreover, for big maps it seems costly to project all the
	 XXX map for each hypo. Isn'it possible to do it another way ?) */

      /* Compute Map relatively to robot pos hypothesis */
      tr_abs_to_loc(ig2,Loc_map,hyp->Tr,'S');

      /* Changes the origine of Current_data (does not change the segments) */
      Set_pos_image(ig1,hyp->Tr[0],hyp->Tr[1],hyp->Tr[2],hyp->Tr[2]);

      if(op){
	if(hypo_is_valid(ig1,Loc_map,distmax)){
	  list_valid = dlhcons(list_valid,hyp);
	}
      }
      else
      {
	list_valid = dlhcons(list_valid,hyp);
      }

     }  
    l = dls(l);
  }while(l != NULL && l != phyp);
  return(list_valid);
}

/*********
hypo_is_valid : if a segment of igp is behind a segment of igm 
                (with a distance of at least thr) returns 0 
                Else returns 1.
 *********/
long hypo_is_valid(igla *igp,igla *igm,double thr)
  {
    PDBL *lp,*lm;
    Sgl *sp,*sm;

    for(lp = igp->lsgl; lp != NULL; lp = dls(lp)){
      sp = (Sgl *)dlel(lp);
      for(lm = igm->lsgl; lm != NULL; lm = dls(lm)){
	sm = (Sgl *)dlel(lm);
	if (test_behind(sp,sm,thr))
	  return(0);
      }
    }
    return(1);
  }


#ifdef UNUSED

/*************
  returns 1 if there exist 2 pairs of
  matchings in the hypothesis hyp 
  defining an angle the sinus of
  which is more than the threshold
  Thra defined in the header of
  this file.
  
  This condition is understood as a 
  a necessary condition for the hypothesis
  to be well defined. 

  returns 0 otherwise
**************/


static long Well_defined_hypo(Hyp *hyp)
{
  double min_vx = 2.,max_vx = -1.,vx;
  double vx_min=1000,vy_min=1000,vx_max=-1,vy_max=-1;
  PDBL *l;
  Sgl *s;
  Smatch *m;
  

  for(l = hyp->lsm; l != NULL; l = dls(l))
    {
      
      m  = (Smatch  *)dlel(l);
      s = m->sgm;
      if((vx = fabs(s->vx)) > max_vx){
	max_vx = vx;
	vx_max = s->vx;
	vy_max = s->vy;
      }
      
      if (vx < min_vx){
	
	min_vx = vx;
	vx_min = s->vx;
	vy_min = s->vy;
      }
      
    }
  if(fabs(vx_min * vy_max - vy_min * vx_max)  < Thra)
    return(0);

  return(1);

}

#endif /* UNUSED */

/** 
  apply ComputeFinalLocation to the list of
  hypothesis lhyp 
**/

void ComputeFinalListHypo(PDBL *lhyp)
{
  PDBL *l;

   if(Speed_level < 2) /** all this has already been done **/
    return;

  for(l = lhyp; l != NULL; l = dls(l)){
    ComputeFinalLocation((Hyp *)(l->el));
  }

}




/**** 
  In case the speed level is high we didn't compute
  very accurately the location corresponding to 
  each hypothesis. The variances are not
  computed yet either. However, in each hypothesis
  structure, there is a list of matchings (hyp->lsm)
  from which this can be done. This is the purpose
  of this function.
****/

static void ComputeFinalLocation(Hyp *hyp)
{
  Smatch *m1,*m2;
  PDBL *l,*l_end;
  Sgl *sg1,*sg2;
  Sgl *sm1,*sm2;

 
    
  /** the two initial pairs of segments are at the end 
      of the list **/

  l_end = dlfin(hyp->lsm);

  l = dlp(l_end);

  m1 = (Smatch *)(l_end->el);
  m2 = (Smatch *)(l->el);

  sm1 = m1->sgm;
  sg1 = m1->sgp;

  sm2 = m2->sgm;
  sg2 = m2->sgp;


  /** computing the initial location if Speed_level = 3 **/
  if(Speed_level == 3)
    Localize2(sg1,sm1,sg2,sm2,hyp->Tr,hyp->Vtr);

  l_end = l;

  for(l = l_end->last; l != NULL; l = dlp(l)){
    m1 = (Smatch *)(l->el);
    Update2_pos(hyp->Tr,hyp->Vtr,m1->sgp,m1->sgm);
  }

}
  
void Expurge_multiplicity(PDBL *lhyp)
{
  
  PDBL *l1,*l2;
  Hyp *hyp1,*hyp2;


  l1 = lhyp;
  if(l1 == NULL)
    return;

  for(l1 = lhyp; dls(l1) != NULL;l1 = dls(l1)){
    hyp1 = (Hyp *)dlel(l1);
    for(l2 = dls(l1); l2 != NULL; ){
      hyp2 = (Hyp *)dlel(l2);
      if(compatible_hyp(hyp1,hyp2)){
	fusion_hypo(hyp1,hyp2);
	l2 = dldel_s(l2,l2);
	
      }
      else
	l2 = dls(l2);
    }
    if(dls(l1) == NULL)
      return;
  }

}



#ifdef UNUSED

static Hyp *choose_best_hypo(PDBL *lhyp)
{
  PDBL *l;
  long score = 0;
  Hyp *hypc,*hypr=NULL;

  if(lhyp == NULL)
    return(NULL);

  for(l = lhyp; l != NULL; l = dls(l)){
    hypc = (Hyp *)dlel(l);
    if(hypc->nmatched > score){
      score = hypc->nmatched;
      hypr = hypc;
    }
  }

  return(hypr);
}
#endif

/**********************
  functions for setting parameters
***********************/



/***********************************************
  helper functions for classifying, creating etc...
***********************************************/

/***** 
testing if the angle defined by the matching mct is 
to be put just after, just before or is compatible with
the pre-hypothesis pointed to by l
*****/
       
static long test_ang_mct(Smatch *mct,PDBL *l)
     

{
  
  Prehyp *m1,*m2;
  
  
  m1 = Prhel(l);
  m2 = Prhel(l->last);
  
  /* Warning : the angle should be between 0 and 2 Pi */
  
  
  
  if(m1->ang > mct->eca)
    return(test_between(m1,m2,mct));

  else if(l->next == Phyp){ /* we have to test the compatibility
			      with m1 and the next one
			      
			      */
    m2 = Prhel(l->next);
    if(test_between(m1,m2,mct) == 1)
      return(3); /* no compatibility */
    else
      return(2);
  }
  return(0);/* no compatibility */
}

	 
  
static long test_between(Prehyp *m1,Prehyp *m2,Smatch *mct)
    
{
  
  double a,va,vb;
  double d1,d2;
  
  /* We assume that the closest hypothesis (Dist_maha)  from mct is one 
     of those on its side in the list : either m1 or m2 */
  
  
  
  va = (mct->sgp)->vo + (mct->sgm)->vo;
  a = norm_ang(mct->eca,m1->ang - M_PI);
  d1 = Dist_maha(idt1,didty1,didtx1,&a,&va,&(m1->ang),&(m1->vang),NULL,&a,&va,1,1,1);
  /* if the gap is too big, we decide for imcompatibility in any case */

  if (fabs(a - m1->ang) > MaxAngle)
    d1 += Thr;

  a = norm_ang(mct->eca,m2->ang - M_PI);

  d2 = Dist_maha(idt1,didty1,didtx1,&a,&va,&(m2->ang),&(m2->vang),NULL,&a,&va,1,1,1);
  
  if (fabs(a - m2->ang) > MaxAngle)
      d2 += Thr;
  
  if(d1 < d2){
    if (d1 < Thr) /* m1 is compatible and closer than m2 */
      {
	/* we include mct into m1 */
	a = norm_ang(mct->eca,m1->ang - M_PI);
	fusion(idt1,didty1,didtx1,&a,&va,&(m1->ang),&(m1->vang),NULL,&(m1->ang),&vb,1,1,1);
	if(va < m1->vang)
	  m1->vang = va;
	/******
	  we conserve the order 
	*******/

	m1->lsm = dlbcons(m1->lsm,mct);
	(m1->nmatched)++ ;
	return(2);
	

      }
    
    else 
      /* no compatible hypothesis, but the new one 
	 will have to be between m1 and m2*/
      return(1);
    
  }
  
  else {
    if (d2 < Thr) /* m2 is compatible and closer than m1 */
      {
	a = norm_ang(mct->eca,m2->ang - M_PI);
	fusion(idt1,didty1,didtx1,&a,&va,&(m2->ang),&(m2->vang),NULL,&(m2->ang),&vb,1,1,1);
	if(va < m2->vang)
	  m2->vang = va;
	m2->lsm = dlbcons(m2->lsm,mct);
	(m2->nmatched)++ ;
	return(2);
	    
	  }
    
    else 
      /* no compatible hypothesis, but the new one 
	 will have to be between m1 and m2*/
      return(1);
    
  }

  /* no compatible hypothesis, but the new one 
     will have to be between m1 and m2*/
  return(1);
}

			    

static void Fill_match(Smatch *mct,Sgl *sp,Sgl *sm)
    
{
  
  mct->Lhyp = NULL;
  mct->sgp = sp;
  mct->sgm = sm;
  /* angle between 0 and 2 Pi */
  mct->eca = norm_ang(sm->o - sp->o,0.);
  
  
}
  
static PDBL * Add_Phyp(PDBL *hyp,Smatch *mct)
 
{
  Prehyp *php;
  

  php = (Prehyp *)malloc(sizeof(Prehyp));
  php->lsm = dlhcons(NULL,mct);
  php->licomp = NULL;
  php->ang = mct->eca;
  php->vang = (mct->sgp)->vo + (mct->sgm)->vo;
  php->nmatched = 1;

  if(hyp == Phyp)
    return ((Phyp = dlhcons(hyp,php)));
  else
    return dlhcons(hyp,php);

}

/**************
selecting two "good" segments (long and with a good angle)
for localizing
**************/


       



static long SelectionPercepts(igla *igp)
    
{
  PDBL *l;
  Sgl * s;
  double length2;
  int i = 0;
  
  


  if(igp == NULL)
    return(-1);

  /* eliminating segments less that MinLength inches long */

  for(l = igp->lsgl; l != NULL; l  = dls(l)){
    s  = (Sgl *)dlel(l);
    length2 = (s->first.x - s->last.x)*(s->first.x - s->last.x)
             +(s->first.y - s->last.y)*(s->first.y - s->last.y);

    if(length2 < MinLength)
      s->flg = 1;
    else
      s->flg = 0;
    s->ifirst = i++;
  }
  Npercepts = i;
  return(0);
}



/*********
  we order the list of segment of igp in non-decreasing
  unnacuracy in angle
*********/
  
static void angle_reliability_order(igla *igp)
{
  
  PDBL *li = NULL,*l,*la;

  
  la = igp->lsgl;
  igp->lsgl = NULL;
  
  for(l = la; l != NULL;){
    li = l;
    l = dls(li);
    igp->lsgl = insere_dbl(igp->lsgl,li,test_unacuracy);
  }
  
}




static long test_unacuracy(Sgl *s1,Sgl *s2)
  {
    if (s1->vo < s2->vo)
      return(1);
    return(0);
  }


/*** 
  matching array manager for ensuring not generating too
  much hypothesis
***/

static void set_array_prehyp(Prehyp *phyp)
{
  int size,n=0;
  PDBL *l;
  static int  Gsize = 0;
  Smatch *m;

  size = phyp->nmatched;

  if(size > Gsize){
    free(Local_array);
    Local_array = (char *)malloc(sizeof(char) * (size * (size+1)/2));
    Gsize = size;
  }

  memset((void *)Local_array,0,(size * (size+1)/2));
  
  for(l = phyp->lsm; l != NULL; l = dls(l)){
    m = (Smatch *)(l->el);
    m->n = n;
    n++;
  }

}


static void Fill_array_prehyp(PDBL *lsm)
{
  PDBL *l1,*l2;
  Smatch *m1,*m2;
  int n;

  for(l1 = lsm; l1->next != NULL; l1 = dls(l1)){
    m1 = (Smatch *)l1->el;
    n = m1->n;
    for(l2 = l1->next; l2 != NULL; l2 = dls(l2)){
      m2 = (Smatch *)l2->el;

      /* sara 3/11/2003 bug : left part not modifiable ! */
/*       Smelt(Local_array,n,m2->n) = 1; */
      Local_array[Smelt_ind(n,m2->n)] = 1; 
    }
  }

}
  

static double angleLimit360(double angle);

#ifndef RAD_TO_DEG
#define RAD_TO_DEG(x) (x*180.0/M_PI)
#endif

#ifndef MIN
#define MIN(x,y) (y<x?y:x)
#endif

/* #define MIN_SEP_WIDTH    1 */  /* sep */
/* #define NB_MIN_GROUPS   10  */ /* Nb min of groups (depends also on number of matching) */

/* list of the group separator angles */
#define NB_STEPS      4*360  /* Nb steps in distribution (discretisation=1deg) */
static int separators[NB_STEPS]; 
static int nbGroups;


static int angleHisto()
{
  Smatch *mct;  /* one matching */
  unsigned char ang_histo[NB_STEPS];  /* distribution of the matching angles */
  unsigned char min_histo[NB_STEPS];  /* the minimums of the distribution 
				       (0: NOT A MINIMUM
				        1: 0 matching with this angle
				        2: 1 matching with this angle
					3: 2 matching with this angle ...) */
  int i, ang, prevAng, nextAng, gpe, angDeg;
  int nbMaxMatchInSep;
  int minMatchings;
  int nbMinSep;
  int sepWidth;
  int vicinity;
  int prevNbSep;
  int nbSeps;
  int reachEnd;
  FILE *pfile;
  int maxMatch, maxMaxMatch;
  int MIN_SEP_WIDTH;
  int minM, maxW, minW, maxM;
  int NB_MIN_GROUPS;
  int minGps, maxGps;

  minM = 20;
  maxW = 16;  /* 4 degres */
  minGps = 1;

  maxM = 200;
  minW = 1;    /* 1/4 de degre */
  maxGps = 100;

  /* Parameters to create separators */

  /* width min */
  if (Nmct<minM)
    MIN_SEP_WIDTH = maxW;
  else if (Nmct>maxM)
    MIN_SEP_WIDTH = minW;
  else {
    MIN_SEP_WIDTH = (int)((double)(maxW-minW)/(double)(minM-maxM)*(double)(Nmct-maxM)+minW);
  }
#if VERBOSE>0
  printf ("MIN_SEP_WIDTH (nb match %d)= %d (%f deg)\n", 
	  Nmct, MIN_SEP_WIDTH, MIN_SEP_WIDTH/4.0);
#endif

  /* nb min gpes */
  if (Nmct<minM)
    NB_MIN_GROUPS = minGps;
  else if (Nmct>maxM)
    NB_MIN_GROUPS = maxGps;
  else {
    NB_MIN_GROUPS = (int)((double)(maxGps-minGps)/(double)(maxM-minM)*(double)(Nmct-minM)+minGps);
  }
#if VERBOSE>0
  printf ("NB_MIN_GROUPS (nb match %d)= %d\n", Nmct, NB_MIN_GROUPS);
#endif

  nbMinSep = NB_MIN_GROUPS;

  /* Init distribution */
  for (ang=0; ang<NB_STEPS; ang++) {
    ang_histo[ang] = 0;
    min_histo[ang] = 0;
  }

  /**
   **  Computes the distribution of the matching angles 
   **/
  mct = match;
  for(i=0; i<Nmct; i++) {
    angDeg = (int)angleLimit360(RAD_TO_DEG(mct->eca));
    (ang_histo[angDeg])++;
#if VERBOSE>8
    printf ("%d : %f -> %d \n", i, mct->eca, angDeg);
#endif
    mct++;
  }


#if VERBOSE>5
  /* Pour tracer l'histogramme avec gnuplot */
  pfile = fopen ("/home/sara/tmp2", "w");
  if (pfile==NULL) 
    printf ("cannot open /home/sara/tmp2\n");

  for (ang=0; ang<NB_STEPS; ang++) {
#if VERBOSE>8
    if (ang_histo[ang]) {
      printf ("ang%d : %d matchings\n", ang, ang_histo[ang]);
    }
#endif

    /* Pour tracer l'histogramme avec gnuplot */
    fprintf (pfile, "%d %d\n", ang, ang_histo[ang]);
  }
  fclose (pfile);
#endif

  /**
   **  Gets the minimums 'min_histo' in the angles distribution
   **/

  /* Minimum number of separators (ie, of groups) */
/*   nbMinSep = MIN(NB_MIN_GROUPS, (int)(Nmct/8.0)); */
/*   if(nbMinSep==0) nbMinSep=1; */

  /* Look for group separators : the minimums of the histogram */
  nbMaxMatchInSep = 0;
  nbGroups = 0;
  prevNbSep = -1;
  nbSeps = 0;

  /* While not enought separators (and if nbGroups still increase) */
#if VERBOSE>5
  printf ("nbMinGpe = %d \n", nbMinSep);
#endif
  while (nbGroups < nbMinSep && prevNbSep < nbGroups) {

    /* Look for minimum in angles distribution */
    prevNbSep = nbGroups;

    for (ang=0; ang<NB_STEPS; ang++) {
      
      /* not already a separator, but could be */
      if (!min_histo[ang] && ang_histo[ang] <= nbMaxMatchInSep) {

	/* record the separator (with its 'level') */
	nbSeps++;
	min_histo[ang] = nbMaxMatchInSep+1;

	/* get previous and next angle group */
	if (ang == 0) {
	  prevAng = NB_STEPS-1;
	  nextAng = ang + 1;
	}
	else {
	  prevAng = ang - 1;
	  if (ang == NB_STEPS-1) nextAng = 0;
	  else nextAng = ang + 1;
	}

	/* update the number of separators (that are not adjacent) */
	if (!min_histo[prevAng]) {
	  nbGroups++;
#if VERBOSE>6
	    printf ("nbGroups ++\n");
#endif
	}

	/* both adjacent groups are already separators : merging */
	else 
	  if (min_histo[nextAng]) {
#if VERBOSE>6
	    printf ("nbGroups --\n");
#endif
	    nbGroups--;
	  }
      }
    }
    nbMaxMatchInSep++;
  }

#if VERBOSE>1
  printf ("%d max match in sep\n", nbMaxMatchInSep);
  printf ("step 1 : %d separators\n", nbGroups);
#endif

  /* Degenerated cases :
     No more groups (only separators) 
     or the number of groups has decreased */
  if (nbSeps == NB_STEPS || nbGroups < prevNbSep) {

    int prevSep = min_histo[NB_STEPS-1];

#if VERBOSE>1
    if (nbSeps == NB_STEPS)
      printf ("no more group !!\n");
    else
      printf ("nb groups decreased %d -> %d !\n", prevNbSep, nbGroups);
#endif

    /* Then the separators with nbMaxMatchInSep are no more separators */
    nbMaxMatchInSep--;
    for (ang=0; ang<NB_STEPS; ang++) {
      if (min_histo[ang] == nbMaxMatchInSep) {

	/* no more a separator */
	min_histo[ang] = 0;

	/* if previous and next are separators
	   then starting a new gpe */
	nextAng = ang+1;
	if (nextAng == NB_STEPS) nextAng = 0;
	if (min_histo[nextAng] == nbMaxMatchInSep) {
	  min_histo[nextAng] = 0;
	}
	if (prevSep && min_histo[nextAng] != 0) {
	  nbGroups++;
	}
      }
      prevSep = min_histo[ang];
    } /* for */

#if VERBOSE>1
    printf ("%d max match in sep\n", nbMaxMatchInSep);
    printf ("step 1 bis : %d separators\n", nbGroups);
#endif
  }

  /**
   **  Gather minimum as group separators
   **/

  /* Gathered the separators in one unique array */
  nbGroups = 0;
  reachEnd = 0;
  separators[0] = 0;
  separators[1] = 0;
  for (ang=0; ang<NB_STEPS; ang++) {

    /* this angle belongs to a group separator */
    if ((minMatchings = min_histo[ang])) {

      /* Init angular width  of the separator */
      sepWidth = 0;
      separators[nbGroups] = ang;
#if VERBOSE>6
      printf ("init separators[%d] = %d\n", nbGroups, separators[nbGroups]);
#endif

      /* look for the minimum separator in the vicinity 
	 and the width of the separator */
      while (1) {
	
	/*   */
	sepWidth++;
	ang++;
	if (ang == NB_STEPS) {
	  ang = 0;
	  /* Already a loop !!!! */
	  if (reachEnd) {
	    printf ("loop\n");
	    break;
	  }
	  reachEnd = 1;
	}

	/* no more adjacent separator */
	if (!(vicinity = min_histo[ang]))
	  break;

	/* it is a better separator */
	if (vicinity < minMatchings) {
	  minMatchings = vicinity;
	  separators[nbGroups] = ang;
#if VERBOSE>6
	  printf ("new separators[%d] = %d\n", nbGroups, separators[nbGroups]);
#endif
	}
      }

      /* if width is enough then ready for next separator */
      if (sepWidth >= MIN_SEP_WIDTH) {
	/* Test if the last group has reached the first one */
	if (reachEnd && ang >= separators[0]) {
#if VERBOSE>2
	  printf ("first and last groups are fused\n");
#endif
	}
	/* it is a new group */
	else
	  nbGroups++;
      }
#if VERBOSE>6
      else {
	printf (" REMOVE separator = %d (width %d)\n", nbGroups, sepWidth);
      }
      printf (" nbGroups = %d\n", nbGroups);
#endif
    }

    if(reachEnd)
      break;
  } /* for */

#if VERBOSE>6
  for (gpe=0; gpe<nbGroups; gpe++) {
    printf ("separators[%d] = %d\n", gpe, separators[gpe]);
  }
#endif
#if VERBOSE>1
  printf ("step 2 : %d separators\n", nbGroups);
#endif
#if VERBOSE>2
  printf ("%d separators (degre) : ", nbGroups);
  for (gpe=0; gpe<nbGroups; gpe++) {
    printf ("%d ", separators[gpe]/4);
  }
  printf ("\n");
#endif
  return nbGroups;
}

/* ----------------------------------------------------------------------
 * Ang_match_histo  -  
 *
 * Computes the partitioning of all the matchings pair in angular groups
 * by analysing the angles distribution (instead of considering the angular
 * compatibility in the sens of kalman, the result of this approach was too
 * instable and dependent to the order in which we consider the matchings. 
 * Very near matching were sometimes put in distinct groups because of their 
 * uncertainty.). This new method looks for the local minimum in the
 * the distribution (histogram) of the angle for every matching. These 
 * minimums are the boundaries between the angle groups : closed angles 
 * remain in a same group, independently of their uncertainty.
 *
 */

static int Ang_match_histo(igla *igp,igla *igm)
{
  Smatch *mct;
  int angDeg, i;
  int group, nextGroup;
  static PDBL *hyp_ang[NB_STEPS];
  Prehyp *m1, *phyp;
  PDBL *hypAng, *hyp, *aux;

  /* Computes histogram of angles */
  angleHisto();

  for (i=0; i<nbGroups; i++)
    hyp_ang[i] = NULL;

  /* test useless inputs for historical reasons ! */
  if(igp == NULL || igm == NULL) {
    printf ("Ang_match : no data or no aspect map\n");
    return(0);
  }
  
  /* for each matching */
  mct = match;
  for(i=0; i<Nmct; i++, mct++) {
    
    /* Ignoring gaps more than GapAngle around MeanAngle if LocAng set to 1 
       !!!!!!! */
    if(LocAng) {
      if(fabs(norm_ang(mct->eca - MeanAngle,-M_PI)) > GapAngle){
	continue;
      }
    }
    
    /* Look for angular group */
    angDeg = (int) angleLimit360(RAD_TO_DEG(mct->eca));
    group = 0;
    while (group < nbGroups && angDeg > separators[group]) group++;
    if (group == nbGroups) group = 0;
#if VERBOSE>6
    printf ("(nbGroups %d) ang %f = %d -> group %d\n", 
	    nbGroups, mct->eca, angDeg, group);
#endif

    /* Head of hyp list */
    hyp = Phyp;

    /* Init hyp list with first matching */
    if (hyp == NULL) {
      
      /* first hyp */
      Add_Phyp(NULL,mct);
      
      /* we close the list */
      Phyp->next = Phyp;
      Phyp->last = Phyp;

      /* record this hyp */
      hyp_ang[group] = Phyp;
#if VERBOSE>6
      printf ("new new group %d\n", group);
#endif

      continue;
    }
    
    /* Hyp list already initialised */

    /* Already this hyp : just add matching */
    if ((hypAng=hyp_ang[group])) {
      m1 = Prhel(hypAng);
      m1->lsm = dlbcons(m1->lsm,mct);
      (m1->nmatched)++ ;
#if VERBOSE>6
      printf ("add hyp %d in group %d\n", m1->nmatched, group);
#endif
      continue;
    }

    /* not yet hyp for this angle : add one */
    
    /* look for next hyp (angular ordering) */
    nextGroup = group+1;
    while(nextGroup<nbGroups && hyp_ang[nextGroup] == 0) 
      nextGroup++;
	
    /* A next hyp has been found : insert just before it */
    if (nextGroup != nbGroups) {
      hyp_ang[group] = Add_Phyp(hyp_ang[nextGroup], mct);
#if VERBOSE>6
      printf (">> insert new group %d (before %d)\n", group, nextGroup);
#endif
      continue;
    }

    /* mct is actually at the end (or head) of the (circular) list */
    aux = Phyp;
    Add_Phyp(Phyp,mct);
    hyp_ang[group] = Phyp;
    Phyp = aux;
#if VERBOSE>6
    printf (">> new group %d at head\n", group);
#endif
    continue;

  } /* for each matching */


  /****
    at this point, for each Prehyp hyp, hyp->lsm 
    is pointing the end of the list of matching.
    We come back to the begining
    ****/
  
  hyp = Phyp;
  if(hyp == NULL) {
    printf ("Ang_match : no hyp found\n");
    return(0);
  }
  do{
    phyp = (Prehyp *)dlel(hyp);
    phyp->lsm = dldeb(phyp->lsm);
    hyp = dls(hyp);
  }while(hyp != NULL && hyp != Phyp);
  return(1);

}


static double angleLimit360(double angleIn)
{
  double angle = angleIn * NB_STEPS/360;

  while(angle >= NB_STEPS)
    angle -= NB_STEPS;
  while (angle < 0)
    angle += NB_STEPS;
  return angle;
}

