/* server_skel.c
 *
 * example of how to write an rpc server for cognos
 *
 *                             ** COPYRIGHT NOTICE **
 *
 * Copyright (c) 1998, Nomadic Technologies, Inc.
 *
 * The contents of this file is copyrighted by Nomadic Communications, Inc, and
 * is protected by United States copyright laws.  Unauthorized copying is
 * prohibited.  All rights reserved.
 *
 */

#include <assert.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef VXWORKS
#include <sys/time.h>
#endif
#include <sys/types.h>
#include <unistd.h>
#include <math.h>
#ifdef VXWORKS 
#include <taskLib.h>
#define M_PI 3.141592653589793
#endif

#include "constants.h"
#include "buffer.h"
#include "command.h"
#include "socket.h"
#include "lasdef.h"
#include "User.h"
#include "odometer.h"
#include "connect.h"
#include "hash.h"
#include "worldcomm.h"
#include "map_server.h"

#include "odometer_p.h"
#include "sick_robots_p.h"
#include "Map_mgr_p.h"
#include "las_io_p.h"
#include "image_mgr_p.h"
#include "session_p.h"
#include "robots_p.h"
#include "lasegm_p.h"

#include "match_p.h"
#include "api_P.h"
#include "config_p.h"

/*include the definition of the comms double resolution
  to be included and used by any client communicating directly (client_skel.c)
 */
#include "com_common.h"

#ifdef VXWORKS 
#undef LOCAL
#undef OK
#endif

/* the  robot this server is connected to as a client */
static int RobId = -1;

/* Robot machine name */

static char RobotName[50];
/* Robot Port number*/ 

static long RobPort;

/* to access internal laser image structure */
/* igla *PtoIm(int); */

/* strings under which commands are registered */
/** see RegisteredCmd.c **/ 




/* the return command for synchronous commmands call */

#define SKEL_SET_RESPONSE     "SetResponse"

/* a callback for asynchronous call for robot position */

#define ROB_POS_RESPONSE "SetRobPos"

/* the port to listen on */
#define SKEL_PORT             40000


#define GLOBAL 1
#define LOCAL 2

/* socket communication buffers */
static char skel_out_buffer[10000];

static void  SKEL_InitResponse(char *buff);


/*since everything is done through callbacks we need
  static local variable for data exchange */

/*index on the last image data received */
static long sick_image_index = -1;

/*pointer to the current map as an array of longs */
 
static long *MapArray = NULL;

/* function to get the index on the map in internal format */
long GetMapIndex();


/* allows for reading 200 segs locally : should be enough ! */
#define NB_MAX_SEGS 200
static long segmentArray[4*NB_MAX_SEGS];


/* allows for reading 730 points locally : should be enough ! */
#define NB_MAX_PTS 730
static long pointArray[2*NB_MAX_PTS];




/* We have to define this in order to use worldcomm even if it 
   doesn't use it yet, but I don't know yet what I am going to do 
   with this ...*/

static struct WLD_World *map_serv_world;

static void RegisterMapServerCommands();

/*
  Flag for allowing (in continuous mode) to match segments one by one according 
  to the current robot position/variance. Useful when going along a long corridor
  in order to keep the robot position ok in 2 dimension.
  WARNING : might induce some instability in the result
*/

int SubRelocOk = 1;

static void BuildSegResponse(long *seg)
{
  int i;
  for(i = 0; i < 4 * seg[0] + 1; i++){
    BUF_AddLong(seg[i]);
    BUF_AddSpace();
  }

}

/* 
   there should be a mode here to say 
   the points are expressed as a regular
   serie of distances (for comm speed purpose)
*/

static void BuildPtResponse(long *seg)
{
  int i;
  for(i = 0; i < 2 * seg[0] + 1; i++){
    BUF_AddLong(seg[i]);
    BUF_AddSpace();
  }

}

static void BuildRobotPosResponse(double *pos)
{
  BUF_AddLong((long)(pos[0] * REAL2LONG));
  BUF_AddSpace();
  BUF_AddLong((long)(pos[1] * REAL2LONG));
  BUF_AddSpace();
  BUF_AddLong((long)(pos[2] * REAL2LONG));
  BUF_AddSpace();
  BUF_AddLong((long)(pos[3] * REAL2LONG));
  BUF_AddSpace();


}

/*** MAP EDITING ***********************************************/

/**** 
      removes n segments from the map and returns the corrected map
      To be consistent, the values should be in double (metric) in the call to RemoveSegment.
      input values will be divided by REAL2LONG
*****/


static char *RemoveSegmentsCmd(int sender_id, long robot_id, int argc, const char **argv)
{

  long n_seg,i;
  double x1,x2,y1,y2;

  assert (argc > 0);
  
  n_seg = atol (argv[1]);
  for(i = 0; i < n_seg ; i++){
    x1 = (double)(atol(argv[4 * i + 2]))/REAL2LONG;
    y1 = (double)(atol(argv[4 * i + 3]))/REAL2LONG;
    x2 = (double)(atol(argv[4 * i + 4]))/REAL2LONG;
    y2 = (double)(atol(argv[4 * i + 5]))/REAL2LONG;
    RemoveSegment(x1,y1,x2,y2);       
  }
  return NULL;
}

static char *RotateMapCmd(int sender_id, long robot_id, int argc, const char **argv)
{

  assert (argc > 0);
  
  RotateMap(atol(argv[1])/REAL2LONG);
  return NULL;

}

/*
  take two coordinates system : 
   -starting point and angle in the current map coordinate 
   system
   - what we want this stating point/angle to become in the new 
   coord sys.
*/

static char *ChangeMapCoordinatesCmd(int sender_id, long robot_id, int argc, const char **argv)
{

  
  double start[3], end[3];

  assert (argc == 7);
  
  start[0] = atol(argv[1])/REAL2LONG;
  start[1] = atol(argv[2])/REAL2LONG;
  start[2] = atol(argv[3])/REAL2LONG;
  end[0] = atol(argv[4])/REAL2LONG;
  end[1] = atol(argv[5])/REAL2LONG;
  end[2] = atol(argv[6])/REAL2LONG;

  ChangeMapCoordinates(start, end);

  return NULL;

}

/*****
  merging segments : takes n segments and merge them into one.
  Resulting segment gets the maximum possible extension.
  Format of call: 
  (long)number of segments
  (long)x1 y1 x2 y2 .....

  Return the updated map.
*****/



static char *MergeSegmentsCmd(int sender_id, long robot_id, int argc, const char **argv)
{

  long n_seg,i;
  double *seg;

  assert (argc > 0);
  
  n_seg = atol (argv[1]);
  
  /* building the array */
  seg = (double *)malloc((n_seg *4 + 1) * sizeof(double));
  seg[0] = n_seg;

  for(i = 1; i < 4*n_seg + 1 ; i++){
    seg[i] = atol(argv[i+1])/REAL2LONG;
  }

  MergeSegments(seg);

  free(seg);
  
  return NULL;
  

}


/** intersects 2 segments and return the new map **/

static char *InterSegmentsCmd(int sender_id, long robot_id, int argc, const char **argv)
{

  long *map;


  assert (argc > 0);
    
  IntersectsSegments(atol(argv[1]),atol(argv[2]),atol(argv[3]),atol(argv[4]),
		    atol(argv[5]),atol(argv[6]),atol(argv[7]),atol(argv[8]));       
 

  /** preparing for more data integration if the user wants **/
  /* UpdateLocalMap(); */

  /** getting the current map **/
  map = GetCurrentMap();
  /** returning result ***/

  SKEL_InitResponse(skel_out_buffer);
  /* if there was a pb ... I should go through all this ! */
  /* 
     for now I put 0 if pb . otherwise the first long will be for 
     defining the coordinate system the image is defined in 
  */
  
  BUF_AddLong(GLOBAL);
  BUF_AddSpace();
  BuildRobotPosResponse(Glob_pos);
  BuildSegResponse(map+4);
  free(map);
  return skel_out_buffer;

}


/*****************Sick Commands *******************************/


static char *InitSickCmd(int sender_id, long robot_id, int argc, const char **argv)
{

  
    SKEL_InitResponse(skel_out_buffer);

    if(InitSick())      
      BUF_AddLong(0);
    else
      BUF_AddLong(1);
   return skel_out_buffer;
   

}
#define SG_ONLY 0
#define PT_ONLY 1
#define PT_SG 2

/*
  Gets sick data from 2 sources : saved file or actual sick.

  Registered under GetRecordData and GetSickData .

  Options to come :

  - segments only (default)
  - points 
  - segments + points
 
  There will be a pb if reading a session. For now, points are not saved so we will 
  start by disableing the point option when using GetRecordData. 
  There should be an option to save points as well as segments (or maybe just points
  and then resegment on demand ?).

*/


static char *GetSickDataCmd(int sender_id, long robot_id, int argc, const char **argv)
{
  double abs_rob_pos[4];
  int op = SG_ONLY;
  assert(argc > 0);

  if( argc > 1)
    op = atoi(argv[1]);

  /* here I  don't know yet what to do : should I keep the images in memory in case
     the user wants to look at them. In that case I should define a server call
     for freeing them at user request and I should also return an Id on each 
     image transmitted to the user. To be done */
  /* for now I free them automatically when a new call is issued */
  
  if ( sick_image_index >= 0 && sick_image_index != GetMapIndex())
     Gfree_image(sick_image_index); 
     
    
  
  if(!strcmp(argv[0],"GetSickData")){

      sick_image_index = GetSickData(op);

      if(sick_image_index < 0)
      {
	  SKEL_InitResponse(skel_out_buffer);
	  BUF_AddLong(0);  
	  printf ("GetSickDataCmd : not connected to sick\n");
	  return skel_out_buffer;
      }
      /*saving if a session is open .
	NOTE : currently, this will save only the segments and only 
	if segments where requested. To be updated.
       */
      if(op == SG_ONLY || op == PT_SG)
	if(SessionIsOpen())
	  SaveSession(sick_image_index);
  }
    

  else /* keeping on reading the recorded session */
    {
      sick_image_index = Get_Image();
      ReadSession(sick_image_index);
       
    }
  

  /* computing the new robot position in abs_rob_pos
     from the encoders saved in the image */
  
  GetAbsPos((PtoIm(sick_image_index))->enc,abs_rob_pos);
  
  /* storing the aboslute position in the image */
  SetAbsPosImage(sick_image_index,abs_rob_pos);
  
  SKEL_InitResponse(skel_out_buffer);
  
  /* if there was a pb ... I should go through all this ! */
  /* 
     for now I put 0 if pb . otherwise the first long will be for 
     defining the coordinate system the image is defined in 
  */
  
  BUF_AddLong(LOCAL);
  BUF_AddSpace();

  BuildRobotPosResponse(abs_rob_pos);
  if(op == SG_ONLY || op == PT_SG){
    /* copying in the array : the segments are kept in local frame */
    Copy_image_segment(sick_image_index,segmentArray,NB_MAX_SEGS);
    BuildSegResponse(segmentArray);
  }
  if(op == PT_ONLY || op == PT_SG){
    Copy_image_point(sick_image_index,pointArray,NB_MAX_PTS);
    BuildPtResponse(pointArray);

  }

  return skel_out_buffer;
}

/********************* File commands *******************************/

static char *ReadSegmentFileCmd(int sender_id, long robot_id, int argc, const char **argv)
{
  
  
  double abs_rob_pos[4];
  assert(argc > 0);
    
  
  sick_image_index = Get_Image();
  rd_file_sg_c(PtoIm(sick_image_index),argv[1]);
      

  /* copying in the array : the segments are kept in local frame */
  Copy_image_segment(sick_image_index,segmentArray,NB_MAX_SEGS);
  
  SKEL_InitResponse(skel_out_buffer);
  /* if there was a pb ... I should go through all this ! */
  /* 
     for now I put 0 if pb . otherwise the first long will be for 
     defining the coordinate system the image is defined in 
  */
  GetAbsPos(PtoIm(sick_image_index)->enc,abs_rob_pos);
  BUF_AddLong(LOCAL);
 
  BUF_AddSpace();
  BuildRobotPosResponse(abs_rob_pos);
  BuildSegResponse(segmentArray);
  return skel_out_buffer;
}

static char *ReadMapFileCmd(int sender_id, long robot_id, int argc, const char **argv)
{
  
  assert(argc > 0);
    
  SKEL_InitResponse(skel_out_buffer);
 
  /* POURQUOI L'EXPORTATION DE LA CARTE VERS LE CLIENT A-T-ELLE
          ETE SUPPRIMEE ? */
  /*
    PHIL : parce que l'API a chnage et que demander a server de charger une carte 
    est different de lui demander de nous la transmettre. 
    Il existe un appel GetCurrentMap pour cela . 
  */
  
  if(! RetrieveMap(argv[1])){
    BUF_AddLong(0);


  }
  else
    BUF_AddLong(1);
 
  return skel_out_buffer;

}


static char *SaveMapFileCmd(int sender_id, long robot_id, int argc, const char **argv)
{
  assert (argc > 0);
  SKEL_InitResponse(skel_out_buffer);
  if(SaveMapToFile(argv[1]) < 0)
    BUF_AddLong(0);
  else
    BUF_AddLong(1);
  return skel_out_buffer;

}


/*********************** MAP BUILDING ********************************/
/**************** see also SESSION MANAGEMENT ***********************************/

/***
    erases the current map (if any) from the server and reinitialize
    to an empty map.
****/

static char *EraseMapCmd(int sender_id, long robot_id, int argc, const char **argv)
{
  assert(argc > 0);
  Erase_map(); 
  SKEL_InitResponse(skel_out_buffer);
  return skel_out_buffer;
}

/* adds the last image saved in 'sick_image_index' to the map. Returns the map */
 
static char *DataIntegrationCmd(int sender_id, long robot_id, int argc, const char **argv)
{


  assert(argc > 0);

  MapArray = DataIntegration(sick_image_index);
  SKEL_InitResponse(skel_out_buffer);
  /* 
     for now I put 0 if pb . otherwise the first long will be for 
     defining the coordinate system the image is defined in 
  */
  
  BUF_AddLong(GLOBAL);
  
  BUF_AddSpace();
  /*sending the result */
  BuildRobotPosResponse(Glob_pos);
 
  BuildSegResponse(MapArray + 4);
  free(MapArray);
  
  return skel_out_buffer;
  
}

/*** same as above but takes a new image from the sick ***/

static char * NewDataIntegrationCmd(int sender_id, long robot_id, int argc, const char **argv)
 
{
  double abs_rob_pos[4];
  assert(argc > 0);

  if(ServConnected2Robot()){
    if( sick_image_index >= 0 && sick_image_index != GetMapIndex())
      Gfree_image(sick_image_index);
    sick_image_index = SickGet_image(RobId);
    segmentation(PtoIm(sick_image_index),1);
    /* getting the absolute position according to the robot's encoders */
    GetAbsPos((PtoIm(sick_image_index))->enc,abs_rob_pos);
    /* storing the aboslute position in the image */
    SetAbsPosImage(sick_image_index,abs_rob_pos);
    
    /*saving if a session is open */
    if(SessionIsOpen())
      SaveSession(sick_image_index);
  }
  else {
    SKEL_InitResponse(skel_out_buffer);
    BUF_AddLong(0);
    return skel_out_buffer;
  }
  MapArray = DataIntegration(sick_image_index);
  SKEL_InitResponse(skel_out_buffer);
  /* 
     for now I put 0 if pb . otherwise the first long will be for 
     defining the coordinate system the image is defined in 
  */
  
  BUF_AddLong(GLOBAL);
  
  BUF_AddSpace();
  /*sending the result */
  BuildRobotPosResponse(Glob_pos);
 
  BuildSegResponse(MapArray + 4);
  free(MapArray);
  MapArray = NULL;
  
  return skel_out_buffer;
  
}






/******************** LOCALIZATION *******************************/

#define NB_MAX_POSITIONS 500
static long positionsArray[4*NB_MAX_POSITIONS+1];
static long transformsArray[4*NB_MAX_POSITIONS+1];
static long best = -1;

/* 
   This will take 2 , 4 or 5 arguments. The first 2 arguments give  an 
   estimate x,y position of the robot with respect to the map in millimeter 
   Optional 2 next arguments give an approximate angle (degrees) and a possible deviation
   on each side (degrees). NOTE : if dev = 180, those argument are ignored.
   Last optional argument is an option for filtering result : if equal to 1, the localization
   will reject solution leading to object behind walls. Default to 0
*/ 

 static char *LocalLocalizeCmd(int sender_id, long robot_id, int argc, const char **argv)
{
  double xl,yl,thl;
  int data;
  int i,op = 0;


  assert(argc > 0);
  xl = atol(argv[1])/REAL2LONG;
  yl = atol(argv[2])/REAL2LONG;

  /** if we know something about the angle */
  if(argc >= 5){
    thl = atol(argv[3])*0.18/M_PI;
    
    Set_loc_angle(thl,atol(argv[4]));
  }
  else {
    Unset_loc_angle();

  }
  if(argc == 6)
    op = atol(argv[5]);

  /*** getting an image  ****/
  if(ServConnected2Robot()){
      data = SickGet_image(RobId);
      segmentation(PtoIm(data),1);
  }
  else {
    SKEL_InitResponse(skel_out_buffer);
    BUF_AddLong(0);
    return skel_out_buffer;
  }

  /** updating the robot position according to encoders value stored in 
  the new data image **/

   set_encoders(PtoIm(data)->enc);

  /*** filtering data with length too small ****/

  FilterOutData(data);

  /** turning off the Localisation filter ***/
  /** this has to be done if we have no idea about position confidence value ***/
  /** here, the user provides the aprrox location and no confidence value is known ***/
  /** about it **/ 
  LocFilterOff();


  /*** op here means we  filter or not the solution according to visibility criteria ***/
  best = Local_localization(data,xl,yl,transformsArray,op,NB_MAX_POSITIONS);
  /** getting the positions also ***/
  GetPositionArray(positionsArray, NB_MAX_POSITIONS);

  Gfree_image(data);

  SKEL_InitResponse(skel_out_buffer);
  /* 
     for now I put 0 if pb . otherwise the first long will be for 
     defining the coordinate system the image is defined in 
  */
  if(best < 0){
    BUF_AddLong(0);
    return skel_out_buffer;
  }
  BUF_AddLong(1);
  BUF_AddSpace();
  /** 
    well, here I return only the best position ,
    I think that a list of solution is mainly useful for
    absolute localization.
    There will be a server call to return the current list of 
    possibilities. To be done .
    */
  printf("Pos: %.3f %.3f %.3f  %ld segs\n",
	 positionsArray[best*4 + 1]/REAL2LONG,
	 positionsArray[best*4 + 2]/REAL2LONG,
	 positionsArray[best*4 + 3]/REAL2LONG,
	 positionsArray[best*4 + 4]);
  for(i = 0; i < 4; i++){
    BUF_AddLong(positionsArray[best*4 + 1+ i]);
    BUF_AddSpace();
  }
  return skel_out_buffer;
}

 static char *GlobalLocalizeCmd(int sender_id, long robot_id, int argc, const char **argv)
{
  double thl;
  int data;
  int op = 0;


  assert(argc > 0);
  

  /** if we know something about the angle */
  if(argc >= 3){
    thl = atol(argv[1])*0.18/M_PI;
    
    Set_loc_angle(thl,atol(argv[2]));
  }
  else {
    Unset_loc_angle();

  }

  /* visi option on here otherwise will produce too many results */
  op = 1;

  /*** getting an image  ****/
  if(ServConnected2Robot()){
      data = SickGet_image(RobId);
      segmentation(PtoIm(data),1);
  }
  else {
    SKEL_InitResponse(skel_out_buffer);
    BUF_AddLong(0);
    return skel_out_buffer;
  }

  /** updating the robot position according to encoders value stored in 
  the new data image **/

   set_encoders(PtoIm(data)->enc);

  /*** filtering data with length too small ****/

  FilterOutData(data);

  /** turning off the Localisation filter ***/
  /** this has to be done if we have no idea about position confidence value ***/
  /** here, the user provides the aprrox location and no confidence value is known ***/
  /** about it **/ 
  LocFilterOff();


  /*** op here means we  filter or not the solution according to visibility criteria ***/
  best = Global_localization(data,transformsArray,op,NB_MAX_POSITIONS);
  /** getting the positions also ***/
  GetPositionArray(positionsArray, NB_MAX_POSITIONS);

  /*Gfree_image(data);*/

  SKEL_InitResponse(skel_out_buffer);
  /* 
     for now I put 0 if pb . otherwise the first long will be for 
     defining the coordinate system the image is defined in 
  */
  if(best < 0){
    BUF_AddLong(1);
    return skel_out_buffer;
  }
  BUF_AddLong(0);
  /*BUF_AddSpace();*/
  /** 
    well, here I return only the best position ,
    I think that a list of solution is mainly useful for
    absolute localization.
    There will be a server call to return the current list of 
    possibilities. To be done .
    */
  /*printf("Pos: %.3f %.3f %.3f  %ld segs\n",
	 positionsArray[best*4 + 1]/REAL2LONG,
	 positionsArray[best*4 + 2]/REAL2LONG,
	 positionsArray[best*4 + 3]/REAL2LONG,
	 positionsArray[best*4 + 4]);
  for(i = 0; i < 4; i++){
    BUF_AddLong(positionsArray[best*4 + 1+ i]);
    BUF_AddSpace();
  }*/
  return skel_out_buffer;
}



/*************************************************
 takes the last image from GetSickData and tries all current hypothesis
 to see if they are valid . Rejects bad hypothesis and returns as first arg the number of remaining hypothesis . One should call GetSickData before this, and 
 GetRobPositions if they exists
 *********************************************************/


static char * FilterBadHyposCmd(int sender_id, long robot_id, int argc, const char **argv) 
{
        int laser_image;
    int numPossibilities = 0;
    assert(argc > 0);
    

 
    if ( (laser_image = GetLastImageIndex()) >= 0 && laser_image != GetMapIndex())
    {
	
	numPossibilities = FilterBadLocations(laser_image,transformsArray,1.0,NB_MAX_POSITIONS);
	/** getting the positions also ***/
	GetPositionArray(positionsArray, NB_MAX_POSITIONS);
    }
    /*Gfree_image(data);*/
    
    SKEL_InitResponse(skel_out_buffer);
    
    
    BUF_AddLong(numPossibilities);

    /*BUF_AddSpace();*/
    /** 
	well, here I return only the best position ,
	    I think that a list of solution is mainly useful for
	    absolute localization.
	    There will be a server call to return the current list of 
	    possibilities. To be done .
	*/
	/*printf("Pos: %.3f %.3f %.3f  %ld segs\n",
	  positionsArray[best*4 + 1]/REAL2LONG,
	  positionsArray[best*4 + 2]/REAL2LONG,
	  positionsArray[best*4 + 3]/REAL2LONG,
      positionsArray[best*4 + 4]);
      for(i = 0; i < 4; i++){
      BUF_AddLong(positionsArray[best*4 + 1+ i]);
      BUF_AddSpace();
      }*/
	return skel_out_buffer;
}
    
 
/* 
   Tracking mode. Here the approximate position is taken from the robot itself.
   This will take 3 optional arguments. 
   Optional first  argument give a possible deviation
   on each side (degrees) of the current approximate robot
   position. If == 180, this argument is ignored.
   Second argument is an option saying that we must (if == 1) rely on 
   current robot position confidence. Default is true for tracking mode.
   Last optional argument is an option for filtering result : if equal to 1,
   the localization will reject solution leading to object behind walls. Default to 0.
*/ 

static char *TrackOnceCmd(int sender_id, long robot_id,
				int argc, const char **argv)
{
  int data;
  int i,op=0;


  assert(argc > 0);
  /* xl = atol(argv[1])/REAL2LONG;
  yl = atol(argv[2])/REAL2LONG;
  */
  
   /** here, the robot provides the aprrox location and some confidence value is known ***/
  /** about it  **/ 
  /** turning on the Localisation filter ... ***/
  LocFilterOn();

  /** unless the user doesn't want it **/
  if(argc >= 3)
    if(! atol(argv[2]))
      LocFilterOff();


  if(argc >= 4)
    op = atol(argv[3]);

  /*** getting an image  ****/
  if(ServConnected2Robot()){
      data = SickGet_image(RobId);
      segmentation(PtoIm(data),1);
  }
  else {
    SKEL_InitResponse(skel_out_buffer);
    BUF_AddLong(0);
    return skel_out_buffer;
  }

  /*** filtering data with length too small ****/

  FilterOutData(data);

  /** updating the robot position according to encoders value stored in 
  the new data image **/

  set_encoders(PtoIm(data)->enc);

  /** computing the variance matrix in Vpos ***/
  /* done in set_encoders */
  /*get_location_err();*/

  /** if we know something about the angle */
  
  if(argc >= 2){
    
    
    Set_loc_angle(Glob_pos[2]*180.0/M_PI,atol(argv[1]));
  }
  else {
    Unset_loc_angle();

  }


  

  /*** op here means we  filter or not the solution according to visibility criteria ***/
  best = Local_localization(data,Glob_pos[0],Glob_pos[1],transformsArray,
			    op,NB_MAX_POSITIONS);

  GetPositionArray(positionsArray, NB_MAX_POSITIONS);

  Gfree_image(data);

  SKEL_InitResponse(skel_out_buffer);
  /* 
     for now I put 0 if pb . otherwise the first long will be for 
     defining the coordinate system the image is defined in 
  */
  if(best < 0){
    BUF_AddLong(0);
    return skel_out_buffer;
  }
  BUF_AddLong(1);
  BUF_AddSpace();
  /** 
    well, here I return only the best position ,
    I think that a list of solution is mainly useful for
    absolute localization.
    There will be a server call to return the current list of 
    possibilities. To be done .
    */
  printf("Position: %.3f %.3f %.3f %ld\n",
	 positionsArray[best*4+1]/REAL2LONG,
	 positionsArray[best*4+2]/REAL2LONG,
	 positionsArray[best*4+3]/REAL2LONG,
	 positionsArray[best*4+4]);
  for(i = 0; i < 4; i++){
    BUF_AddLong(positionsArray[best*4 + 1+ i]);
    BUF_AddSpace();
  }
  return skel_out_buffer;
}


/* 
   Continous Tracking mode. This function returns immediately after setting 
   TRACKING_FLG to 1 . The main loop will flush rpc calls and call TrackOnce as long
   as the flag TRACKING_FLG is on. A second call to this function will set TRACKING_FLG to 0.

   Here the approximate position is taken from the robot itself.
   This will take 4 optional arguments. 

   Optional first  argument give a possible deviation
   on each side (degrees) of the current approximate robot
   position. If == 180, this argument is ignored. No argument means no angle 
   restriction.

   Second argument is an option saying that we must (if == 1) rely on 
   current robot position confidence. Default is true for tracking mode.

   Third optional argument is an option for filtering result : if equal to 1,
   the localization will reject solution leading to object behind walls. Default to 0.

   Fourth argument : Flag to get the variance as well : if set, returns 6 doubles after the localization: 
   vxx, vxy,vxt,vyy,vyt,vtt

   NOTE: in case of multiple clients those options might have to change between clients.
   This is a pb. For now, the last change will override previous setting
*/ 

static long Deviation , VisOp;
BOOL TRACKING_FLG = 0;
static int track_client_id = -1;
BOOL ReturnVariance = 0;
static char *ContinuousTrackCmd(int sender_id, long robot_id,
				int argc, const char **argv)
{
  MSERV_ClientDescription *client_d;
  struct CONN_ConnInfo *client_info;
  int error;


  assert(argc > 0);
  /* getting the conn info structure */

  
  error = CONN_FDtoConnInfo(sender_id,&client_info);
  assert(error == N_NO_ERROR);
  client_d = (MSERV_ClientDescription *)(client_info->user_data);

  assert(client_d);
  
  
  /** if called twice : deactivate **/

  if (client_d->auto_track_wanted){
    client_d->auto_track_wanted = FALSE;
    SKEL_InitResponse(skel_out_buffer);
    return skel_out_buffer;
    
  }
  
  client_d->auto_track_wanted = TRUE;

  if(argc >= 2)
    Deviation = atol(argv[1]);
  else
    Deviation = 180;

   /** here, the robot provides the aprrox location and some confidence value is known ***/
  /** about it  **/ 
  /** turning on the Localisation filter ... ***/
  LocFilterOn();



  /** unless the user doesn't want it **/
  if(argc >= 3)
    if(! atol(argv[2]))
      LocFilterOff();


 
  if(argc >= 4)
    VisOp = atol(argv[3]);

  /* do we want the variance as well ? */
  if(argc >= 5)
      ReturnVariance = atol(argv[4]);

  SKEL_InitResponse(skel_out_buffer);
  


  return skel_out_buffer;

  

}

void TrackOnce(long dev,long vis_op)
{
  long data;
  char *buffer;
  struct CONN_ConnInfo *conn_info;
  long nloc, nsuccess;
  
  /*** getting an image  ****/
  if(ServConnected2Robot()){
    data = SickGet_image(RobId);
    segmentation(PtoIm(data),1);
  
    
    /*** filtering data with length too small ****/
    
    FilterOutData(data);
    
    /** updating the robot position according to encoders value stored in 
      the new data image **/
    
    set_encoders(PtoIm(data)->enc);
    
    /** computing the variance matrix in Vpos ***/
    /* done in set_encoders */
    /*get_location_err();*/
    
    /** if we know something about the angle */
    
    Set_loc_angle(Glob_pos[2]*180.0/M_PI,dev);
    
    /*** op here means we  filter or not the solution according to visibility criteria ***/
    best = Local_localization(data,Glob_pos[0],Glob_pos[1],
			      transformsArray,vis_op,NB_MAX_POSITIONS);

    

    /** responding to clients **/
    /*Oops : if nothing good found, I should not respond anything to the clients !!!*/
    /**
       In fact, this should always say if it found a solution and how many where found 
	This way, the client could know if there is a pb. I add 2 numbers in front of the message 
	- nloc > 0 => localization successful, n possibilities
	- nloc <= 0 => localization failed, nloc possibilities rejected AFTER filtering.
	- nsuccess > 0 : number of time in a row the localization was successful
	- nsuccess < 0 : number of time in a row the localization was unsuccessful
    */

    nloc = GetNpossibilities();
    nsuccess = GetNSuccess();

    printf("%3lu  (%2lu poss)\n",nsuccess, nloc);

    /* 
       if there was no possible matchings even BEFORE filtering : we assume 
       there was not enough info in the map/data (like in a corridor) .
       If the flag SubRelocOk is on, we try to use the current variance 
       of the robot in order to match segments one by one 
    */
    if(SubRelocOk && nloc == 0)
      {
	MapTrData(PtoIm(data),Glob_pos,Vpos,12.);
	/** re-initializing the transform from robot encoders to global position **/
	init_location_st(Glob_pos,Vpos);
      }
    Gfree_image(data);
    BUF_Reset(NULL);
    BUF_AddString(ROB_POS_RESPONSE);
    BUF_AddSpace();
    
    BUF_AddLong(nloc);
    
    BUF_AddSpace();
    
    BUF_AddLong(nsuccess);
    
  
    
      
      BUF_AddSpace();


      printf("transf: %.3f %.3f %.3f\n",
	     Loc2Abs[0], Loc2Abs[1], Loc2Abs[2]);

      

      BUF_AddLong((long)(Loc2Abs[0]*REAL2LONG));
      
      BUF_AddSpace();
      
      BUF_AddLong((long)(Loc2Abs[1]*REAL2LONG));
      
      BUF_AddSpace();
      
      BUF_AddLong((long)(Loc2Abs[2]*REAL2LONG));


      /* If we want the variance */
      
      if(ReturnVariance){
	  printf("variance server : %f %f %f %f %f %f  \n",Vpos[0],Vpos[1],Vpos[3],Vpos[2],Vpos[4],Vpos[5]);
	    BUF_AddSpace();
	    BUF_AddDouble(Vpos[0]);
	    BUF_AddSpace();
	    BUF_AddDouble(Vpos[1]);
	    BUF_AddSpace();
	    BUF_AddDouble(Vpos[3]);
	    BUF_AddSpace();
	    BUF_AddDouble(Vpos[2]);
	    BUF_AddSpace();
	    BUF_AddDouble(Vpos[4]);
	    BUF_AddSpace();
	    BUF_AddDouble(Vpos[5]);
      }

      BUF_AddChar('\n');

      /*PHIL: this is a little bit too bad we have to copy the same buffer n times */
      /*NOTE : there is no client settings here, I broadcast.  but it should ...To be done */
      
      InitReadClientTable();
      while ((conn_info = GetNextClient())){
	if ((buffer = strdup(BUF_Buffer())) == NULL)
	  
	  return ;
	
	
	
	CONN_SendToConnection(conn_info,buffer,BUF_Index(),0);
      }
  }
    

	       
	     
  else {
    TRACKING_FLG = 0;
  }
    
    
}

BOOL MAPING_FLG = 0;


static char *ContinuousMapingCmd(int sender_id, long robot_id,
				int argc, const char **argv)
{
 

  assert(argc > 0);
  
  /** if called twice : deactivate **/

  if (MAPING_FLG){
    MAPING_FLG = FALSE;
    
    SKEL_InitResponse(skel_out_buffer);
    return skel_out_buffer;
    
  }
  track_client_id = sender_id;

  

  SKEL_InitResponse(skel_out_buffer);
  
  MAPING_FLG = 1;

  return skel_out_buffer;

  

}


void MapOnce()
{
  long data;

  /*** getting an image  ****/

  if(ServConnected2Robot()){
    data = SickGet_image(RobId);
    segmentation(PtoIm(data),1);

     /*saving if a session is open */
    if(SessionIsOpen())
      SaveSession(data);

    if(MapArray)
      free(MapArray);

    MapArray = DataIntegration(data);

  }
	       
	     
  else {
    MAPING_FLG = 0;
  }
    
    
}


/**
   Connecting to the robot. 
   At that point we need to establish consistency between the robot's internal position 
   and the global position.
   Here my choice is : I set the global position to be 0 and then I compute the transform 
   according to the internal position (done by init_encoders())
   However, this should happen only ONCE. Otherwise what would happen when 
   multiple clients connect ??

   Normally, if a client connects to the server, it should ask if the server is
   already connected before. In any case, we prevent multiple connections here (done in 
   ConnectServer2Robot).
   Note that it also prevents the server from beeing connected to several robots (good !!)
 */


static char *ConnectRobotCmd(int sender_id, long robot_id, int argc, const char **argv)

{
  
  SKEL_InitResponse(skel_out_buffer);

  
    

  assert(argc > 0);
  if (argc < 4)
    {
      /*sprintf(skel_out_buffer, "Error %s %d", argv[0], N_MISSING_ARGUMENT); */
      BUF_AddLong(0);
      return skel_out_buffer;
    }
  
 
  
  
  RobId = atol(argv[3]);

  
  
  if(!ConnectServer2Robot(argv[1],atoi(argv[2]), atol(argv[3]))){
    /* there was a pb */
    BUF_AddLong(0);
    
    return skel_out_buffer;
  }
    
  BUF_AddLong(1);
    
  return skel_out_buffer;
}







static char *GetRobotPositionCmd(int sender_id, long robot_id, int argc, const char **argv)
{

   double enc[4];
   double map_pos[4];
   assert(argc > 0);
   GetRobotEncoders(enc,RobId);
   GetAbsPos(enc,map_pos);

   SKEL_InitResponse(skel_out_buffer);

   BUF_AddLong(1);
   BUF_AddSpace();
   BUF_AddLong((long)(map_pos[0]*1000));
   BUF_AddSpace();
   BUF_AddLong((long)(map_pos[1]*1000));
   BUF_AddSpace();
   BUF_AddLong((long)(map_pos[2]*1000));
   BUF_AddSpace();
   BUF_AddLong((long)(map_pos[3]*1000));
   

   return skel_out_buffer;
}


static char *GetRobotTransformCmd(int sender_id, long robot_id, int argc, const char **argv)
{

   assert(argc > 0);
  
   SKEL_InitResponse(skel_out_buffer);

   BUF_AddLong(1);
   BUF_AddSpace();
   BUF_AddLong((long)(Loc2Abs[0]*1000));
   BUF_AddSpace();
   BUF_AddLong((long)(Loc2Abs[1]*1000));
   BUF_AddSpace();
   BUF_AddLong((long)(Loc2Abs[2]*1000));
   return skel_out_buffer;
}


/** tries to open the directory if it exists or return a failure code **/
/** format : (string)dirname optional (long)erase optional (long)start_index **/
/** if erase == 1 : erase any map in memory and start from scratch else start from
current map . default = 1 **/
/** if start_index >= 0 : sets the SessionCounter to start_index allowing to continue a session from there. Default = 0 **/    
/** sends a code 0 as first return argument on failure or 1 on success */


static char *OpenSessionCmd(int sender_id, long robot_id, int argc, const char **argv)
{
  BOOL OK =FALSE; 
   assert(argc > 0);
   if (argc < 2)
     {
       sprintf(skel_out_buffer, "OpenSessionCmd: Error %s %d", argv[0], N_MISSING_ARGUMENT);
       return skel_out_buffer;
     }
   if(argc > 3)
     OK = OpenSaveSession(argv[1],atol(argv[3]));
   else
     OK = OpenSaveSession(argv[1],0);
   
  SKEL_InitResponse(skel_out_buffer);
  if(!OK)
    BUF_AddLong(0);
  else
    BUF_AddLong(1);

   /*** erasing and reseting a possible previous map ***/
   if(argc > 2)
     if(atol(argv[2]))
       Erase_map();

   
   return skel_out_buffer;
 }

/** tries to open the directory if it exists or create it if it doesn't **/
/** format : (string)dirname optional (long)erase optional (long)start_index **/
/** if erase == 1 : erase any map in memory and start from scratch else start from
current map . default = 1.
In that case, the server position will be set to the origin.
 Otherwise, it will use as a starting value. This means that either the current value is 
assumed good enough for tracking or a manual localization should occur 
with respect to the current map. 
 **/
/** if start_index >= 0 : sets the SessionCounter to start_index allowing to continue a session from there. Default = 0 **/    
/** sends a code 0 as first return argument on failure or 1 on success */
/** 
  NOTE : this function requires the server to be connected to the robot. 
  If not, you should use OpenSession for off-line map work 
***/
static char *InitRecSessionCmd(int sender_id, long robot_id, int argc, const char **argv)
{
  long start_index = 0;
  long erase= 1;

   assert(argc > 0);
   if (argc < 2)
     {
       sprintf(skel_out_buffer, "Error %s %d", argv[0], N_MISSING_ARGUMENT);
       return skel_out_buffer;
     }
  if(!ServConnected2Robot() )
    {
      SKEL_InitResponse(skel_out_buffer);
      BUF_AddLong(0);
      return skel_out_buffer;

    }

  
   GetRobotEncoders(rob_enc,RobId);
  
   /* 
      we decide the starting point of this session is the origin 
      of the absolute frame ;
      Glob_pos is set to the origin point;
      Loc2Abs is computed to make the transform between rob_enc and 
      Glob_pos.
   */

   /*OUTPHIL, this is no good : 
     init_encoders(rob_enc);
     init_encoders occurs automatically when restarting a map */

   SKEL_InitResponse(skel_out_buffer);
  if(argc > 3)
    start_index = atol(argv[3]);
  if( ! InitRecSession(argv[1],start_index)) 
     BUF_AddLong(0);
  else {
     BUF_AddLong(1);
     /*** erasing and reseting a possible previous map ***/
     /*** 
       this should be an option : one could want to continue on building 
       an already started map 
       ***/
     if(argc > 2)
       erase = atol(argv[2]);
     if(erase){
       Erase_map();
     }
   }
   return skel_out_buffer;

}



/** reads the next file in the session if there is one open . Then integrates 
    it in the map . Failure if no session open or map/localization failure .
**/

/** sends a code 0 as first return argument on failure or 1 on success */



static char *ReadSessionCmd(int sender_id, long robot_id, int argc, const char **argv)
{
   assert(argc > 0);
   
   if(sick_image_index != GetMapIndex() && sick_image_index != -1)
     Gfree_image(sick_image_index);
   sick_image_index = Get_Image();
  SKEL_InitResponse(skel_out_buffer);
  if( ! ReadSession(sick_image_index))
     {
       BUF_AddLong(0);
       return skel_out_buffer;
     }
  MapArray = DataIntegration(sick_image_index);

  /* 
     for now I put 0 if pb . otherwise the first long will be for 
     defining the coordinate system the image is defined in 
  */
  
    BUF_AddLong(GLOBAL);
  
   BUF_AddSpace();
   /*sending the result */

   BuildRobotPosResponse(Glob_pos);
   BuildSegResponse(MapArray + 4);
   free(MapArray);
   return skel_out_buffer;
 }

/**************** DATA ACCESS ******************************************/

/*** 
     Two ways to call this function : as "GetRobPositions" or as "GetRobTransforms".
     
     Returns an array of either absolute positions with respect to the map or 
     an array of tranforms bewtween the robot encoders and the absolute positions in
     the map .

     the first long incates the number of possibilities. 

     the second long indicates the index of the best bet according to the internal localization software.
      
     
     Then it goes like this for each possibility "p" :

     a) absolute position : 
              array[4*p + 2] = x (millimeters)
	      array[4*p + 3] = y (millimeters)
	      array[4*p + 4] = th (milliradiants)
	      array[4*p + 5] = number of matchings

      b) transform (assuming xc, yc and thc as encoder values) :
              
              array[4*p +2] + cos(array[4*p + 4]) * xc - sin (array[4*p + 4]) * yc = x
              array[4*p +3] + sin(array[4*p + 4]) * xc + cos (array[4*p + 4]) * yc = y
	      array[4*p + 4] + thc = th

     NOTE : this assumes a localization took place before. It will just read the arrays "positions" 
     and "transforms".
              
 **/

static char *GetRobotPositionsCmd(int sender_id, long robot_id, int argc, const char **argv)
{
  long *array, i;

  assert(argc > 0);
  /* absolute */
  if( ! strcmp(argv[0],"GetRobPositions")){
    if( best < 0 ) /* no valid localization available */
      {
	SKEL_InitResponse(skel_out_buffer);
	BUF_AddLong(0);
	return skel_out_buffer;
      }
    array = positionsArray;


  }
  /* transforms */
  else {

    if( best < 0 ) /* no valid localization available */
      {
	SKEL_InitResponse(skel_out_buffer);
	BUF_AddLong(0);
	return skel_out_buffer;
      }
    array = transformsArray;
    

    
  }
  SKEL_InitResponse(skel_out_buffer);
  BUF_AddLong(array[0]);
  BUF_AddSpace();
  BUF_AddLong(best);
  for(i = 0; i < 4*array[0] ; i++){ 
    BUF_AddSpace();
    BUF_AddLong(array[i + 1]);
  }
  
  return skel_out_buffer;
}





static char *GetCurrentMapCmd(int sender_id, long robot_id, int argc, const char **argv)

{
  long *map;
  if (!(map = GetCurrentMap())){
    SKEL_InitResponse(skel_out_buffer);
    BUF_AddLong(GLOBAL);
    return skel_out_buffer;
  }
 

  /* 
     for now I put 0 if pb . otherwise the first long will be for 
     defining the coordinate system the image is defined in 
  */
  SKEL_InitResponse(skel_out_buffer);
  BUF_AddLong(GLOBAL);
  
  BUF_AddSpace();

  /*sending the result */
  
  BuildRobotPosResponse(Glob_pos);
  BuildSegResponse(map + 4);
  free(map);
  return skel_out_buffer;
  
}

static char *GetCurrentSegmentsCmd(int sender_id, long robot_id, int argc, const char **argv)
{
  
  double abs_rob_pos[4]; 

  assert(argc > 0);    

  /* copying in the array : the segments are kept in local frame */
  Copy_image_segment(sick_image_index,segmentArray,NB_MAX_SEGS);
  
  SKEL_InitResponse(skel_out_buffer);
  /* if there was a pb ... I should go through all this ! */
  /* 
     for now I put 0 if pb . otherwise the first long will be for 
     defining the coordinate system the image is defined in 
  */
  GetAbsPos(PtoIm(sick_image_index)->enc,abs_rob_pos);
  BUF_AddLong(LOCAL);
 
  BUF_AddSpace();
  BuildRobotPosResponse(abs_rob_pos);
  BuildSegResponse(segmentArray);
  return skel_out_buffer;

}

/*********************************************************
  returns the robot name, port number and robot id we are 
  connected to.
  If no connection, returns 0 as the first parameter. Otherwise:

  first return parameter : robot id
  first return parameter : robot name
  first return parameter : port

********************************************************/


static char *GetConnectedRobotCmd(int sender_id, long robot_id, int argc, const char **argv)

{
  assert(argc > 0);

  SKEL_InitResponse(skel_out_buffer);
  if(!(RobId = GetConnectedRobot(RobotName,&RobPort))){
    BUF_AddLong(0);
    return skel_out_buffer;
  }

  /*This should be ONLY if running on hilare ...*/
#ifdef __HILARE2
  RobId = 1;
  sprintf(RobotName,"%s","hilare2");
  RobPort = 0;
#endif
  BUF_AddLong(RobId);
  BUF_AddSpace();
  BUF_AddString(RobotName);
  BUF_AddSpace();
  BUF_AddLong(RobPort);
  return skel_out_buffer;



}

/************************************************************************************/
/***        Helpers for RPC connection. Thanks to Jake & Keith.        **************/
/***        Note : I put all CMD_Register calls in a separate file which ************/
/***        is included in rpc_initialize()                              ************/
/************************************************************************************/ 


/* 
   in this server appplication, the response will always use SKEL_SET_RESPONSE 
   just not to have to rewrite a different function on the client side.
   Well, it wouldn't be bad to have a different call back when returning a segment image
   or a map (which is the same). To be done .
*/

static void  SKEL_InitResponse(char *buff)
{
  BUF_Reset(skel_out_buffer);
  BUF_AddString(SKEL_SET_RESPONSE);
  BUF_AddSpace();
}


/*callback called from worldcomm.c (if registered: look in RegisterMapServerCommands) upon
  connection. A client structure is linked to the conn_info structure and initialized.
  Here, the id is not used.
*/

static char *MapServer_ConnectionCBK(long id, struct CONN_ConnInfo *conn_info)
{
  MSERV_ClientDescription *client_d;
  client_d = (MSERV_ClientDescription *)malloc(sizeof(MSERV_ClientDescription));
  client_d->auto_track_wanted = FALSE;
  client_d->auto_map_wanted = FALSE;
  conn_info->user_data = client_d;
  return NULL;
}



#ifndef VXWORKS
static void sighandler()
{
  printf ("Bye\n");
  exit (0);
}
#endif

/* 
   Here the server will read a file called MapServer.ini if present in the current 
   directory. This will be done AFTER reading possible argumentson the command line. 
   It looks like this is a pb for vxworks which doesn't deal with variable argument on 
   the command line (Matthieu ???)
*/



#ifndef VXWORKS
int main(int argc , char **argv)
#else
STATUS
MapServer(char *robotName, int port, int robId)
#endif
{

  int error;
  struct CONN_ConnInfo *client_info;
  MSERV_ClientDescription *client_desc;

#ifndef VXWORKS
  char *robotName=NULL;
  int port;
#endif

  if((error = CONN_Initialize()) != N_NO_ERROR)
    return error;
 
  CONN_Listen(40000);
 
  /* Get parameters */

#ifndef VXWORKS
  if(argc == 4) {
    RobId = atol(argv[3]);
    robotName = argv[1];
    port = atoi(argv[2]);
  }
#else
  RobId = robId;
#endif

  /* Sig handler */

#ifndef VXWORKS
  signal(SIGINT,sighandler);
#endif

  /* connecting to the robot if any specified */

   
  if(robotName != NULL) {
  
    if(ConnectServer2Robot(robotName,port,RobId)){

      printf ("MapServer: connected to robot %s\n", RobotName);
    }
    else{

      printf ("MapServer: connection to robot %s failed\n", RobotName);

    }
  }

  else {
    printf ("MapServer: no robot specified (-> no connection)\n");
  }
  
#ifndef VXWORKS  /* xxx Does it work with VxWorks ?  To be tested later ... */
  parseConfigFile("MapServer.ini");
#endif

  /* fetching the connected robot if any */
  RobId = GetConnectedRobot(RobotName,&RobPort);


  /* initializing a dummy world structure. I guess, my map should be included in it. */
  map_serv_world = WLD_InstanciateWorld();
  
  /* Here I have a pb : if we are not connected to a real robot but I still want to initialize
     a world, what should I put here as robot id ? 
     I think I will go through the worldcomm.c and remove restrictions about the wldc_robot_id .
     We are mixing up server and client functions there. Worldcomm.c should probably be
     split in two parts : client and server. 
     For now, I define a negative id. Will see this later.
     Also, I am cheating here by saying this is a real server. The concept of "real" means in fact
     we are a "server" as opposed to a client.
     */
  
  WLDC_Initialize(map_serv_world,-1,1);

  RegisterMapServerCommands();
  while (1){
    CONN_Poll();

    /* the TRACKING_FLG was set when a client requested automatic tracking in the call
       ContinuousTrackingCmd above. Now, I  define a client description structure
       so each client would say if it is interested. This test  checks if
       at least one client is interested.*/
       
    /* if(TRACKING_FLG) */
    
      /*
	PHIL: this  browses the client list to know if somebody is interested in 
	getting the updated position. Jake : this is my point here. No need for polling
	the world : I know when something changed. World polling is not a  common behaviour.
	Each server might want to set this its own way.
      */

    InitReadClientTable();
    while((client_info = GetNextClient())){
      client_desc = (MSERV_ClientDescription *)client_info->user_data;
      if(client_desc->auto_track_wanted){
	TrackOnce(Deviation,VisOp);
	break;
      }
    }
      /* here I should do the same */
    if(MAPING_FLG){
      MapOnce();
      /*sleep(3);*/  
    }


#ifdef VXWORKS
    taskDelay(20);
#endif

  }


}


static void RegisterMapServerCommands()
{

  CMD_Register("OpenSession",OpenSessionCmd );
  CMD_Register("ReadSession",ReadSessionCmd );
  CMD_Register("GetSickData",GetSickDataCmd );
  CMD_Register("GetRecordData",GetSickDataCmd );
  CMD_Register("InitSick",InitSickCmd );
  CMD_Register("GetRobotPosition",GetRobotPositionCmd );
  CMD_Register("ConnectRobot",ConnectRobotCmd );
  CMD_Register("DataIntegration",DataIntegrationCmd );
  CMD_Register("InitRecSession",InitRecSessionCmd );
  CMD_Register("ReadSegmentFile",ReadSegmentFileCmd);
  CMD_Register("RemoveSegments",RemoveSegmentsCmd);
  CMD_Register("InterSegments",InterSegmentsCmd);
  CMD_Register("MergeSegments",MergeSegmentsCmd);
  CMD_Register("ReadMapFile",ReadMapFileCmd);
  CMD_Register("SaveMapFile",SaveMapFileCmd);
  CMD_Register("LocalLocalize",LocalLocalizeCmd);
  CMD_Register("TrackOnce",TrackOnceCmd);
  CMD_Register("ContinuousTrack",ContinuousTrackCmd);
  CMD_Register("NewDataIntegration",NewDataIntegrationCmd);
  CMD_Register("GetRobPositions",GetRobotPositionsCmd);
  CMD_Register("GetRobTransforms",GetRobotPositionsCmd);
  CMD_Register("GetCurrentSegments",GetCurrentSegmentsCmd);
  CMD_Register("GetCurrentMap",GetCurrentMapCmd);
  CMD_Register("ContinuousMaping",ContinuousMapingCmd);
  CMD_Register("GetConnectedRobot",GetConnectedRobotCmd);
  CMD_Register("RotateMap",RotateMapCmd);
  CMD_Register("ChangeMapCoordinates",ChangeMapCoordinatesCmd);
  CMD_Register("GetRobotTransform",GetRobotTransformCmd);
  CMD_Register("GlobalLocalize",GlobalLocalizeCmd);
  CMD_Register("FilterBadHypos",FilterBadHyposCmd);
  RegisterCreateObjectCallback((void *)MapServer_ConnectionCBK);
}











